/* 
 Copyright by The HDF Group.
 All rights reserved.                                                      
                                                                           
 This file is part of the HDF-EOS5 Augmentation Tool. The full HDF-EOS5 
 Augmentation Tool copyright notice including terms governing use, modification,
 and redistribution, is contained in the file COPYING.     
 COPYING can be found at the root of the source code    
 distribution tree. If you do not have access to this file, you may request
 a copy from eoshelp@hdfgroup.org. 
*/
#include "HE5_HdfEosDef.h"
#include "util.h"
#include "read.h"

/** 
 * @brief call the passed @inq function and store the result in @dest
 * 
 * @param ctxid context identifier
 * @param fname HDF-EOS5 filename
 * @param inq inquiry API such as GDinqgrid, SWinqswath or PTinqpoint
 * @param[out] dest this function will allocate storage and fill the result of @inq
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_namelist(hid_t ctxid, const char *fname, long (*inq)(const char *, char *, long *), char **dest)
{
  long bufsize = 0;
  int succeeded = 0;

  *dest = NULL;
  do {
    long numentries = 0;

    if ((numentries = inq(fname, NULL, &bufsize)) == -1) err1_break(1, "inq fails")
    if (numentries == 0) bufsize = 0;
    if ((*dest = malloc(bufsize + 1)) == NULL) nomem_break()
    (*dest)[bufsize] = 0;
    if (numentries > 0) {
      if (inq(fname, *dest, &bufsize) == -1) err1_break(1, "inq_fails (2)")
    }
    succeeded = 1;
  } while (0);

  if (!succeeded) {
    if (*dest != NULL) free(*dest);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get EOS5 the list of dimensions by using the given set of functions
 * 
 * @param ctxid context identifier
 * @param id EOS5 id returned by GDattach or SWattach
 * @param entries GDnentries or SWnentries
 * @param inq GDinqdims or SWinqdims
 * @param[out] dims vector of dimension_t (pairs of name and size)
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_dimension(hid_t ctxid, hid_t id, long (*entries)(hid_t, int, long *), long (*inq)(hid_t, char *, hsize_t *), dimensionalloc *dims)
{
  char *namelist = NULL;
  hsize_t *dimsize = NULL;
  int succeeded = 0;

  do {
    long numdims = 0;
    long bufsize = 0;

    if ((numdims = entries(id, HE5_HDFE_NENTDIM, &bufsize)) == -1) err1_break(1, "dimension entries failure")
    if (numdims > 0) {
      int i = 0;
      int j = 0;
      int failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("dimension namelist")
      namelist[bufsize] = 0;
      if ((dimsize = malloc(numdims * sizeof(hsize_t))) == NULL) nomem1_break("dimension size")
      if (inq(id, namelist, dimsize) == -1) err1_break(1, "dimension inq failure")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  dimension_t *dim;
	  if (!dimensionalloc_readyplus(dims, 1)) nomem1_flagbreak(failinside, "dimension_t")
	  dim = &dims->s[dims->len++];
	  init_dimension_t(dim);
	  ASSERT(dims->len <= (size_t)numdims)
	  {
	    if (!challoc_copyb(&dim->name, namelist + i, j - i)) nomem1_flagbreak(failinside, "namelist")
	    dim->dimsize = dimsize[dims->len - 1];
            dim->maxdimsize = dim->dimsize;
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (namelist != NULL) 
        free(namelist);
    if (dimsize != NULL) 
        free(dimsize);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get field info of EOS5 grid or swath/zonal by using given set of functions
 * 
 * @param ctxid context identifier
 * @param id EOS5 id returned by GDattach or SWattach
 * @param entries GDnentries or SWnentries
 * @param inq GDinqfields, SWinqgeofields or SWinqdatafields
 * @param fldinfo GDfieldinfo or SWfieldinfo
 * @param inqtype HE5_GDinqdatatype
 * @param readfld GDreadfield or SWreadfield
 * @param getfill GDgetfillvalue or SWgetfillvalue
 * @param geofield determine the second argument of @entries function
 * If geofield is nonzero, HDFE_NENTGFLD is used for the second argument of @entries function; otherwise, HDFE_NENTFLD is used
 * @param getattr function to be called to fetch attributes e.g. get_gridlocalattribute
 * @param[out] fields vector of fieldinfo_t
 * 
 *
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_fieldinfo(hid_t ctxid, hid_t id, 
                         long (*entries)(hid_t, int, long *), 
                         long (*inq)(hid_t, char *, int *, hid_t *), 
                         herr_t (*fldinfo)(hid_t, const char *, int *, hsize_t *, hid_t *, char *, char *), 
                         herr_t (*inqtype)(hid_t, const char *, const char *, int, hid_t *, H5T_class_t *, H5T_order_t *, size_t *), 
                         herr_t (*readfld)(hid_t, const char *, const hssize_t *, const hsize_t *, const hsize_t *, void *), 
                         herr_t (*getfill)(hid_t, const char *, void *), 
                         int geofield, 
                         int (*getattr)(hid_t, hid_t, const char *, attributealloc *), 
                         fieldinfoalloc *fields, 
                         namelistalloc* ext_dimnames)
{
  char *namelist = NULL;
  namelistalloc dimnames, maxdimnames;
  int succeeded = 0;

  namelistalloc_init(&dimnames);
  namelistalloc_init(&maxdimnames);

  do {
    long numfields = 0;
    long bufsize = 0;

    /* Obtain the number of fields */
    if ((numfields = entries(id, geofield ? HE5_HDFE_NENTGFLD : HE5_HDFE_NENTDFLD, &bufsize)) == -1) err1_break(1, "fieldinfo entries failure")
    if (numfields > 0) {
      int i = 0;
      int j = 0;
      int failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("fieldinfo namelist")
      namelist[bufsize] = 0;
      if (inq(id, namelist, NULL, NULL) == -1) err1_break(1, "fieldinfo inq failure")

      for (i = j = 0; j <= bufsize; ++j) {

        /* ',' separates the fields, so use ',' to find the next field*/
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  fieldinfo_t *field = NULL;
	  if (!fieldinfoalloc_readyplus(fields, 1)) nomem1_flagbreak(failinside, "fieldinfo_t")
	  field = &fields->s[fields->len++];
	  init_fieldinfo_t(field);
	  ASSERT(fields->len <= (size_t)numfields)
	  {
	    size_t k = 0;
	    int failinside2 = 0;
	    hsize_t numelem = 0;
	    hsize_t dimsize[16]; /* 16 should be enough for the maximum number of dimensions*/
	    char dimlist[512]; /* This number 512 came from one HDF-EOS developer */
            char maxdimlist[512];
	    size_t typesize = 0;
            size_t elemsize = 0;

	    if (!challoc_copyb(&field->name, namelist + i, j - i)) nomem1_flagbreak(failinside, "namelist")

	    /* The prototype of HE5_GDfieldinfo states that 5th argument is of hid_t[] type,
	     * but I think it should not be an array. Its code seems to write only the first
	     * element, and there is no way to recognize how many elements are available.
	     * I will guess only the first element will be used.
	     */
	    {
	      hid_t he5type = 0;
	      if ((fldinfo(id, field->name.s, &field->rank, dimsize, &he5type, dimlist, maxdimlist)) == -1) err1_flagbreak(1, failinside, "fieldinfo")
	    }

	    namelistalloc_empty(&dimnames);
	    namelistalloc_empty(&maxdimnames);
   
            /* Save dimlist and maxdimlist to dimnames C++ like vector structure */
	    if (parse_list(dimlist, strlen(dimlist), ',', &dimnames) == FAIL) err_flagbreak(1, failinside)
	    if (parse_list(maxdimlist, strlen(maxdimlist), ',', &maxdimnames) == FAIL) err_flagbreak(1, failinside)

	    /* assert that rank is equal to number of names from comma-separated dimension list */
	    if (dimnames.len != field->rank) err1_flagbreak(1, failinside, "rank != dimension name list")
	    if (maxdimnames.len != field->rank) err1_flagbreak(1, failinside, "rank != max dimension name list")

            /* Fill in dimension. */
	    for (k = 0; k < dimnames.len; ++k) {
	      dimension_t *dim = NULL;
	      if (!dimensionalloc_readyplus(&field->dims, 1)) nomem1_flagbreak(failinside2, "dimension_t extension")
	      dim = &field->dims.s[field->dims.len++];
	      init_dimension_t(dim);
	      {
		if (!challoc_copy(&dim->name, &dimnames.s[k].name)) nomem1_flagbreak(failinside2, "dimension name copy")
		if (!challoc_copy(&dim->max, &maxdimnames.s[k].name)) nomem1_flagbreak(failinside2, "max dimension name copy")
		dim->dimsize = dimsize[k];

                /* Here we store the information of unlimited dimension for future use.
                   This matters when creating a dimension scale that is unlimited dimension.
                   KY 2012-11-6 */
                if (strcmp(dim->max.s,"Unlim") == 0) 
                    dim->maxdimsize = H5S_UNLIMITED;
                else
                    dim->maxdimsize = dim->dimsize;
	      }
	    }

            /* We encounter a case that the dimension scale field is extensible. This
               will cause the initial dimension size of the field different than the final
               dimension size of the field since the dimension size we obtain is the initial 
               dimension size(based on the HDF-EOS5 APIs). 
               The only way to obtain the extensible field info. from the HDF-EOS5 API is here.
               Since the current design will fill in the data until the last step and once the 
               HDF5 dataset is created, the data space cannot be changed. So the extensible info.
               has to be filled in in the data creation stage. 

               Because so far this happens only in 1-D extensible field and for EOS users, 1-D extensible
               field is a typical case, so we just check the extensible
               info. for 1-D field. Basically we will store the dimension name associated with
               1-D field into a name list.
               KY 2012-11-06
            */

            if (1 == field->rank) {

                /*extensible dim name is only saved when the maximum dimename is Unlim */
                if(strcmp(maxdimnames.s[0].name.s,"Unlim") == 0) {

                    int ext_index = 0;
                    int ext_name_exist = 0;

                    for ( ext_index = 0; ext_index < ext_dimnames->len;ext_index++) {
                        if (strcmp(dimnames.s[0].name.s,ext_dimnames->s[ext_index].name.s) == 0) {
                            ext_name_exist = 1;
                            break;
                        }
                    }

                    /*If no extensible dims, make ext_dimnames the same as dimnames.*/
                    if (0 == ext_name_exist) {
                       namelist_t *elem;
                       if (!namelistalloc_readyplus(ext_dimnames, 1)) { failinside = 1; earlynomem_break() }
                       elem = &ext_dimnames->s[ext_dimnames->len++];
                       init_namelist_t(elem);
                       {
                            if (!challoc_copy(&elem->name, &dimnames.s[0].name)) nomem1_flagbreak(failinside2, "dimension name copy")
                       }
                    }
                }
            }

           
	    if (failinside2) 
            { 
                failinside = failinside2; 
                break; 
            }

	    for (k = 0, numelem = 1; k < field->dims.len; ++k)
	        numelem *= field->dims.s[k].dimsize;

	    {
	      H5T_class_t classid;
	      H5T_order_t order;
	      if (inqtype(id, field->name.s, NULL, geofield ? HE5_HDFE_GEOGROUP : HE5_HDFE_DATAGROUP, &field->type, &classid, &order, &typesize) == -1) err1_flagbreak(1, failinside, "inqtype")
	    }

	    if ((elemsize = H5Tget_size(field->type)) == 0) err1_flagbreak(1, failinside, "H5Tget_size")

#if 0
#ifdef EAGER_DATA_FETCH
	    /* read data */
	    if (!challoc_ready(&field->data, numelem * elemsize)) nomem1_flagbreak(failinside, "fieldinfo data")
	    if (readfld(id, field->name.s, NULL, NULL, NULL, field->data.s) == -1) err1_flagbreak(3, failinside, "readfield failure")
	    field->data.len = (size_t)numelem * elemsize;
#else
	    field->datalen = (size_t)numelem * elemsize;
#endif
#endif

	    /* filler may not exist */
	    if (!challoc_ready(&field->filler, elemsize)) nomem1_flagbreak(failinside, "fieldinfo filler");

	    if (getfill(id, field->name.s, field->filler.s) != -1)
	      field->filler.len = elemsize;

	    /* attribute */
	    if (getattr) {
	      if (getattr(ctxid, id, field->name.s, &field->localattr) == -1) err_flagbreak(1, failinside);
	    }

	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (namelist != NULL) free(namelist);
    namelistalloc_free(&dimnames);
    namelistalloc_free(&maxdimnames);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get a vector of attributes of EOS5 global, grid,swath or zonal average by using given set of functions
 * TODO: I think nobody likes this function which has too many parameters. Is there any type-safe and generic way?
 * 
 * The following parameters that start with HE5_GD can also be replaced by HE5_ZA or HE5_SW to represent 
 * zonal average or swath.
 * @param ctxid context identifier
 * @param id EOS5 id returned by GDattach or SWattach
 * @param kind global or HE5_HDFE_ATTRGROUP, HE5_HDFE_GRPATTRGROUP, HE5_HDFE_LOCATTRGROUP
 * @param fieldname field name for local attributes
 * @param inq3      HE5_EHinqglbattrs, HE5_GDinqattrs, HE5_GDinqgrpattrs 
 * @param inq4                                                            HE5_GDinqlocattrs 
 * @param attrinfo4 HE5_EHglbattrinfo, HE5_GDattrinfo, HE5_GDgrpattrinfo, 
 * @param attrinfo5                                                       HE5_GDlocattrinfo
 * @param inqtype6 HE5_EHinqglbdatatype,
 * @param inqtype8                       HE5_GDinqdatatype
 * @param readattr3 HE5_EHreadglbattr, HE5_GDreadattr, HE5_GDreadgrpattr
 * @param readattr4                                                       HE5_GDreadlocattr
 * @param[out] attrs vector of attributes
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_attribute(hid_t ctxid, hid_t id, int kind, const char *fieldname,
  long (*inq3)(hid_t, char *, long *),
  long (*inq4)(hid_t, const char *, char *, long *),
  herr_t (*attrinfo4)(hid_t, const char *, hid_t *, hsize_t *),
  herr_t (*attrinfo5)(hid_t, const char *, const char *, hid_t *, hsize_t *),
  herr_t (*inqtype6)(hid_t, const char *, hid_t *, H5T_class_t *, H5T_order_t *, size_t *),
  herr_t (*inqtype8)(hid_t, const char *, const char *, int, hid_t *, H5T_class_t *, H5T_order_t *, size_t *),
  herr_t (*readattr3)(hid_t, const char *, void *),
  herr_t (*readattr4)(hid_t, const char *, const char *, void *),
  attributealloc *attrs)
{
  char *namelist = NULL;
  int succeeded = 0;

  ASSERT((inq3 == 0) ^ (inq4 == 0))
  ASSERT((attrinfo4 == 0) ^ (attrinfo5 == 0))
  ASSERT((inqtype6 == 0) ^ (inqtype8 == 0))
  ASSERT((readattr3 == 0) ^ (readattr4 == 0))

  do {
    long numattrs = 0;
    long bufsize = 0;

    if (inq3) {
      /* group attributes */
      if ((numattrs = inq3(id, NULL, &bufsize)) == -1) err1_break(1, "attribute inq failure")
    }
    else {
      /* local field attributes */
      if ((numattrs = inq4(id, fieldname, NULL, &bufsize)) == -1) err1_break(1, "attribute inq failure")
    }
    if (numattrs > 0) {
      int i = 0;
      int j = 0;
      int failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("attribute namelist")
      namelist[bufsize] = 0;
      if (inq3) {
	if (inq3(id, namelist, &bufsize) == -1) err1_break(1, "attribute inq failure (2)")
      }
      else {
	if (inq4(id, fieldname, namelist, &bufsize) == -1) err1_break(1, "attribute inq failure (2)")
      }

      for (i = j = 0; j <= bufsize; ++j) {

        /*',' separates the attributes, so use ',' to find the next attribute. */
	if ((j == bufsize && bufsize) || namelist[j] == ',') {

	  attribute_t *attr = NULL;
	  if (!attributealloc_readyplus(attrs, 1)) nomem1_flagbreak(failinside, "attribute_t")
	  attr = &attrs->s[attrs->len++];
	  init_attribute_t(attr);
	  ASSERT(attrs->len <= (size_t)numattrs)
	  {
	    hsize_t count = 0;
	    size_t typesize = 0;

	    if (!challoc_copyb(&attr->name, namelist + i, j - i)) nomem1_flagbreak(failinside, "attribute name")
	    {
	      hid_t he5type;
	      if (attrinfo4) {
                /* group attribute info */
		if (attrinfo4(id, attr->name.s, &he5type, &count) == -1) err1_flagbreak(1, failinside, "attrinfo failure")
	      }
	      else {
                /* local attribute info */
		if (attrinfo5(id, fieldname, attr->name.s, &he5type, &count) == -1) err1_flagbreak(1, failinside, "attrinfo failure")
	      }
	    }
	    {
	      H5T_class_t classid;
	      H5T_order_t order;
	      if (inqtype6) {
                /* group attribute datatype */
		if (inqtype6(id, attr->name.s, &attr->type, &classid, &order, &typesize) == -1) err1_flagbreak(1, failinside, "inqtype")
	      }
	      else {
                /* local attribute datatype */
		if (inqtype8(id, fieldname, attr->name.s, kind, &attr->type, &classid, &order, &typesize) == -1) err1_flagbreak(1, failinside, "inqtype")
	      }
	    }

	    if (!challoc_ready(&attr->value, (size_t)count * typesize)) nomem1_flagbreak(failinside, "attribute value")
	    if (readattr3) {
              /* group attribute value */
	      if (readattr3(id, attr->name.s, attr->value.s) == -1) err1_flagbreak(1, failinside, "readattr failure")
	    }
	    else {
              /* local attribute value */
	      if (readattr4(id, fieldname, attr->name.s, attr->value.s) == -1) err1_flagbreak(1, failinside, "readattr failure")
	    }
	    attr->value.len = (size_t)count * typesize;
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    if (namelist != NULL) free(namelist);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get EOS5 grid information by using GDgridinfo API
 * 
 * @param ctxid context identifier
 * @param gdid EOS5 id returned by HE5_GDattach
 * @param[out] info result
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_gridinfo(hid_t ctxid, hid_t gdid, gridinfo_t *info)
{
  if (HE5_GDgridinfo(gdid, &info->xdim, &info->ydim, info->upleft, info->lowright) == -1)
    err1_return(1, "GDgridinfo failure")
  return SUCCEED;
}

/** 
 * @brief get projection information by using GDprojinfo API
 * 
 * @param ctxid context identifier
 * @param gdid EOS5 id returned by HE5_GDattach
 * @param grid grid information from HE5_GDgridinfo API
 * @param[out] proj projection information
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_gridprojinfo(hid_t ctxid, hid_t gdid, const gridinfo_t *grid, projinfo_t *proj)
{
  if (HE5_GDprojinfo(gdid, &proj->proj, &proj->zone, &proj->sphere, proj->param) == -1)
    err1_return(1, "GDprojinfo failure")

  if (HE5_GDpixreginfo(gdid, &proj->pix) == -1)
    err1_return(1,"GDpixreginfo failure");

  if(HE5_GDorigininfo(gdid,&proj->origin) == -1)
    err1_return(1,"GDorigininfo failure")

/* The following code can be removed after the next release. The code
   was added because of a bug in HDF-EOS5 APIs. KY 2013-10-18 
*/
#if 0
/* Testing with TES-Aura_L3-CO_r0000002345_F01_05.he5 file. */
if(proj->origin == HE5_HDFE_GD_UL)
  printf("correct \n");
#endif


#if 0
  herr_t          statmeta = FAIL;
  char            *metabuf = (char*)NULL;
  char            *metaptrs[2] = {NULL,NULL};
  char            *utlstr = (char *)NULL;

  context_t *ctx = get_context(ctxid);
  metabuf = (char *)HE5_EHmetagroup(ctx->gridfd, gdname, "g", NULL, metaptrs);

  /* printf("metaptrs[0]=%s metaptrs[1]=%s\n", metaptrs[0], metaptrs[1]); */

  utlstr  = (char *)calloc( HE5_HDFE_UTLBUFSIZE, sizeof(char) );
  statmeta = HE5_EHgetmetavalue(metaptrs, "PixelRegistration", utlstr);

  if(statmeta!=FAIL)
  {
    HE5_GDpixreginfo(gdid, &proj->pix); 
  } else
  {
    proj->pix = HE5_HDFE_CENTER;
  }

  statmeta = HE5_EHgetmetavalue(metaptrs, "GridOrigin", utlstr);

  if(statmeta!=FAIL)
  {
    HE5_GDorigininfo(gdid, &proj->origin);
  } else
  {  
    proj->origin = HE5_HDFE_GD_UL;
  }
#endif 

  return SUCCEED;
}

/** 
 * @brief get swath dimension map info
 * Note: the current augmentation tool doesn't handle the swath dimension map case.
 * @param ctxid context identifier
 * @param swid EOS5 id returned by HE5_SWattach
 * @param[out] maps swath dimension map information from the HE5_SWinqmaps API
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
static int get_swathdimmap(hid_t ctxid, hid_t swid, dimmapalloc *maps)
{
  char *namelist = NULL;
  long *offset = NULL;
  long *increment = NULL;
  namelistalloc name;
  int succeeded = 0;

  namelistalloc_init(&name);
  do {
    long nummaps = 0;
    long bufsize = 0;

    if ((nummaps = HE5_SWnentries(swid, HE5_HDFE_NENTMAP, &bufsize)) == -1) err1_break(3, "dimmap entries")
    if (nummaps > 0) {
      int i = 0;
      int j = 0;
      int failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("dimmap namelist")
      if ((offset = malloc(nummaps * sizeof(long))) == NULL) nomem1_break("dimmap offset")
      if ((increment = malloc(nummaps * sizeof(long))) == NULL) nomem1_break("dimmap increment")
      if (HE5_SWinqmaps(swid, namelist, offset, increment) == -1) err1_break(1, "dimmap")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  dimmap_t *map;
	  if (!dimmapalloc_readyplus(maps, 1)) nomem1_flagbreak(failinside, "dimmap")
	  map = &maps->s[maps->len++];
	  init_dimmap_t(map);
	  ASSERT(maps->len <= (size_t)nummaps)
	  {
	    namelistalloc_empty(&name);
	    if (parse_list(namelist + i, j - i, '/', &name) == FAIL) err_flagbreak(1, failinside)
	    ASSERT(name.len == 2)
	    if (!challoc_copy(&map->geo, &name.s[0].name)) nomem1_flagbreak(failinside, "indexmap geo")
	    if (!challoc_copy(&map->data, &name.s[1].name)) nomem1_flagbreak(failinside, "indexmap data")
	    map->offset = offset[maps->len - 1];
	    map->increment = increment[maps->len - 1];
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    namelistalloc_free(&name);
    if (namelist) free(namelist);
    if (offset) free(offset);
    if (increment) free(increment);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief get swath index map info
 * Note: the current augmentation tool doesn't handle the swath index map case.
 *       In fact, the index map is not used by NASA.
 * @param ctxid context identifier
 * @param swid EOS5 id returned by HE5_SWattach
 * @param[out] maps swath index map information from the HE5_SWinqidxmaps API
 * 
 * @return FAIL if failed, SUCCEED if successful
 */

static int get_swathindexmap(hid_t ctxid, hid_t swid, indexmapalloc *indices)
{
  char *namelist = NULL;
  namelistalloc name;
  int succeeded = 0;

  namelistalloc_init(&name);
  do {
    long numindices = 0;
    long bufsize = 0;

    if ((numindices = HE5_SWnentries(swid, HE5_HDFE_NENTIMAP, &bufsize)) == -1) err1_break(1, "SWnentries HDFE_NENTIMAP failure")
    if (numindices > 0) {
      int i = 0;
      int j = 0;
      int failinside = 0;

      if ((namelist = malloc(bufsize + 1)) == NULL) nomem1_break("indexmap namelist")
      if (HE5_SWinqidxmaps(swid, namelist, NULL) == -1) err1_break(1, "SWinqdimmaps failure")

      for (i = j = 0; j <= bufsize; ++j) {
	if ((j == bufsize && bufsize) || namelist[j] == ',') {
	  indexmap_t *index = NULL;
	  if (!indexmapalloc_readyplus(indices, 1)) nomem1_flagbreak(failinside, "indexmap_t")
	  index = &indices->s[indices->len++];
	  init_indexmap_t(index);
	  ASSERT(indices->len <= (size_t)numindices)
	  {
	    hsize_t dimsize = 0;
	    namelistalloc_empty(&name);
	    if (parse_list(namelist + i, j - i, '/', &name) == FAIL) err_flagbreak(1, failinside)
	    ASSERT(name.len == 2)
	    if (!challoc_copy(&index->geo, &name.s[0].name)) nomem1_flagbreak(failinside, "indexmap geo")
	    if (!challoc_copy(&index->data, &name.s[1].name)) nomem1_flagbreak(failinside, "indexmap data")

	    if ((dimsize = HE5_SWdiminfo(swid, index->geo.s)) == -1) err1_flagbreak(1, failinside, "SWdiminfo failure")
	    if (!longalloc_ready(&index->indices, (size_t)dimsize)) nomem1_flagbreak(failinside, "indexmap data")
	    if (HE5_SWidxmapinfo(swid, index->geo.s, index->data.s, index->indices.s) == -1) err1_flagbreak(1, failinside, "SWidxmapinfo failure")
	    index->indices.len = (size_t)dimsize;
	  }
	  i = j + 1;
	  continue;
	}
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    namelistalloc_free(&name);
    if (namelist != NULL) free(namelist);
  }
  return succeeded ? SUCCEED : FAIL;
}

/* The routines 
   HE5_GDinqdims_long, 
   get_griddimension, 
   get_globalattribute, 
   get_gridattribute,
   get_gridgroupattribute, 
   get_gridlocalattribute,
   HE5_GDinqfields_long

   are wrapper routines of the corresponding HDF-EOS5 APIs and functions defined in this tool.

*/
static long HE5_GDinqdims_long(hid_t gdid, char *dimnames, hsize_t dims[])
{
  return HE5_GDinqdims(gdid, dimnames, dims);
}

static int get_griddimension(hid_t ctxid, hid_t gdid, dimensionalloc *dims)
{
  return get_dimension(ctxid, gdid, HE5_GDnentries, HE5_GDinqdims_long, dims);
}

static int get_globalattribute(hid_t ctxid, hid_t eosfd, attributealloc *attrs)
{
  return get_attribute(ctxid, eosfd, 0, NULL, HE5_EHinqglbattrs, NULL, HE5_EHglbattrinfo, NULL, HE5_EHinqglbdatatype, NULL, HE5_EHreadglbattr, NULL, attrs);
}

static int get_gridattribute(hid_t ctxid, hid_t gdid, attributealloc *attrs)
{
  return get_attribute(ctxid, gdid, HE5_HDFE_ATTRGROUP, NULL, HE5_GDinqattrs, NULL, HE5_GDattrinfo, NULL, NULL, HE5_GDinqdatatype, HE5_GDreadattr, NULL, attrs);
}

static int get_gridgroupattribute(hid_t ctxid, hid_t gdid, attributealloc *attrs)
{
  return get_attribute(ctxid, gdid, HE5_HDFE_GRPATTRGROUP, NULL, HE5_GDinqgrpattrs, NULL, HE5_GDgrpattrinfo, NULL, NULL, HE5_GDinqdatatype, HE5_GDreadgrpattr, NULL, attrs);
}

static int get_gridlocalattribute(hid_t ctxid, hid_t gdid, const char *fieldname, attributealloc *attrs)
{
  return get_attribute(ctxid, gdid, HE5_HDFE_LOCATTRGROUP, fieldname, NULL, HE5_GDinqlocattrs, NULL, HE5_GDlocattrinfo, NULL, HE5_GDinqdatatype, NULL, HE5_GDreadlocattr, attrs);
}

static long HE5_GDinqfields_long(hid_t gdid, char *fieldlist, int rank[], hid_t ntype[])
{
  return HE5_GDinqfields(gdid, fieldlist, rank, ntype);
}

/** 
 * @brief Obtain FillValue with the HE5_GDgetfillvalue function
 * @param gdid HDF-EOS5 grid id obtained from HE5_GDattach
 * @param fieldname The field name that the FillValue is obtained from
 * @param[out] the fillvalue 
 * 
 * @return FAIL if failed, SUCCEED if successful
 */

static herr_t HE5_GDgetfillvalue_const(hid_t gdid, const char *fieldname, void *fillval)
{
    int status=-1;

    char *attrnames = NULL;
    long strbufsize = 0;

    status = HE5_GDinqlocattrs(gdid, fieldname, NULL, &strbufsize);
    attrnames = (char*)malloc(sizeof(char)*(strbufsize+1));
    status = HE5_GDinqlocattrs(gdid, fieldname, attrnames, &strbufsize);

    if(strstr(attrnames, "_FillValue")!=NULL)
      status =  HE5_GDgetfillvalue(gdid, (char *)fieldname, fillval);

    free(attrnames);

    return status;
}

/* The routines 

   get_gridfieldinfo
   get_zonaldimension
   get_swathdimension
   get_zonalattribute
   get_swathattribute
   get_zonalgroupattribute
   get_swathgroupattribute
   get_zonallocalattribute
   get_swathlocalattribute
   HE5_ZAfieldinfo_const
   HE5_SWfieldinfo_const
   HE5_ZAreadfield_const
   HE5_SWreadfield_const

   are wrapper routines of the corresponding HDF-EOS5 APIs and functions defined in this tool.

*/

static int get_gridfieldinfo(hid_t ctxid, hid_t gdid, fieldinfoalloc *fields, namelistalloc* ext_dimnames)
{
  return get_fieldinfo(ctxid, gdid, HE5_GDnentries, HE5_GDinqfields_long, HE5_GDfieldinfo, HE5_GDinqdatatype, HE5_GDreadfield, HE5_GDgetfillvalue_const, 0, get_gridlocalattribute, fields, ext_dimnames);
}

static int get_zonaldimension(hid_t ctxid, hid_t zaid, dimensionalloc *dims)
{
  return get_dimension(ctxid, zaid, HE5_ZAnentries, HE5_ZAinqdims, dims);
}

static int get_swathdimension(hid_t ctxid, hid_t swid, dimensionalloc *dims)
{
  return get_dimension(ctxid, swid, HE5_SWnentries, HE5_SWinqdims, dims);
}

static int get_zonalattribute(hid_t ctxid, hid_t zaid, attributealloc *attrs)
{
  return get_attribute(ctxid, zaid, HE5_HDFE_ATTRGROUP, NULL, HE5_ZAinqattrs, NULL, HE5_ZAattrinfo, NULL, NULL, HE5_ZAinqdatatype, HE5_ZAreadattr, NULL, attrs);
}

static int get_swathattribute(hid_t ctxid, hid_t swid, attributealloc *attrs)
{
  return get_attribute(ctxid, swid, HE5_HDFE_ATTRGROUP, NULL, HE5_SWinqattrs, NULL, HE5_SWattrinfo, NULL, NULL, HE5_SWinqdatatype, HE5_SWreadattr, NULL, attrs);
}

static int get_zonalgroupattribute(hid_t ctxid, hid_t zaid, attributealloc *attrs)
{
    return get_attribute(ctxid, zaid, HE5_HDFE_GRPATTRGROUP, NULL, HE5_ZAinqgrpattrs, NULL, HE5_ZAgrpattrinfo, NULL, NULL, HE5_ZAinqdatatype, HE5_ZAreadgrpattr, NULL, attrs);
}

static int get_swathgroupattribute(hid_t ctxid, hid_t swid, attributealloc *attrs)
{
  return get_attribute(ctxid, swid, HE5_HDFE_GRPATTRGROUP, NULL, HE5_SWinqgrpattrs, NULL, HE5_SWgrpattrinfo, NULL, NULL, HE5_SWinqdatatype, HE5_SWreadgrpattr, NULL, attrs);
}

static int get_zonallocalattribute(hid_t ctxid, hid_t zaid, const char *fieldname, attributealloc *attrs)
{
    return get_attribute(ctxid, zaid, HE5_HDFE_LOCATTRGROUP, fieldname, NULL, HE5_ZAinqlocattrs, NULL, HE5_ZAlocattrinfo, NULL, HE5_ZAinqdatatype, NULL, HE5_ZAreadlocattr, attrs);
}

static int get_swathlocalattribute(hid_t ctxid, hid_t swid, const char *fieldname, attributealloc *attrs)
{
  return get_attribute(ctxid, swid, HE5_HDFE_LOCATTRGROUP, fieldname, NULL, HE5_SWinqlocattrs, NULL, HE5_SWlocattrinfo, NULL, HE5_SWinqdatatype, NULL, HE5_SWreadlocattr, attrs);
}

static herr_t HE5_ZAfieldinfo_const(hid_t zaid,  const char *fieldname, int *rank, hsize_t dims[], hid_t ntype[], char *dimlist, char *maxdimlist)
{
  return HE5_ZAinfo(zaid, (char *)fieldname, rank, dims, ntype, dimlist, maxdimlist);
}

static herr_t HE5_SWfieldinfo_const(hid_t swid, const char *fieldname, int *rank, hsize_t dims[], hid_t ntype[], char *dimlist, char *maxdimlist)
{
  return HE5_SWfieldinfo(swid, (char *)fieldname, rank, dims, ntype, dimlist, maxdimlist);
}

static herr_t HE5_ZAreadfield_const(hid_t zaid, const char *fieldname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[],  void *data)
{
    return HE5_ZAread(zaid, (char *)fieldname, start, stride, edge, data);
}

static herr_t HE5_SWreadfield_const(hid_t swid, const char *fieldname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[],  void *data)
{
  return HE5_SWreadfield(swid, (char *)fieldname, start, stride, edge, data);
}

/** 
 * @brief Obtain FillValue with the HE5_ZAgetfillvalue function
 * @param zaid HDF-EOS5 zonal average id obtained from HE5_ZAattach
 * @param fieldname The field name that the FillValue is obtained from
 * @param[out] the fillvalue 
 * 
 * @return FAIL if failed, SUCCEED if successful
 */

static herr_t HE5_ZAgetfillvalue_const(hid_t zaid, const char *fieldname, void *fillval)
{
  int status=-1;

  char *attrnames = NULL;
  long strbufsize = 0;

  status = HE5_ZAinqlocattrs(zaid, fieldname, NULL, &strbufsize);
  attrnames = (char*)malloc(sizeof(char)*(strbufsize+1));
  status = HE5_ZAinqlocattrs(zaid, fieldname, attrnames, &strbufsize);

  if(strstr(attrnames, "_FillValue")!=NULL)
    status =  HE5_ZAgetfillvalue(zaid, (char *)fieldname, fillval);

  free(attrnames);
  return status; 
}

/** 
 * @brief Obtain FillValue with the HE5_SWgetfillvalue function
 * @param gdid HDF-EOS5 swath id obtained from HE5_SWattach
 * @param fieldname The field name that the FillValue is obtained from
 * @param[out] the fillvalue 
 * 
 * @return FAIL if failed, SUCCEED if successful
 */

static herr_t HE5_SWgetfillvalue_const(hid_t swid, const char *fieldname, void *fillval)
{
  int status=-1;
  
  char *attrnames = NULL;
  long strbufsize = 0;
  status = HE5_SWinqlocattrs(swid, fieldname, NULL, &strbufsize);
  attrnames = (char*)malloc(sizeof(char)*(strbufsize+1));
  status = HE5_SWinqlocattrs(swid, fieldname, attrnames, &strbufsize);

  if(strstr(attrnames, "_FillValue")!=NULL)
    status =  HE5_SWgetfillvalue(swid, (char *)fieldname, fillval);
  
  free(attrnames);
  return status; 
}

/* get_zonalfieldinfo and get_swathfieldinfo are wrappers of get_fieldinfo */

static int get_zonalfieldinfo(hid_t ctxid, hid_t zaid, int geofield, fieldinfoalloc *fields, namelistalloc* ext_dimnames)
{
    return get_fieldinfo(ctxid, zaid, HE5_ZAnentries, HE5_ZAinquire, HE5_ZAfieldinfo_const, HE5_ZAinqdatatype, HE5_ZAreadfield_const, HE5_ZAgetfillvalue_const, geofield, get_zonallocalattribute, fields,ext_dimnames);  
}

static int get_swathfieldinfo(hid_t ctxid, hid_t swid, int geofield, fieldinfoalloc *fields, namelistalloc* ext_dimnames)
{
  return get_fieldinfo(ctxid, swid, HE5_SWnentries, geofield ? HE5_SWinqgeofields : HE5_SWinqdatafields, HE5_SWfieldinfo_const, HE5_SWinqdatatype, HE5_SWreadfield_const, HE5_SWgetfillvalue_const, geofield, get_swathlocalattribute, fields,ext_dimnames);
}

/* The following routine is no longer used. Now comment them out. Remove them in the next release. KY 2013-11-14*/
#if 0
/**
 * @brief test whether the object names in a mapping file are propvided properly
 * @param ctxid context identifier
 * @param object name list
 *
 * @return FAIL if failed, SUCCEED if successful
 */
static int test_object_name_list(hid_t ctxid, char *namelist)
{
  int i = 0;
  int num = 0;
  context_t *ctx = get_context(ctxid);
  char *token = NULL;
  char *c = NULL;

  c = (char*)calloc(strlen(namelist)+1, sizeof(char));
  if(c == NULL) err1_return(3,"Cannot allocate the memory for the name list. ");

  strcpy(c, namelist);

  /* If the HDF-EOS5 file is a single HDF-EOS5 object(e.g. grid), 
     There is no need to provide grid or swath names in the mapping file.
     So ignore the comparsion. Return SUCCEED.
  */
  if(strstr(c, ",")==NULL)
    return SUCCEED; /* single object, no need to test */

  /* 
   * The following testing is only for multiple objects augmented 
   * against the file option. 
   */
  for (token = strtok(/*namelist*/c,","); token != NULL; token = strtok(NULL, ","))
  {
    num++;
    for(i=0; i<ctx->grid_swath_len; i++)
      if(strcmp(ctx->grid_swath_name[i], token)==0)
	break; /* Find the object name in the specified mapping file*/
  
    if(i==ctx->grid_swath_len) /* Do not find the object name in the specified mapping file */
    {
      free(c);
      return FAIL;
    }
  }
   
  free(c);
  return (num==ctx->grid_swath_len)? SUCCEED: FAIL;
}
#endif


/** 
 * @brief opens the given HDF-EOS5 file and fetches the list of grids and swaths
 * 
 * @param ctxid context identifier
 * @param fname the HDF-EOS5 filename to be opened
 * @param aug_option either 1(index), 2(file) or 0(default)
 * @return FAIL if failed, SUCCEED if successful
 */
int open_hdfeos5(hid_t ctxid, const char *fname, const int aug_option)
{
  hid_t gdfd = -1;
  hid_t swfd = -1;
  hid_t zafd = -1;
  char *gridlist = NULL;
  char *szlist = NULL;
  char *szlist2 = NULL;
  char *attrlist = NULL;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  const int FILE_OPTION = 2;

  ASSERT(ctx->gridfd == -1)
  ASSERT(ctx->swathfd == -1)

  do {
    if (ctx == NULL) err1_break(3, "invalid context id")

    /* 1. Open the grid object. 
       2. Obtain the grid name list in C++ vector. 
       Do the same thing for swath and zonal.
    */
    if ((gdfd = HE5_GDopen(fname, H5F_ACC_RDONLY)) == FAIL) err1_break(1, "HE5_GDopen")
    if (get_namelist(ctxid, fname, HE5_GDinqgrid, &gridlist) == FAIL) err1_break(1, "HE5_GDinqgrid")
    if (parse_list(gridlist, strlen(gridlist), ',', &ctx->gridlist) == FAIL) err_break(3)

    if ((swfd = HE5_SWopen(fname, H5F_ACC_RDONLY)) == FAIL) err1_break(1, "HE5_SWopen")
    if (get_namelist(ctxid, fname, HE5_SWinqswath, &szlist) == FAIL) err1_break(1, "HE5_SWinqswath")
    if (parse_list(szlist, strlen(szlist), ',', &ctx->szlist) == FAIL) err_break(1)

    ctx->za_index = ctx->szlist.len;
    
    if((zafd = HE5_ZAopen(fname, H5F_ACC_RDONLY)) == FAIL) err1_break(1, "HE5_ZAopen")
    if(get_namelist(ctxid, fname, HE5_ZAinqza, &szlist2) == FAIL)  err1_break(1, "HE5_ZAinqza")
    if (parse_list(szlist2, strlen(szlist2), ',', &ctx->szlist) == FAIL) err_break(1)

    /*gdfd/zafd/swfd is actually the HDF-EOS5 file ID. So using any one of them is OK to obtain the global attributes. */
    if (get_globalattribute(ctxid, gdfd, &ctx->globalattr) == FAIL) err_break(1)
    
    /* attributealloc attrs = ctx->globalattr;
    for(k=0; k<attrs.len; k++)
    {
      attribute_t *a = &attrs.s[k];
      printf("name=%s value=%s\n", a->name.s, a->value.s);
    } */    

    succeeded = 1;
  } while (0);


  if (1) {
    if (gridlist != NULL) 
       free(gridlist);    
    if (szlist != NULL) 
       free(szlist);
    if (szlist2 != NULL) 
       free(szlist2);
    if (attrlist != NULL) 
       free(attrlist);
  }
  if (succeeded) {
    ctx->infname = fname;
    ctx->gridfd = gdfd;
    ctx->swathfd = swfd;
    ctx->zonalfd = zafd;
  }
  else {
    if (gdfd != -1) HE5_GDclose(gdfd);
    if (swfd != -1) HE5_SWclose(swfd);
    if (zafd != -1) HE5_ZAclose(zafd);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief read data in HDF-EOS5 grid
 * 
 * @param ctxid context identifier
 * @param gridname grid name
 * @param[out] grid all grid information retrieved from the HDF-EOS5 library
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int read_hdfeos5_onegrid(hid_t ctxid, /*const*/ char *gridname, grid_t *grid)
{

  hid_t gdid = -1;
  int succeeded = 0;

  init_grid_t(grid);
  do {
    context_t *ctx = get_context(ctxid);

    if (ctx == NULL) err1_break(3, "invalid context id")

    /* Need to obtain grid info., grid proj. info, grid dimension, grid field info.,
                      grid attribute and grid group attribute.
    */
    if ((gdid = HE5_GDattach(ctx->gridfd, gridname)) == FAIL) err1_break(1, "GDattach failure")
    if (get_gridinfo(ctxid, gdid, &grid->info) == FAIL) err_break(1)
    if (get_gridprojinfo(ctxid, gdid, &grid->info, &grid->proj) == FAIL) err_break(1)
    if (get_griddimension(ctxid, gdid, &grid->dim) == FAIL) err_break(1)
    if (get_gridfieldinfo(ctxid, gdid, &grid->field,&grid->extendible_dimnames) == FAIL) err_break(1)
    if (get_gridattribute(ctxid, gdid, &grid->attr) == FAIL) err_break(1)
    if (get_gridgroupattribute(ctxid, gdid, &grid->grpattr) == FAIL) err_break(1)

    succeeded = 1;
  } while (0);

  if (!succeeded)
    free_grid_t(grid);
  if (1) {
    if (gdid != -1) 
       HE5_GDdetach(gdid);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief read data in HDF-EOS5 zonal average
 * 
 * @param ctxid context identifier
 * @param zonalname zonal average name
 * @param[out] zonal all zonal information retrieved from the HDF-EOS5 library
 * 
 * @return FAIL if failed, SUCCEED if successful
 */

int read_hdfeos5_onezonal(hid_t ctxid, const char *zonalname, swath_zonal_t *zonal)
{
  hid_t zaid = -1;
  int succeeded = 0;

  init_swath_or_zonal_t(zonal);
  do {
      context_t *ctx = get_context(ctxid);
	    
      if (ctx == NULL) err1_break(1, "invalid context id")

      /* Need to obtain zonal dimension, zonal field info.,
                        zonal attribute and zonal group attribute.
      */	  
      if ((zaid = HE5_ZAattach(ctx->zonalfd, zonalname)) == FAIL) err1_break(1, "ZAattach failure")
      if (get_zonaldimension(ctxid, zaid, &zonal->dim) == -1) err_break(1)
      if (get_zonalfieldinfo(ctxid, zaid, 0, &zonal->datafield,&zonal->extendible_dimnames) == -1) err_break(1)
      if (get_zonalattribute(ctxid, zaid, &zonal->attr) == -1) err_break(1)
      if (get_zonalgroupattribute(ctxid, zaid, &zonal->grpattr) == -1) err_break(1)
								  
      succeeded = 1;
    } while (0);
  if (!succeeded)
    free_swath_or_zonal_t(zonal);
  if (1) {
    if (zaid != -1) HE5_ZAdetach(zaid);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief read data in HDF-EOS5 swath
 * 
 * @param ctxid context identifier
 * @param swathname swath name
 * @param[out] swath all swath information retrieved from the HDF-EOS5 library
 * 
 * @return FAIL if failed, SUCCEED if successful
 */
int read_hdfeos5_oneswath(hid_t ctxid, const char *swathname, swath_zonal_t *swath)
{
  hid_t swid = -1;
  int succeeded = 0;

  init_swath_or_zonal_t(swath);
  do {
    context_t *ctx = get_context(ctxid);

    if (ctx == NULL) err1_break(1, "invalid context id")

    /* Need to obtain swath dimension, swath field info.,swath dim. map,
                        swath attributes and swath group attributes.
       swath dimension map and index map are called. But the current augmentation tool
       will not do anything special for the case that have dimension map and index map.
       The file will be augmented just like the swath without the dimension map.
    */
    if ((swid = HE5_SWattach(ctx->swathfd, swathname)) == FAIL) err1_break(1, "SWattach failure")
    if (get_swathdimension(ctxid, swid, &swath->dim) == -1) err_break(1)
    if (get_swathdimmap(ctxid, swid, &swath->dimmap) == -1) err_break(1)
    if (get_swathindexmap(ctxid, swid, &swath->index) == -1) err_break(1)
    if (get_swathfieldinfo(ctxid, swid, 1, &swath->geofield, &swath->extendible_dimnames) == -1) err_break(1)
#if 0
printf("extendible dim. name= %s\n",swath->extendible_dimnames.s[0].name.s); 
printf("extendible dim. len= %d\n",swath->extendible_dimnames.len); 
#endif
    if (get_swathfieldinfo(ctxid, swid, 0, &swath->datafield,&swath->extendible_dimnames) == -1) err_break(1)
    if (get_swathattribute(ctxid, swid, &swath->attr) == -1) err_break(1)
    if (get_swathgroupattribute(ctxid, swid, &swath->grpattr) == -1) err_break(1)

    succeeded = 1;
  } while (0);

  if (!succeeded)
    free_swath_or_zonal_t(swath);
  if (1) {
    if (swid != -1) 
        HE5_SWdetach(swid);
  }
  return succeeded ? SUCCEED : FAIL;
}

/** 
 * @brief close HDF-EOS5 file
 * 
 * @param ctxid context identifier
 * 
 * @return FAIL if failed, SUCCEED if successful
 */

int close_hdfeos5(hid_t ctxid)
{
  context_t *ctx = get_context(ctxid);
  if (ctx != NULL) {
    if (ctx->gridfd != -1) 
       HE5_GDclose(ctx->gridfd);
    ctx->gridfd = -1;
    if (ctx->swathfd != -1)
       HE5_SWclose(ctx->swathfd);
    ctx->swathfd = -1;
    if(ctx->zonalfd != -1) 
       HE5_ZAclose(ctx->zonalfd);
    ctx->zonalfd = -1;
    return 0;
  }
  return -1;
}

