/* 
 Copyright by The HDF Group.
 All rights reserved.                                                      
                                                                           
 This file is part of the HDF-EOS5 Augmentation Tool. The full HDF-EOS5 
 Augmentation Tool copyright notice including terms governing use, modification,
 and redistribution, is contained in the file COPYING.     
 COPYING can be found at the root of the source code    
 distribution tree. If you do not have access to this file, you may request
 a copy from eoshelp@hdfgroup.org. 
*/
#include "hdf5.h"
#include "HE5_HdfEosDef.h"
#include "util.h"
#include "read.h"
#include "write.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#define _C_VECTOR_IMPL
#include "cvct.h"

/*  This constant repesents the initial maximum number of elements if users want 
    to fill in its own dimensional scale with an ASCII file.  
    This maximum number of element of the own dimension scale can increase dynamically.
*/
const int INIT_MAX_NUM_ELMTS_OWN_SCALE = 100;

/* Using the global variable is not a good practice, may tackle this in the next release 
   KY 2013-10-02 
*/
int g_no_log_info=1;


/* This variable will check if the file is augmented or not. This is not the best approach .
   Given that this variable is only used to check if the file is augmented, it shouldn't
   affect other parts of the code. Just leave it here for the time being.
   I may tackle it in the next release if necessary. KY 2013-10-04
*/
int augmented=0;

/* These two global variables are used for the -f option when users can use '2*' to 
   indicate that users just want to fill in nature numbers(index) for the dimension scales 
   of unspecified dimensions in the mapping file.

   May find a way to avoid using global variables.
*/
int *asterisk = NULL;
int asterisk_len = 0;

#define UNLIMIT_NAME "Unlim"

/* If finding that the grid name in the mapping file exists, we need to record the position of the field, ASCII file or index
   indicated in the mapping file. This position will be used in the write.c to fetch the correct field, ASCII file, index
   and to fill in the dim. scale from the corresponding field, ASCII file and index.

   start_ptr_field,end_ptr_field are used to record the starting and the ending position of the field for one grid or swath
   in the mapping file.

   start_ptr_file, end_ptr_file are used to record the starting and the ending position of the data file that store the dim. scale 
   for one grid or swath in the mapping file.

   start_ptr_index, end_ptr_index are used to record the starting and the ending position of the index for one grid or swath 
   in the mapping file.

   These variables are then passed to write.c (with extern) to obtain the corresponding data field, data file or index file info.
*/
int start_ptr_field = 0;
int end_ptr_field   = 0;

int start_ptr_file=0;
int end_ptr_file = 0;

int start_ptr_index=0; 
int end_ptr_index = 0;

/** 
 * @brief one instance of this structure keeps all necessary information for each grid
 * collect_grid_aux() function fills this structure using the HDF-EOS5 library.
 * @see collect_grid_aux
 */
typedef struct {
  challoc gridname;
  grid_t grid;
  int orthogonal;
  int ydimmajor;
  doublealloc lons;
  doublealloc lats;
} grid_aug_t;

void init_grid_aug_t(grid_aug_t *e)
{
  challoc_init(&e->gridname);
  init_grid_t(&e->grid);
  doublealloc_init(&e->lons);
  doublealloc_init(&e->lats);
}
void free_grid_aug_t(grid_aug_t *e)
{
  challoc_free(&e->gridname);
  free_grid_t(&e->grid);
  doublealloc_free(&e->lons);
  doublealloc_free(&e->lats);
}
DECLARE_VECTOR_ALLOCATOR_FREE(gridaugalloc, grid_aug_t)

/** 
 * @brief one instance of this structure keeps all necessary information for each swath.
 * collect_swath_aux() function fills this structure using the HDF-EOS5 library.
 * @see collect_swath_aux
 */
typedef struct {
  challoc swath_or_zonal_name;
  swath_zonal_t swath_or_zonal;
} swath_zonal_aug_t; /* Zonal shares this structure with swath, so it is renamed as swath_zonal_aug_t instead of swath_aug_t. */

void init_swath_or_zonal_aug_t(swath_zonal_aug_t *e)
{
  challoc_init(&e->swath_or_zonal_name);
  init_swath_or_zonal_t(&e->swath_or_zonal);
}

/* Sees this free_swath_zonal_aug_t function has never been used. Need to investigate. KY 2013-10-07 
   Actually it seems that szaugalloc_free or DECLARE_VECTOR_ALLOCATOR_FREE already clean up the memory. 
   KY 2013-10-07 again. */
void free_swath_zonal_aug_t(swath_zonal_aug_t *e)
{
  challoc_free(&e->swath_or_zonal_name);
  free_swath_or_zonal_t(&e->swath_or_zonal);
}
DECLARE_VECTOR_ALLOCATOR_FREE(szaugalloc, swath_zonal_aug_t)

/* This function will shrink lat and lon from 2-D arrays [(YDim,XDim) or (XDim,YDim)] to 1-D arrays */

static int shrink_lonlat(hid_t ctxid, const grid_t *grid, int ydimmajor, doublealloc *lons, doublealloc *lats)
{
  int succeeded = 0;

  do {

    int i = 0;

    /* YDIM major, lat(YDim, XDim) */
    if (ydimmajor == 1) {
      for (i = 0; i < grid->info.ydim; ++i)
	lats->s[i] = lats->s[i * grid->info.xdim];
      lats->len = grid->info.ydim;
      lons->len = grid->info.xdim;
    }
    /* XDIM major. lat(XDim, YDim), really rare KY 2013-10-04 */
    else if (ydimmajor == 0) {
      for (i = 0; i < grid->info.xdim; ++i)
	lons->s[i] = lons->s[i * grid->info.ydim];
      lons->len = grid->info.xdim;
      lats->len = grid->info.ydim;
    }
    else {
      ASSERT(0);
    }

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief retrieve all necessary information for the given grid.
 * 
 * @param ctxid context identifier
 * @param gridname grid name
 * @param[out] gridaux all necessary information for the given grid
 * 
 * @return 0 if successful
 */
static int collect_grid_aux(hid_t ctxid, /*const*/ char *gridname, grid_aug_t *gridaux)
{
  doublealloc whole;
  int succeeded = 0;

  doublealloc_init(&whole);
  do {
    if (!challoc_copys(&gridaux->gridname, gridname)) nomem1_break("grid name")

    /* Read one grid information to gridaux */
    if (read_hdfeos5_onegrid(ctxid, gridname, &gridaux->grid) == -1) err1_break(1, "hdfeos5 read")

    /* After reading all information from the given HDF-EOS5 file, it calculates
     * longitude and latitude values using an HDF-EOS5 function, then detects
     * the major dimension and finally checks whether one-dimensional coordinate
     * variable can be used instead of two-dimensional ones.
     */
    if (generate_lonlat(ctxid, &gridaux->grid, &gridaux->ydimmajor, &gridaux->lons, &gridaux->lats) == -1) err1_break(1, "generate lonlat")
    if ((gridaux->orthogonal = check_ortho_dimscale(&gridaux->grid, gridaux->ydimmajor, &gridaux->lons, &gridaux->lats)) == -1) err1_break(1, "check ortho")
    if (gridaux->orthogonal) {
      if (shrink_lonlat(ctxid, &gridaux->grid, gridaux->ydimmajor, &gridaux->lons, &gridaux->lats) == -1) err_break(1)
    }
    else {
      /* TODO: this has not been tested, yet. */
      /* Seems that we don't need to do anything else. Just keep lat/lon as it is. KY 2013-10-04 */
    }

    succeeded = 1;
  } while (0);

  if (1) {
    doublealloc_free(&whole);
  }
  return succeeded ? 0 : -1;
}
/** 
 * @brief retrieve all necessary information for all grids.
 * 
 * @param ctxid context identifier
 * @param grids includes the name and other info. that will be retrieved from the HDF-EOS5 library and then gathered by
 *              collect_grid_aux. 
 * 
 * @return 0 if successful
 */

static int read_grid(hid_t ctxid, gridaugalloc *grids)
{
  int succeeded = 0;

  do {

    int i = 0;
    int failinside = 0;

    context_t *ctx = get_context(ctxid);

    for (i = 0; i < ctx->gridlist.len; ++i) {

      grid_aug_t *gridaux;

      if (!gridaugalloc_readyplus(grids, 1)) err1_flagbreak(1, failinside, "lonlat")
      gridaux = &grids->s[grids->len++];
      init_grid_aug_t(gridaux);

      if (collect_grid_aux(ctxid, ctx->gridlist.s[i].name.s, gridaux) == -1) err1_flagbreak(1, failinside, "lonlat")
         if(!g_no_log_info)
            msg2("read grid", ctx->gridlist.s[i].name.s);
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief retrieve all necessary information for the given zonal average object.
 * 
 * @param ctxid context identifier
 * @param zaname zonal average name
 * @param[out] zonalaux all necessary information for the given zonal average object
 * 
 * @return 0 if successful
 */

static int collect_zonal_aux(hid_t ctxid, const char *zaname, swath_zonal_aug_t *zonalaux)
{
  doublealloc wholelat;
  int succeeded = 0;

  doublealloc_init(&wholelat);
  do {
    if (!challoc_copys(&zonalaux->swath_or_zonal_name, zaname)) nomem1_break("zonal name")

    /* Read one zonal average info. to zonalaux */
    if (read_hdfeos5_onezonal(ctxid, zaname, &zonalaux->swath_or_zonal) == -1) err1_break(1, "hdfeos5 read")

    succeeded = 1;
  } while (0);

  if (1) {
    doublealloc_free(&wholelat);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief retrieve all necessary information for the given swath.
 * 
 * @param ctxid context identifier
 * @param swathname swath name
 * @param[out] swathaux all necessary information for the given swath
 * 
 * @return 0 if successful
 */
static int collect_swath_aux(hid_t ctxid, const char *swathname, swath_zonal_aug_t *swathaux)
{
  doublealloc wholelat;
  int succeeded = 0;

  doublealloc_init(&wholelat);
  do {
    if (!challoc_copys(&swathaux->swath_or_zonal_name, swathname)) nomem1_break("swath name")

    /* Read one swath info. to swathaux */
    if (read_hdfeos5_oneswath(ctxid, swathname, &swathaux->swath_or_zonal) == -1) err1_break(1, "hdfeos5 read")

    succeeded = 1;
  } while (0);

  if (1) {
    doublealloc_free(&wholelat);
  }
  return succeeded ? 0 : -1;
}
/** 
 * @brief retrieve info. for all zonal average objects by calling collect_zonal_aux.
 * 
 * @param ctxid context identifier
 * @param[out] zas zonal average info.: includes the name and other info. that will be retrieved from the HDF-EOS5 library.
 * 
 * @return 0 if successful
 */
static int read_zonal(hid_t ctxid, szaugalloc *zas)
{
  int succeeded = 0;

  do {
    int i = 0;
    int failinside = 0;
    context_t *ctx = get_context(ctxid);

    /* swath and zonal average are treated mostly the same. Zonal average doesn't have dimension map,
       so we don't need to check if a dimension map exists. In the case when there are both swath and zonal average objects
       in an HDF-EOS5 file, zonal will always be indexed after the swath objects. KY 2013-10-07*/
    for (i = /*0*/ctx->za_index; i < ctx->szlist.len; ++i) {
      swath_zonal_aug_t *zonalaux;

      if (!szaugalloc_readyplus(zas, 1)) err1_flagbreak(1, failinside, "lonlat")
      zonalaux = &zas->s[zas->len++];
      init_swath_or_zonal_aug_t(zonalaux);

      /* collect zonal average info. */
      if (collect_zonal_aux(ctxid, ctx->szlist.s[i].name.s, zonalaux) == -1) 	err1_flagbreak(1, failinside, "lonlat")

      if(!g_no_log_info)
	msg2("read zonal", ctx->szlist.s[i].name.s);

      /* if (swathaux->swath.dimmap.len > 0) err2_flagbreak(3, failinside, swathaux->swathname.s, "has a dimension map.") */
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}
/** 
 * @brief retrieve info. for all swath objects by calling collect_swath_aux.
 * 
 * @param ctxid context identifier
 * @param[out] swaths swath info.: includes the name and other info. that will be retrieved from the HDF-EOS5 library.
 * 
 * @return 0 if successful
 */
static int read_swath(hid_t ctxid, szaugalloc *swaths)
{
  int succeeded = 0;

  do {
    int i =0;
    int failinside = 0;
    context_t *ctx = get_context(ctxid);

    /* swath and zonal average are treated mostly the same. Zonal average doesn't have dimension map,
       so we don't need to check if a dimension map exists. In the case when there are both swath and zonal average objects
       in an HDF-EOS5 file, zonal will always be indexed after the swath objects. KY 2013-10-07*/
    for (i = 0; /*i < ctx->szlist.len*/i < ctx->za_index; ++i) {
      swath_zonal_aug_t *swathaux;

      if (!szaugalloc_readyplus(swaths, 1)) err1_flagbreak(1, failinside, "lonlat")
      swathaux = &swaths->s[swaths->len++];
      init_swath_or_zonal_aug_t(swathaux);

      /* Collect swath info. */
      if (collect_swath_aux(ctxid, ctx->szlist.s[i].name.s, swathaux) == -1) 	err1_flagbreak(1, failinside, "lonlat")
			
      if(!g_no_log_info)
	msg2("read swath", ctx->szlist.s[i].name.s);

      /* For swath, we need to check the dimension map */
      if (swathaux->swath_or_zonal.dimmap.len > 0) err2_flagbreak(1, failinside, swathaux->swath_or_zonal_name.s, "has a dimension map. This is not supported by the tool.")
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief Check whether grid/swath/zonal names exist in the mapping file
 * 
 * @param ctxid context identifier
 * @param grids all necessary information of grids retrieved using the HDF-EOS5 library
 * @param swzas all necessary information of swaths and zonal average objects retrieved using the HDF-EOS5 library 
 * 
 * @return 0 if successful
 */

static int check_obj_names(hid_t ctxid,const gridaugalloc *grids,const szaugalloc *swzas) {

    int succeeded = 0;

    do {

        int i = 0;
        int obj_count = 0;

        context_t *ctx = get_context(ctxid);
 
        /* For one EOS5 object case, no need to check.  */
        if(0==ctx->grid_swath_len) {
            succeeded = 1;
            break;
        }

        int eosobj_index = 0;
        int num_obj_in_ctx = 0;

        /* Check how many grids specified in the mapping file */
        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(ctx->ptr_obj_flag[eosobj_index] == 1) 
                num_obj_in_ctx++;
        }

        /* Check if the grid name specified in the mapping file is the same as that in the HDF-EOS5 file. */
        /* Every grid name needs to be found */
        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(ctx->ptr_obj_flag[eosobj_index] == 1) {
                for (i = 0; i < grids->len; ++i) {
                    const grid_aug_t *gridaux = &grids->s[i];
                    if(strcmp(ctx->grid_swath_name[eosobj_index], gridaux->gridname.s)==0) {
                        obj_count++;             
                        break;
                    }
                }
            }
        }

        if (num_obj_in_ctx != obj_count) 
            break;

        /* Check if the swath names in the mapping file are the same as those in the HDF-EOS5 file. */
        num_obj_in_ctx = 0;
        obj_count      = 0;
        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(ctx->ptr_obj_flag[eosobj_index] == 2) 
                num_obj_in_ctx++;

        }

        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(ctx->ptr_obj_flag[eosobj_index] == 2) {
                for (i = 0; i < ctx->za_index; ++i) {
                    const swath_zonal_aug_t *swathzaaux = &swzas->s[i];
                    if(strcmp(ctx->grid_swath_name[eosobj_index], swathzaaux->swath_or_zonal_name.s)==0) {
                        obj_count++;
                        break;
                    }
                }
            }
        }
        if (num_obj_in_ctx != obj_count) 
            break;


        /* Check if the zonal names in the mapping file are the same as those in the HDF-EOS5 file. */
        num_obj_in_ctx = 0;
        obj_count  = 0;
        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(ctx->ptr_obj_flag[eosobj_index] == 3) 
                num_obj_in_ctx++;
        }

        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(ctx->ptr_obj_flag[eosobj_index] == 3) {
 
                for (i = ctx->za_index; i < ctx->szlist.len; ++i) {
                    const swath_zonal_aug_t *zonalaux = &swzas->s[i];
                    if(ctx->ptr_obj_flag[eosobj_index] == 3) {
                        if(strcmp(ctx->grid_swath_name[eosobj_index], zonalaux->swath_or_zonal_name.s)==0) {
                            obj_count++;
                            break;
                        }
                    }
                }
            }
        }

        if(obj_count != num_obj_in_ctx) 
           break;

        succeeded = 1;

    } while(0);


    return succeeded ?0:-1;

}

/** 
 * @brief Message to users when finding duplicate dimension or field names in the mapping file.
 * 
 * @param dup_name the duplicate name
 * 
 */

void print_dup_message(const char*dup_name) {

    printf("Duplicate dimension name or field name found in the mapping file. \n");
    printf("The duplicate name is %s\n",dup_name);
    printf("Please check the mapping file and make the corrections. \n");

}

/** 
 * @brief Check if the dimension name of an HDF-EOS5 grid in the context exists in the HDF-EOS5 file.
 * 
 * @param ctx context ID
 * @param gridaux all necessary information for the given grid object
 * @param sindex  starting index
 * @param eindex  ending index
 * @param dimscale_type dimensional scale source type(1: scale from a data field, 2: scale from an ASCII file, 3: scale from an index)
 * @return 0 if successful
 */

static int check_one_grid_dim_name(const context_t *ctx,const grid_aug_t *gridaux,const int sindex, const int eindex, const int dimscale_type) {

    int ret_val = 0;
    int fdi     = 0;
    int i       = 0;

    char* dimname = NULL;
    for (fdi = sindex; fdi <eindex; fdi++) {

        /*dim. scale from a field */
        if (dimscale_type == 1) 
            dimname = ctx->dimnames[fdi];
        /* dim. scale from a file */
        else if (dimscale_type == 2)
            dimname = ctx->dim_data_name[fdi];
        /* dim. scale from an index */
        else if (dimscale_type == 3)
            dimname = ctx->dim_index_name[fdi];
        else {
            fprintf(stderr,"Dimensional scale can only come from field, ASCII file or an index, error occurs at %d of file %s .\n",__LINE__,__FILE__);
            ret_val = -2;
        }
            
        if((strcmp(dimname,"XDim") == 0) || (strcmp(dimname,"YDim") == 0))
            continue;
        else {/* Other dim. names. */
            for (i = 0; i<gridaux->grid.dim.len;++i) {
                const dimension_t *dim = &gridaux->grid.dim.s[i];
                if (strcmp(dimname,dim->name.s) == 0)
                    break;
            }
            /* Cannot find this dimension name in the HDF-EOS5 file. */
            if ( i == gridaux->grid.dim.len) {
                ret_val = -1;
                break;
            }
        }
    }

    return ret_val;

}

/** 
 * @brief Check if the dimension name of a swath or a zonal average object in the context exists in the HDF-EOS5 file.
 * 
 * @param ctx context ID
 * @param szaux all necessary information for the given swath or zonal average objects
 * @param sindex  starting index
 * @param eindex  ending index
 * @param dimscale_type dimensional scale source type(1: scale from a data field, 2: scale from an ASCII file, 3: scale from an index)
 * @return 0 if successful
 */

static int check_one_sz_dim_name(const context_t *ctx,const swath_zonal_aug_t *szaux,const int sindex, const int eindex,const int dimscale_type) {

    int ret_val = 0;
    int fdi = 0;
    int i = 0;

    char *dimname = NULL;
    for (fdi = sindex; fdi <eindex; fdi++) {

        /*dim. scale from a field */
        if (dimscale_type == 1) 
            dimname = ctx->dimnames[fdi];
        /* dim. scale from a file */
        else if (dimscale_type == 2)
            dimname = ctx->dim_data_name[fdi];
        /* dim. scale from an index */
        else if (dimscale_type == 3)
            dimname = ctx->dim_index_name[fdi];
        else {
            fprintf(stderr,"Dimensional scale can only come from field, ASCII file or an index, error occurs at %d of file %s .\n",__LINE__,__FILE__);
            ret_val = -2;
        } 

        for (i = 0; i<szaux->swath_or_zonal.dim.len;++i) {
            const dimension_t *dim = &szaux->swath_or_zonal.dim.s[i];
            if (strcmp(dimname,dim->name.s) == 0)
                  break;
        }

        /* Cannot find this dimension name in the HDF-EOS5 file. */
        if ( i == szaux->swath_or_zonal.dim.len) {
            ret_val = -1;
            break;
        }
    }

    return ret_val;
}

/** 
 * @brief Compare the dimension name in the context with the dimension name in the HDF-EOS5 file
 *        Check if the data field name used as the dimension scale in the context exists
 * @param obj_flag object flag (1 is grid, 2 is swath, 3 is zonal average)
 * @param eosobjindex The index of the HDF-EOS5 object in the mapping file 
          (-1 means this file only contains one HDF-EOS5 object,not necessary to specify the object name)
 * @param ctx context struct pointer
 * @param gridaux all necessary information of grids retrieved using the HDF-EOS5 library
 * @param swzas all necessary information of swaths and zonal average objects retrieved using the HDF-EOS5 library 
 * 
 * @return 0 if the dimension name and the field name in the context struct exist
 */
static int compare_one_obj_dim_field(int obj_flag,int eosobjindex,const context_t *ctx,const grid_aug_t *gridaux, const swath_zonal_aug_t *szaux ){


    int ret_val    = 0;

    int sindex_for_field = 0;
    int eindex_for_field = 0;
    int sindex_for_dfile = 0;
    int eindex_for_dfile = 0;
    int sindex_for_index = 0;
    int eindex_for_index = 0;

    if(-1 == eosobjindex) {
        eindex_for_field = ctx->dimgeo_len;
        eindex_for_dfile = ctx->dim_data_len;
        eindex_for_index = ctx->dim_index_len;
    }
    else if(eosobjindex >=0) {
        if(eosobjindex >0) {
            sindex_for_field = ctx->ptr_field[eosobjindex-1];
            sindex_for_dfile = ctx->ptr_file[eosobjindex-1];
            sindex_for_index = ctx->ptr_index[eosobjindex-1];
        }
        eindex_for_field = ctx->ptr_field[eosobjindex];
        eindex_for_dfile = ctx->ptr_file[eosobjindex];
        eindex_for_index = ctx->ptr_index[eosobjindex];
    }
    else {
        fprintf(stderr,"The HDF-EOS5 object index in the mapping file is not right. The error occurs at %d of file %s\n",__LINE__,__FILE__);
        ret_val = -2;
        return ret_val;
    }

    int fdi = 0;
    int i = 0;
    
    /* Check whether the grid dimension name and field name in the file matches those in the mapping file. */
    if (1 == obj_flag) {

        /*First check the dimension name that uses the data field as the dimensional scale */
        ret_val = check_one_grid_dim_name(ctx,gridaux,sindex_for_field,eindex_for_field,1);
        if (ret_val <0) 
            return ret_val;

        /*First check the dimension name that uses the data file as the dimensional scale */
        ret_val = check_one_grid_dim_name(ctx,gridaux,sindex_for_dfile,eindex_for_dfile,2);
        if (ret_val <0)
            return ret_val;

        /*First check the dimension name that uses the index as the dimensional scale */
        ret_val = check_one_grid_dim_name(ctx,gridaux,sindex_for_index,eindex_for_index,3);
        if (ret_val <0)
            return ret_val;

        for (fdi = sindex_for_field; fdi <eindex_for_field; fdi++) {
            for (i = 0; i<gridaux->grid.field.len;++i) {
                const fieldinfo_t *field = &gridaux->grid.field.s[i];
                if (strcmp(ctx->geonames[fdi],field->name.s) == 0)
                    break;
            }
            /* Cannot find this dimension name in the HDF-EOS5 file. */
            if ( i == gridaux->grid.field.len) {
                ret_val = -1;
                break;
            }
        }
        if (ret_val <0)
            return ret_val;
    }

    else if (2 == obj_flag || 3 == obj_flag) {

        /*First check the dimension name that uses the data field as the dimensional scale */
        ret_val = check_one_sz_dim_name(ctx,szaux,sindex_for_field,eindex_for_field,1);
        if (ret_val <0) 
            return ret_val;

        /*First check the dimension name that uses the data file as the dimensional scale */
        ret_val = check_one_sz_dim_name(ctx,szaux,sindex_for_dfile,eindex_for_dfile,2);
        if (ret_val <0)
            return ret_val;

        /*First check the dimension name that uses the index as the dimensional scale */
        ret_val = check_one_sz_dim_name(ctx,szaux,sindex_for_index,eindex_for_index,3);
        if (ret_val <0)
            return ret_val;

        for (fdi = sindex_for_field; fdi <eindex_for_field; fdi++) {
            /* For swath, we pick geofield  */
            if (2 == obj_flag) {
                for (i = 0; i<szaux->swath_or_zonal.geofield.len;++i) {
                    const fieldinfo_t *field = &szaux->swath_or_zonal.geofield.s[i];
                    if (strcmp(ctx->geonames[fdi],field->name.s) == 0)
                        break;
                }
                /* Cannot find this dimension name in the HDF-EOS5 file. */
                if ( i == szaux->swath_or_zonal.geofield.len) {
                    ret_val = -1;
                    break;
                }
            }
            else {
            /* For zonal, only datafield */
                for (i = 0; i<szaux->swath_or_zonal.datafield.len;++i) {
                    const fieldinfo_t *field = &szaux->swath_or_zonal.datafield.s[i];
                    if (strcmp(ctx->geonames[fdi],field->name.s) == 0)
                        break;
                }
                /* Cannot find this dimension name in the HDF-EOS5 file. */
                if ( i == szaux->swath_or_zonal.datafield.len) {
                    ret_val = -1;
                    break;
                }
            }

        }
        if (ret_val <0)
            return ret_val;
    }

    else {

    }
    return 0;
}
/** 
 * @brief Check if dimension names and names of the data fields used as dimension scales exist
 * 
 * @param ctxid context identifier
 * @param grids includes the name and other info. that will be retrieved from the HDF-EOS5 library and then gathered by
  *               collect_grid_aux.
 * @param swathzas  includes the name and other info. that will be retrieved from the HDF-EOS5 library and then gathered by
  *                 collect_swath_aux and collect_zonal_aux.
 * 
 * @return 0 if dimension names or field names exist
 */

static int check_dim_field_names(hid_t ctxid, const gridaugalloc *grids, const szaugalloc *swathzas) {

    int ret_val = 0;

    /* Redundant with this do-while loop, use this way to make this function the same style of others */
    do {

        int i = 0;
        int failinside = 0;

        context_t *ctx = get_context(ctxid);

        if ( 0 == ctx->grid_swath_len) {

            for(i = 0; i < grids->len;i++) { 
                const grid_aug_t *gridaux = &grids->s[i];
                /* -1 in the first parameter, means to check all dim names, field names etc.*/
                /* we need to check all grids and make sure all grids have these dim. names. */
                ret_val = compare_one_obj_dim_field(1,-1,ctx,gridaux,NULL);
                if(ret_val < 0) {
                    failinside = 1;
                    break;
                }
            }
            if (failinside == 1)
                break;

            /* Add swath zonal here. */
            for(i = 0; i < swathzas->len;i++) { 
                const swath_zonal_aug_t *szaux = &swathzas->s[i];
                /* -1 in the first parameter, means to check all dim names, field names etc.*/
                /* we need to check all grids and make sure all grids have these dim. names. */
                ret_val = compare_one_obj_dim_field((i<ctx->za_index)?2:3,-1,ctx,NULL,szaux);
                if(ret_val <0) {
                    failinside = 1;
                    break;
                }
            }

            if (failinside == 1)
                break;

       }
        else {

            int eosobj_index = 0;
          
            for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
                if(1 == ctx->ptr_obj_flag[eosobj_index]) {
                    for (i = 0; i < grids->len; ++i) {
                        const grid_aug_t *gridaux = &grids->s[i];

                        /* If we find this grid, we should go to the outer for loop. */
                        if(strcmp(ctx->grid_swath_name[eosobj_index], gridaux->gridname.s) == 0) { /*Find the grid */
                            ret_val = compare_one_obj_dim_field(1,eosobj_index,ctx,gridaux,NULL) ;
                            if(ret_val <0) {
                                failinside = 1;
                            }
                            break;
                        }
                    }
                    if (1 == failinside)
                        break;
                }
                else if((2 == ctx->ptr_obj_flag[eosobj_index]) || (3 == ctx->ptr_obj_flag[eosobj_index])) {
                    for (i = 0; i < swathzas->len; ++i) {
                        const swath_zonal_aug_t *szaux = &swathzas->s[i];

                        /* If we find this swath or zonal average, we should go to the outer for loop. */
                        if(strcmp(ctx->grid_swath_name[eosobj_index], szaux->swath_or_zonal_name.s) == 0) { /*Find the swath or zonal */
                            ret_val = compare_one_obj_dim_field((ctx->ptr_obj_flag[eosobj_index]==2)?2:3,
                                                           eosobj_index,ctx,NULL,szaux) ;
                            if(ret_val <0) {
                                failinside = 1;
                            }
                            break;
                        }
                    }
                    if (1 == failinside)
                        break;
                }
            }

            if(1 == failinside) 
                break;
        }
        ret_val = 0;
    } while (0);

   return ret_val;

}

/** 
 * @brief augment one grid using the HDF5 library
 * 
 * @param ctxid context identifier
 * @param gridaux all necessary information retrieved using the HDF-EOS5 library
 * @param aug_option: either index(1) or file (2) or default(0) 
 * 
 * @return 0 if successful
 */
static int augment_onegrid(hid_t ctxid, const grid_aug_t *gridaux, const int aug_option)
{
  challoc coordinates;
  int succeeded = 0;

  challoc_init(&coordinates);
  do {

    int i =0;
    int failinside = 0;
    int xdimexplicit = 0;
    int ydimexplicit = 0;

    context_t *ctx = get_context(ctxid);

    /* augmentmethod reflects 3 solutions documented by an investigation report.
    The document can be found under 
    http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
    Finally only solution 1 was accepted although the tool also implemented other two solutions.
    Only solution 1 is used by NASA and we only test solution 1.

    For solution 1,we provide 3 options(index,file or default) that are flagged by aug_option.
    */

    switch (ctx->augmentmethod) {
      case 1:
	break;
      case 2:
      case 3:
	if (create_aug_group(ctxid, gridaux->gridname.s, 1) == -1) err1_flagbreak(2, failinside, "create aug group")
	break;
      default:
	ASSERT(0);
    }
    if (failinside) break;

    /* Create an HDF5 dimension scale for each HDF-EOS5 dimension */
    for (i = 0; i < gridaux->grid.dim.len; ++i) {

      const dimension_t *dim = &gridaux->grid.dim.s[i];

      int isxdim = -1;
      int isydim = -1;
      int ispure = 1;

      /* This part of logic may not apply to HDF-EOS5 since XDim and YDim don't appear as the dimension name
         in the grid dimension name list. I agree that it is a sound redundance code. 
         Will see if needing to revise it later. KY 2013-10-25 */
      isxdim = strcmp(dim->name.s, "XDim") == 0;
      isydim = strcmp(dim->name.s, "YDim") == 0;
      if ((isxdim || isydim) && gridaux->orthogonal)
         ispure = 0;

      if(strcmp(dim->name.s,UNLIMIT_NAME)==0) {
          continue;
      }

      if (write_hdf5_dimension(ctxid, gridaux->gridname.s, 1, ispure, dim,&(gridaux->grid.extendible_dimnames),aug_option) == -1) err1_flagbreak(2, failinside, "aug dimension")
      if (isxdim) xdimexplicit = 1;
      if (isydim) ydimexplicit = 1;
    }
    if (failinside) break;

    /* Create two HDF5 dimension scales if they are implicit in the HDF-EOS5 file. Generally the dimension names should be
       XDim or YDim.  */
    if (!xdimexplicit) {
      if (write_hdf5_implicit_dimension(ctxid, gridaux->gridname.s, &gridaux->grid, 1, !gridaux->orthogonal,aug_option) == -1) err1_break(2, "aug x dimension")
    }
    if (!ydimexplicit) {
      if (write_hdf5_implicit_dimension(ctxid, gridaux->gridname.s, &gridaux->grid, 0, !gridaux->orthogonal,aug_option) == -1) err1_break(2, "aug y dimension")
    }

    /* Associate created HDF5 dimension scales with corresponding HDF5 datasets */
    for (i = 0; i < gridaux->grid.field.len; ++i) {
      const fieldinfo_t *field = &gridaux->grid.field.s[i];
      if (associate_hdf5_dimension(ctxid, gridaux->gridname.s, 1, field, 1) == -1) err2_flagbreak(2, failinside, "assoc dimension", field->name.s)
    }
    if (failinside) break;

    if (gridaux->orthogonal) {

       /*Fill two special HDF5 dimension scales, XDim and YDim since HDF-EOS5 doesn't store XDim and YDim as their dimension names.  */
       if (fill_hdf5_1d_dimension(ctxid,&gridaux->grid,gridaux->gridname.s,1,&gridaux->lons,aug_option)==-1) err1_break(2,"fill 1d lon");
       if (fill_hdf5_1d_dimension(ctxid,&gridaux->grid,gridaux->gridname.s,0,&gridaux->lats,aug_option)==-1) err1_break(2,"fill 1d lat");

      for (i = 0; i < gridaux->grid.dim.len; ++i) {
      	const dimension_t *dim = &gridaux->grid.dim.s[i];

        if(strcmp(dim->name.s, "XDim") == 0)
          continue;

        if(strcmp(dim->name.s, "YDim") == 0)
          continue;

	if(fill_hdf5_1d_3rd_dimension(ctxid, gridaux->gridname.s, dim,aug_option) == -1) err1_flagbreak(2, failinside,"fill 1d the third dimension")
      }

      if (failinside) break;

      /* Write attributes to conform to the CF Conventions */
      if (conform_cf_conventions_grid_1d(ctxid, gridaux->gridname.s, 1) == -1) err1_break(2, "conform to CF conventions lon")
      if (conform_cf_conventions_grid_1d(ctxid, gridaux->gridname.s, 0) == -1) err1_break(2, "conform to CF conventions lat")
    }
    else {
      /* Fill two coordinate variables with longitude and latitude values */
      if (fill_hdf5_2d_dimension(ctxid, gridaux->gridname.s, 1, &gridaux->lons, gridaux->ydimmajor, gridaux->grid.info.xdim, gridaux->grid.info.ydim) == -1) err1_break(2, "fill 2d lon")
      if (fill_hdf5_2d_dimension(ctxid, gridaux->gridname.s, 0, &gridaux->lats, gridaux->ydimmajor, gridaux->grid.info.xdim, gridaux->grid.info.ydim) == -1) err1_break(2, "fill 2d lat")

      /* Write attributes to conform to the CF Conventions */
      if (conform_cf_conventions_grid_2d_dim(ctxid, gridaux->gridname.s, 1) == -1) err1_break(2, "conform to CF conventions XDim")
      if (conform_cf_conventions_grid_2d_dim(ctxid, gridaux->gridname.s, 0) == -1) err1_break(2, "conform to CF conventions YDim")
      if (conform_cf_conventions_grid_2d_coordvar(ctxid, gridaux->gridname.s, 1, &coordinates) == -1) err1_break(2, "conform to CF conventions lon")
      if (conform_cf_conventions_grid_2d_coordvar(ctxid, gridaux->gridname.s, 0, &coordinates) == -1) err1_break(2, "conform to CF conventions lat")

      /* Write "coordinates" attribute for each data field */
      for (i = 0; i < gridaux->grid.field.len; ++i) {
	const fieldinfo_t *field = &gridaux->grid.field.s[i];
	if (conform_cf_conventions_2d_data(ctxid, gridaux->gridname.s, 1, field, &coordinates) == -1) err2_flagbreak(2, failinside, "conform to CF conventions", field->name.s)
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&coordinates);
  }
  return succeeded ? 0 : -1;
}

/**
 * @brief get attibute value of InstrumentName
 * @param ctxid context identifier
 */
static int get_instrument_name(hid_t ctxid) 
{
  /* This definition causes H5Aread does not return the correct value on 64-bit Linux machine. malloc does not work correctly. [06/28/2012 -LD] 
  char *buf = malloc(80*sizeof(char)); */
  /* If errors occur in this routine, the routine will return nothing. Resources will not be released and memory leaking may occur.
     Fix this by returning FAIL. The routines that call this routine will release the resources. 
     KY 2013-10-09
  */

 int succeeded = 0;
	 
 do {
        context_t *ctx = get_context(ctxid);
        char *path = "/HDFEOS/ADDITIONAL/FILE_ATTRIBUTES/";
        htri_t has_eos_fattr_group = -1;
        hid_t eos_fattr_gid        = -1;
	
        has_eos_fattr_group = H5Lexists(ctx->hdf5fd,path,H5P_DEFAULT);

        if (has_eos_fattr_group > 0) {

            if((eos_fattr_gid = H5GOPEN(ctx->hdf5fd, path)) < 0) err1_break(3, "H5GOPEN")

	          hid_t attr = H5Aexists(eos_fattr_gid, "InstrumentName");
	          if(0 == attr) {
                succeeded = 1;
		            break; /* no such attribute*/
            }
            if (attr < 0) {
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aexists") 
            }

	          if((attr = H5Aopen(eos_fattr_gid, "InstrumentName", H5P_DEFAULT)) < 0)  {
                 H5Gclose(eos_fattr_gid);
                 err1_break(3,"H5Aopen");
            }

            hid_t atype = H5Aget_type(attr);
            if(atype <0) {
                H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }

            hid_t atype_mem = H5Tget_native_type(atype, H5T_DIR_ASCEND);
            if(atype_mem <0) {
                H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }

            hid_t aspace = H5Aget_space(attr);
            if(aspace < 0) {
                H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }

            hssize_t nr_elms = H5Sget_simple_extent_npoints(aspace);
            if(nr_elms < 0) {
                H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }

            hsize_t asize = H5Tget_size(atype);
            if(asize == 0) {
                H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }
 
	          char *buf = malloc((asize*nr_elms+1)*sizeof(char));
            if (NULL == buf) {
                H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }
 
            herr_t ret = H5Aread(attr, atype_mem, buf); 
            if(ret <0) {
	              H5Aclose(attr);
                H5Gclose(eos_fattr_gid);
                err1_break(3,"H5Aget_type");
            }
 
            ctx->instrument_name = strdup(buf);
	          ctx->len_of_instrument_name = nr_elms*asize;

	          H5Aclose(attr);
            H5Gclose(eos_fattr_gid);
            free(buf);
        }

        succeeded = 1;
    } while(0);
    return succeeded ? 0 : -1;
}

/* The following routines and struct visit_data_t,attr_info, test_dimension_list and test_augmentation
   are used to check if the file is augmented or not. If the file is augmented already, the program
   will exit. */

typedef struct {
	hid_t ctxid;
	const challoc *srcpath;
} visit_data_t;

/* @brief This routine will check if an attribute called "DIMENSION_LIST" exists. 
          This is the operator called by H5Aiterate2. 
   @param loc_id      The location identifier for the group or dataset being iterated over
   @param attr_name   The name of the current object attribute
   @param ainfo       The attribute's info struct
   @param opdata      a pointer to the operator data passed into H5Aiterate2.
   
*/
static herr_t attr_info(hid_t loc_id, const char *attr_name, const H5A_info_t *ainfo, void *opdata) {
  if(strcmp(attr_name, "DIMENSION_LIST") == 0){/* attribute DIMENSION_LIST exists! */
      augmented = 1; 
  } 

  return 0;
}

static herr_t attr_info1(hid_t loc_id, const char *attr_name, void *opdata) {
  if(strcmp(attr_name, "DIMENSION_LIST") == 0){/* attribute DIMENSION_LIST exists! */
      augmented = 1; 
  } 

  return 0;
}

/**
 * @brief test whether DIMENSION_LIST exists, this is the operator called by H5Giterate.
 * @param srcgrp The group identifier for the group being iterated over
 * @param name   The name of the current object within the group
 * @param data   The pointer to the operator data passed in to H5Giterate, operator_data.
 */

#if defined H5_USE_16_API
static herr_t test_dimension_list(hid_t srcgrp, const char *name, void *data) {

  challoc srcfullpath;
  hid_t source = -1;
  int succeeded = 0;
  int failinside = 0;

  challoc_init(&srcfullpath);
  do {
    const visit_data_t *visitdata = data;
    hid_t ctxid = visitdata->ctxid;
    context_t *ctx = get_context(ctxid);
    H5G_stat_t objinfo;
    
    if(!challoc_copy(&srcfullpath, visitdata->srcpath)) nomem1_break("srcpath")
    if(!challoc_cats(&srcfullpath, "/")) nomem1_break("srcpath")
    if(!challoc_cats(&srcfullpath, name)) nomem1_break("srcpath")

    if(H5Gget_objinfo(srcgrp, name, 0, &objinfo) < 0) err2_break(2, "get_objinfl", srcfullpath.s)
   
    switch(objinfo.type) {
      case H5G_GROUP: 
      {
    	  visit_data_t nested;
    	  nested.ctxid = ctxid;
    	  nested.srcpath = &srcfullpath;
    
              /* Iterate the whole group until going to an HDF5 dataset.*/
    	  H5Giterate(ctx->hdf5fd, srcfullpath.s, NULL, test_dimension_list, &nested);
	    }
        break;
      case H5G_DATASET:
	    {

    	  if((source = H5DOPEN(ctx->hdf5fd, srcfullpath.s)) < 0) err2_break(2, "open dataset", srcfullpath.s)
    
              /*Iterate all the attributes of this dataset to see if the "DIMENSION_LIST" attribute exists.*/
    	  //herr_t ret =  H5Aiterate2(source, H5_INDEX_NAME, H5_ITER_INC, NULL, attr_info, NULL);
	herr_t ret = H5Aiterate1(source,NULL,attr_info1,NULL);
        if (ret <0) {
            failinside = 1;
            H5Dclose(source);
            err1_break(2,"H5Aiterate2");
        }
              
    	  if(source != -1) 
                H5Dclose(source);

      } 
       break;
       default:
         ;
    }
    if (failinside == 1)
       break;

    succeeded = 1;
  } while(0);

  if(1) {
    challoc_free(&srcfullpath);
  }

  return succeeded? 0: -1; 
}

#else
static herr_t test_dimension_list(hid_t srcgrp, const char *name, const H5L_info_t *linfo, void *data) {

  challoc srcfullpath;
  hid_t source = -1;
  int succeeded = 0;
  int failinside = 0;
  challoc_init(&srcfullpath);
  do {
    const visit_data_t *visitdata = data;
    hid_t ctxid = visitdata->ctxid;
    context_t *ctx = get_context(ctxid);
    //H5G_stat_t objinfo;
    H5O_info_t objinfo;
    
    if(!challoc_copy(&srcfullpath, visitdata->srcpath)) nomem1_break("srcpath")
    if(!challoc_cats(&srcfullpath, "/")) nomem1_break("srcpath")
    if(!challoc_cats(&srcfullpath, name)) nomem1_break("srcpath")

    //if(H5Gget_objinfo(srcgrp, name, 0, &objinfo) < 0) err2_break(2, "get_objinfl", srcfullpath.s)
    //if (H5Oget_info_by_name2(srcgrp, name,  &objinfo,H5O_INFO_BASIC,H5P_DEFAULT) < 0) err2_break(2, "get objinfo", srcfullpath.s)
    if (H5OGET_INFO_BY_NAME(srcgrp, name,  &objinfo,H5P_DEFAULT) < 0) err2_break(2, "get objinfo", srcfullpath.s)
 
    switch(objinfo.type) {
      case H5O_TYPE_GROUP: 
      {
    	  visit_data_t nested;
    	  nested.ctxid = ctxid;
    	  nested.srcpath = &srcfullpath;
    
          /* Iterate the whole group until going to an HDF5 dataset.*/
// Need to use H5Literate_by_name, the same as H5Giterate
// idx_g = H5Literate_by_name(grp, "/Data_new", H5_INDEX_NAME, H5_ITER_INC, NULL, group_info, NULL, H5P_DEFAULT);
// H5Literate_by_name( hid_t loc_id, const char *group_name, H5_index_t index_type, H5_iter_order_t order, hsize_t *idx, 
//                     H5L_iterate_t op, void *op_data, hid_t lapl_id )
// int H5Giterate(hid_t loc_id, const char *name, int *idx, H5G_iterate_t operator, void *operator_data )
    	  //H5Giterate(ctx->hdf5fd, srcfullpath.s, NULL, test_dimension_list, &nested);
          H5Literate_by_name(ctx->hdf5fd, srcfullpath.s, H5_INDEX_NAME, H5_ITER_INC, NULL,test_dimension_list, &nested,H5P_DEFAULT);
	    }
        break;
      case H5O_TYPE_DATASET:
	    {

    	  if((source = H5DOPEN(ctx->hdf5fd, srcfullpath.s)) < 0) err2_break(2, "open dataset", srcfullpath.s)
    
              /*Iterate all the attributes of this dataset to see if the "DIMENSION_LIST" attribute exists.*/
        herr_t ret = H5Aiterate2(source, H5_INDEX_NAME, H5_ITER_INC, NULL, attr_info, NULL);
        if (ret <0) {
            failinside = 1;
            H5Dclose(source);
            err1_break(2,"H5Aiterate2");
        }
              
    	  if(source != -1) 
                H5Dclose(source);

      } 
       break;
       default:
         ;
    }
    if (failinside == 1)
       break;

    succeeded = 1;
  } while(0);

  if(1) {
    challoc_free(&srcfullpath);
  }

  return succeeded? 0: -1; 
}

#endif

/**
 * @brief test whether this HDF-EOS5 file is augmented.
   @param ctxid         context identifier
   @param gridswathname grid or swath name
   @param isgrid        Is this group an HDF-EOS5 grid?
   @param isswath       Is this group an HDF-EOS5 swath?
 */
static int test_augmentation(hid_t ctxid, 
	const char *gridswathname, int isgrid, int isswath) {
  challoc srcpath;
  challoc_init(&srcpath);

  int succeeded = 0;
  do {
    context_t *ctx = get_context(ctxid);
    visit_data_t visitdata;

    if(isgrid){
    	if(!challoc_copys(&srcpath, "/HDFEOS/GRIDS/")) nomem1_break("/")
    } else {
	    if(!challoc_copys(&srcpath, (isswath)? "/HDFEOS/SWATHS/" : "/HDFEOS/ZAS/")) nomem1_break("/")
    }
    challoc_cats(&srcpath, gridswathname);
    challoc_cats(&srcpath, "/Data Fields");
   
    visitdata.ctxid = ctxid;
    visitdata.srcpath = &srcpath;

    /* Iterate the group and check if "DIMENSION_LIST" is found */
#if defined H5_USE_16_API
    H5Giterate(ctx->hdf5fd, srcpath.s, NULL, test_dimension_list, &visitdata);
#else 
    H5Literate_by_name(ctx->hdf5fd, srcpath.s, H5_INDEX_NAME, H5_ITER_INC, NULL,test_dimension_list, &visitdata,H5P_DEFAULT);
#endif
   
    succeeded = 1;
  } while(0);

  if(1) {
    challoc_free(&srcpath);
  }

  return succeeded? 0: -1;
}

/**
  *  @brief Obtain the positions of data fields, data files or data indices from the user's dim to dim. scale mapping file. 
            Note for the default option and the index option, this function is not called.
     @param ctxid context identifier
     @param gridswathname the grid or swath name
  *
  */
void scale_storage_pos_in_mfile(hid_t ctxid, const char *gridswathname) 
{

  context_t *ctx = get_context(ctxid);
  int len = ctx->grid_swath_len;
  int k = 0;
  int find_obj_flag = 0;

  /* If the grid name in the HDF-EOS5 file matches the grid name in the mapping file, out of the loop. */ 
  for(k=0; k<len; k++) 
  {
    if(strcmp(gridswathname, ctx->grid_swath_name[k]) == 0){
      find_obj_flag = 1;
      break;
    }
  }

  /* If finding that the grid name in the mapping file exists, record the position of the field, file or index
     specified in the mapping file. This position will be used in the write.c to fetch the correct field, file, index
     and to fill in the dim. scale from the corresponding field, file and index.
     For example, for the followed mapping file: 
     "
      grid name: NadirDayGrid

      0 nLevels Pressure

      0 nLevel1s Pressure1
      "

      '0' indicates the dim. scale is provided by a data field. The start_ptr_field is 0 and the end_ptr_field is 2.
      The start_ptr_field and the end_ptr_field will be used to identify (nLevels, Pressure) and (nLevel1s,Pressure1).
      
  */

  if(find_obj_flag)
  {
    if(k != 0)
    {
      start_ptr_field = ctx->ptr_field[k-1];
      start_ptr_file  = ctx->ptr_file[k-1];
      start_ptr_index = ctx->ptr_index[k-1];
    }
    end_ptr_field   = ctx->ptr_field[k];
    end_ptr_file    = ctx->ptr_file[k];
    end_ptr_index   = ctx->ptr_index[k];
  } 
  /* This else block may be redundant. Check in the future. KY 2013-11-27 */
  else {
    start_ptr_field = 0;
    start_ptr_file  = 0;
    start_ptr_index = 0;
    end_ptr_field   = 0;
    end_ptr_file    = 0;
    end_ptr_index   = 0;
  }

}

/**
  *  @brief Augment grids.
     @param ctxid context identifier
     @param grids includes the name and other info. that will be retrieved from the HDF-EOS5 library and then gathered by
  *               collect_grid_aux. 
  *  @param aug_option: either index(1) or file(2) or default(0)

     @return 0 if successful, -1 if failed.
  */

static int augment_grids(hid_t ctxid, const gridaugalloc *grids, const int aug_option)
{
  int succeeded = get_instrument_name(ctxid);

  /* If get_instrument_name() fails, returns FAIL immediately */
  if (-1 == succeeded){
      warn1(3,"get_instrument_name");
      return succeeded;
  }

  do {

    int i = 0;
    int failinside = 0;

    context_t *ctx = get_context(ctxid);

    end_ptr_field = ctx->dimgeo_len;
    end_ptr_file  = ctx->dim_data_len;
    end_ptr_index = ctx->dim_index_len;

    int len = ctx->grid_swath_len;

    for (i = 0; i < grids->len; ++i) {
      const grid_aug_t *gridaux = &grids->s[i];

      if(i == 0) /* test the first grid only */
      {
        test_augmentation(ctxid, gridaux->gridname.s, 1, 0);
        if(augmented == 1) 
            break;
      }

      /* Obtain the input from the user's dim to dim. scale mapping file. 
         Note for the default option and the index option, this function is not called.
         KY 2013-10-15
      */
      if(len != 0) 
	      scale_storage_pos_in_mfile(ctxid, gridaux->gridname.s); 

      if(!g_no_log_info)
		    msg_push(gridaux->gridname.s);
      if (augment_onegrid(ctxid, gridaux,aug_option) == -1) err1_flagbreak(2, failinside, "aug onegrid")
      if(!g_no_log_info)
	      msg_pop();

    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}


/** 
 * @brief augment one swath or zonal average using the HDF5 library
 * 
 * @param ctxid context identifier
 * @param swathzaaux all necessary information retrieved using the HDF-EOS5 library
 * @param aug_option: either index(1) or file (2) or default(0)   
 * @return 0 if successful,-1 if failed.
 */
static int augment_oneswathza(hid_t ctxid, const swath_zonal_aug_t *swathzaaux,const int aug_option)
{
  challoc coordinates;
  int succeeded = 0;

  challoc_init(&coordinates);
  do {
    int i =0;
    int failinside = 0;
    context_t *ctx = get_context(ctxid);

    switch (ctx->augmentmethod) {
      case 1:
	        break;
      case 2:
      case 3:
	      err1_flagbreak(2, failinside, "The augmented method 2 and 3 are not supported.")
	        break;
      default:
	      ASSERT(0);
    }
    if (failinside) break;

    /* Create an HDF5 dimension scale for each HDF-EOS5 dimension */
    for (i = 0; i < swathzaaux->swath_or_zonal.dim.len; ++i) {
      const dimension_t *dim = &swathzaaux->swath_or_zonal.dim.s[i];

#if 0
printf("maximum dimension name= %s\n",dim->max.s);
printf("swath name =%s\n",swathzaaux->swath_or_zonal_name.s);
#endif

      if(strcmp(dim->name.s,UNLIMIT_NAME)==0) {
          continue;
/*
printf("dimension name= %s\n",dim->name.s);
*/

      }
      if (write_hdf5_dimension(ctxid, swathzaaux->swath_or_zonal_name.s, 0, 1, dim,&(swathzaaux->swath_or_zonal.extendible_dimnames),aug_option) == -1) err1_flagbreak(2, failinside, "aug dimension")
    }

    if (failinside) break;

    /* Associate created HDF5 dimension scales with corresponding HDF5 datasets originated from HDF-EOS5 data fields */
    for (i = 0; i < swathzaaux->swath_or_zonal.datafield.len; ++i) {
      const fieldinfo_t *datafield = &swathzaaux->swath_or_zonal.datafield.s[i];
      if (associate_hdf5_dimension(ctxid, swathzaaux->swath_or_zonal_name.s, 0, datafield, 1) == -1) err1_flagbreak(2, failinside, "assoc dimension")
    }
    if (failinside) break;

    /* Associate created HDF5 dimension scales with corresponding HDF5 datasets originated from HDF-EOS5 geo-location fields */
    for (i = 0; i < swathzaaux->swath_or_zonal.geofield.len; ++i) {
      const fieldinfo_t *geofield = &swathzaaux->swath_or_zonal.geofield.s[i];
      if (associate_hdf5_dimension(ctxid, swathzaaux->swath_or_zonal_name.s, 0, geofield, 0) == -1) err1_flagbreak(2, failinside, "assoc dimension")
    }
    if (failinside) break;

    /* Fill the values in the newly-created dimension scales */
    for (i = 0; i < swathzaaux->swath_or_zonal.dim.len; ++i) {
      const dimension_t *dim = &swathzaaux->swath_or_zonal.dim.s[i];
      if (fill_hdf5_swath_dimension(ctxid, swathzaaux->swath_or_zonal_name.s, dim,aug_option) == -1) err1_flagbreak(2,failinside, "fill HDF5 swath dimension");
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&coordinates);
  }
  return succeeded ? 0 : -1;
}

/**
  *  @brief Augment swaths or zonal average objects.
     @param ctxid context identifier
     @param swathzas includes the name and other info. that will be retrieved from the HDF-EOS5 library and then gathered by
  *               collect_swath_aux or collect_zonal_aux. 
  *  @param aug_option: either index(1) or file(2) or default(0)

     @return 0 if successful, -1 if failed.
  */
static int augment_swathzas(hid_t ctxid, const szaugalloc *swathzas, const int aug_option)
{
  int succeeded = 0;
  do {

    context_t *ctx = get_context(ctxid);

    int i =0;
    int failinside = 0;

    end_ptr_field = ctx->dimgeo_len;
    end_ptr_file  = ctx->dim_data_len;
    end_ptr_index = ctx->dim_index_len;
    
    int len = ctx->grid_swath_len;
  
    for (i = 0; i < swathzas->len; ++i) {
      const swath_zonal_aug_t *swathzaaux = &swathzas->s[i];

      if(i == 0) /* test the first swath only */
      {
	      test_augmentation(ctxid, swathzaaux->swath_or_zonal_name.s, 0, 1);
      	test_augmentation(ctxid, swathzaaux->swath_or_zonal_name.s, 0, 0);
      	if(augmented == 1) 
          break;
      }

      if(len != 0) 
      	scale_storage_pos_in_mfile(ctxid, swathzaaux->swath_or_zonal_name.s);

      /* if the length of swathzas is > ctx->za_index, this is a zonal average object. */
	    ctx->za_object = (i>ctx->za_index-1)? 1: 0;

      if(!g_no_log_info)
		    msg_push(swathzaaux->swath_or_zonal_name.s);
      if (augment_oneswathza(ctxid, swathzaaux,aug_option) == -1) err1_flagbreak(2, failinside, "aug oneswath")
      if(!g_no_log_info)
	      msg_pop();
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/* This is a convenient function for debugging */
static int print_ptr_out(context_t *ctx)
{
  int i = 0;
  for(i=0; i<ctx->grid_swath_len; i++)
  {
    printf("[%s] [%i] [%i] [%i]\n", ctx->grid_swath_name[i], ctx->ptr_field[i], ctx->ptr_file[i], ctx->ptr_index[i]);
  }
  return 0;
}

/** 
  * @brief Add the grid or swath name specified in the mapping file to the context_t struct. 
  * @param ctx context identifier
  * @param line the line number in the mapping file
  * @param p1 the number of fields 
  * @param p2 the number of files
  * @param p3 the number of dimensions to fill with index numbers
  * @param eosobj_flag the flag to indicate if this is a swath,grid or zonal
  * @return 0 if successful, -1 if failed. 
*/
int add_grid_swath_name(context_t *ctx, const char *line, const int p1, const int p2, const int p3, const int obj_flag) 
{
  char *ptr = strstr(line, ":");
  
  if(!ptr) 
  {
      printf("Grid or swath name is not found. The syntax should be 'grid name: your grid name' etc.\n");
      printf("Check the usage of the augmentation tool for more details. \n");
      return -1;
  }

  ++ptr;
  ptr = trim(ptr);
  /* printf("[%s]\n", ptr);*/

  assert(ctx->grid_swath_len <= ctx->grid_swath_max);
  if(ctx->grid_swath_len == ctx->grid_swath_max) 
    {
      ctx->grid_swath_max += 1;
      ctx->grid_swath_name = realloc(ctx->grid_swath_name, sizeof(char*)*ctx->grid_swath_max);
      ctx->ptr_field = realloc(ctx->ptr_field, sizeof(int)*ctx->grid_swath_max);
      ctx->ptr_file = realloc(ctx->ptr_file, sizeof(int)*ctx->grid_swath_max);
      ctx->ptr_index = realloc(ctx->ptr_index, sizeof(int)*ctx->grid_swath_max);
      if(!ctx->grid_swath_name ||
	       !ctx->ptr_field || !ctx->ptr_file || !ctx->ptr_index )
        return -1; /* allocation error */
    }
 
  int len = ctx->grid_swath_len;
  ctx->grid_swath_name[len] = strdup(ptr);
  ctx->ptr_field[len] = p1;
  ctx->ptr_file[len]  = p2;
  ctx->ptr_index[len] = p3;
  ctx->grid_swath_len++;
  ctx->ptr_obj_flag = realloc(ctx->ptr_obj_flag,sizeof(int)*ctx->grid_swath_len);
  ctx->ptr_obj_flag[len]= obj_flag;

  return 0;
}

/**     
 * @brief parse the info. specified in dim to dim scale mapping file(dimgeonames) to the context struct
 *    
 * @param ctxid context identifier
 * @param dimgeonames dim to dim scale mapping file
 * @return 0 if successful,-1 if failed because of something wrong in the tool or the system.
                           -2 if users don't provide the correct information in the mapping file.
 */ 

static int parse_dim_geo_names(context_t *ctx, const char* dimgeonames)
{

  FILE *fp = NULL;
  FILE *fpointer = NULL;
  char *line = NULL;
  char *dimname = NULL;
  char *geoname = NULL; 
  char *saveptr = NULL;
  int  ret = 0;

  int  flag = -1; 
  char *dim_data_dname = NULL;
  char *dim_data_fname = NULL;
  char *dim_index_dname = NULL;

  /* position(or index) of a field name(corresponding to a dimension)  among all field names specified in the mapping file. 
     The field name is the name of an hdf-eos5 field that provides the dimension scale values.  */
  int pos_field=0;

  /* position(or index) of the dimension scale data file name(corresponding to a dimension)  among all dim. scale data file names 
     specified in the mapping file. 
     The data file is an ASCII file that contains the dimension scale values users specifies for a dimension */
  int pos_data_file=0;

  /* position(or index) of a dimension among all dimensions specified in the mapping file that users want to use nature numbers to fill in the dimension. 
  */
  int pos_index=0; 

  /* Initialize the info. that will be provided by the mapping file in the context struct */ 
  ctx->dimgeo_len = 0;
  ctx->dimgeo_max = 1;
  ctx->dimnames = NULL;
  ctx->geonames = NULL;
  ctx->dimnames = realloc(ctx->dimnames, sizeof(char*)*ctx->dimgeo_max);
  ctx->geonames = realloc(ctx->geonames, sizeof(char*)*ctx->dimgeo_max);

  ctx->dim_data_len = 0;
  ctx->dim_data_max = 2;
  ctx->dim_data_name = NULL;
  ctx->dim_data      = NULL;
  ctx->dim_data_name = realloc(ctx->dim_data_name, sizeof(char*)*ctx->dim_data_max);
  ctx->dim_data      = realloc(ctx->dim_data, sizeof(doublealloc)*ctx->dim_data_max);

  ctx->dim_index_len = 0;
  ctx->dim_index_max = 2;
  ctx->dim_index_name = NULL;
  ctx->dim_index_name = realloc(ctx->dim_index_name, sizeof(char*)*ctx->dim_index_max);

  ctx->grid_swath_len = 0;
  ctx->grid_swath_max = 2;
  ctx->grid_swath_name = realloc(ctx->grid_swath_name, sizeof(char*)*ctx->grid_swath_max);
  ctx->ptr_field = realloc(ctx->ptr_field, sizeof(int)*ctx->grid_swath_max);
  ctx->ptr_file = realloc(ctx->ptr_file, sizeof(int)*ctx->grid_swath_max);
  ctx->ptr_index = realloc(ctx->ptr_index, sizeof(int)*ctx->grid_swath_max);


  /* Open the mapping file */
  fp = fopen(dimgeonames, "r");
  if(fp==NULL) {
    printf("Choose the file option but the dimension mapping file doesn't exist. \n");
    return -2;
  }

  /* Assume the number of characters in  each line doesn't exceed 1024.*/

  char mystring[1024];

  while(fgets(mystring, 1024, fp)!=NULL) {/* while((s = getline(&line, &n, fp)) != -1)  */

    line = mystring;

    if(line[0]=='#') /* comment */
      continue;

    if(strlen(line) <= 1) /* empty line */
      continue;

    /* For multiple swath or grid cases, the user should provide the following
       string: 
       swath name: The real swath name 
       The real swath name needs to be stored inside ctx as well as the position of field, data file and index. 
       KY 2013-10-07 */

    int eosobj_flag = 0;
    char *ptr = strstr(line, "swath");

    /* The flag for swath is 2 */
    if(ptr) 
       eosobj_flag = 2;
    else {

       ptr = strstr(line, "grid");

       /* The flag for grid is 1.*/ 
       if(ptr) 
         eosobj_flag = 1;
       else {
           ptr = strstr(line,"zonal");

           /* The flag for zonal is 3. */
           if(ptr) 
              eosobj_flag = 3;
       }
    }

    if(ptr) 
    {
    	if(strstr(ptr, "name")) {
    	  if(add_grid_swath_name(ctx, line, pos_field, pos_data_file, pos_index,eosobj_flag) < 0) 
     	    {
    	      ret = -1; 
    	      goto parse_out;
    	    }
    	  continue;
    	}
    }

    /* trim out special characters such as space, tab or new line character*/
    char *token = trim(strtok_r(line, " \t\n\r", &saveptr));

    /* Not sure what this "2*" means? 2* means all the rest dimension fields will be filled with index numbers.*/
    /* Note: asterrisk_len is a global variable */
    if(strcmp(token, "2*") == 0) 
    {
        flag = 2;
        
        /* It may be possible that 2* appears multiple times in a mapping file. 
           This may happen when multiple swaths or multiple grids are specified in the mapping file.
           2* may appear for each grid or swath.
           KY 2013-10-07
        */ 
        asterisk_len++;
	      asterisk = realloc(asterisk, sizeof(int)*asterisk_len);
	      asterisk[asterisk_len-1] = ctx->grid_swath_len-1;
        ctx->iaster_flag = 1;
    } else
    {
	      flag = atoi(token);
    }
	
    if(flag == 0) { /*field */
	    if(ctx->grid_swath_len > 0)
	      ctx->ptr_field[ctx->grid_swath_len-1] = (++pos_field);

    	dimname = trim(strtok_r(/*line*/NULL, " \t\n\r", &saveptr));
    	geoname = trim(strtok_r(NULL, " \t\r\n", &saveptr));

    	if(dimname==NULL || geoname==NULL) {
          printf("Either dimension name or dimension scale field name doesn't exist. \nPlease check the mapping file %s and make the corrections.\n",dimgeonames);
      	  ret = -2;
      	  goto parse_out;
    	}	
        /* We will see if we want to print out this info. on the screen */
        /* Still keep it for this release. KY 2013-11-27 */
	    printf("[%i] [%s] [%s]\n", flag, dimname, geoname);
    
    	assert(ctx->dimgeo_len <= ctx->dimgeo_max);

    	if(ctx->dimgeo_len == ctx->dimgeo_max) {
            ctx->dimgeo_max += 1;
            ctx->dimnames = realloc(ctx->dimnames, sizeof(char*)*ctx->dimgeo_max);
            ctx->geonames = realloc(ctx->geonames, sizeof(char*)*ctx->dimgeo_max);
            if(ctx->dimnames == NULL || ctx->geonames == NULL) {
                ret = -1;
		            goto parse_out;
            }		
        }	

    	ctx->dimnames[ctx->dimgeo_len] = strdup(dimname);
    	ctx->geonames[ctx->dimgeo_len] = strdup(geoname);
    	ctx->dimgeo_len++;
    }  
    else if(flag == 1) {/* file */
	    if(ctx->grid_swath_len > 0)
	      ctx->ptr_file[ctx->grid_swath_len-1] = (++pos_data_file);

	    dim_data_dname = trim(strtok_r(NULL, " \t\n\r", &saveptr));
      dim_data_fname = trim(strtok_r(NULL, " \t\r\n", &saveptr));
	    printf("[%i] [%s] [%s]\n", flag, dim_data_dname, dim_data_fname);

      if(dim_data_dname==NULL || dim_data_fname==NULL) {
          printf("Either dimension name or ASCII file that describes the dimension scale doesn't exist. \n Please check the mapping file %s and make the corrections.\n",dimgeonames);
          ret = -2;
          goto parse_out;
      }

	    assert(ctx->dim_data_len <= ctx->dim_data_max);
      if(ctx->dim_data_len == ctx->dim_data_max) {
        ctx->dim_data_max += 1;
        ctx->dim_data_name = realloc(ctx->dim_data_name, sizeof(char*)*ctx->dim_data_max);
        ctx->dim_data = realloc(ctx->dim_data, sizeof(doublealloc)*ctx->dim_data_max);
        if(ctx->dim_data_name == NULL || ctx->dim_data == NULL) {
          ret = -1;
          goto parse_out;
        }
      }
	    ctx->dim_data_name[ctx->dim_data_len]=strdup(dim_data_dname);
	
	    doublealloc data;
      doublealloc_init(&data);

      if((fpointer = fopen(dim_data_fname, "r"))==NULL) {

        /* Should close the dimension-dimension scale file handler. 
               Also change the error message to be meaningful. KY 2013-10-04*/
        fprintf(stderr,"Cannot open the dim. scale data file: %s \n",dim_data_fname);   
        fprintf(stderr,"The data file may not exist or may be corrupted. Please check the file: %s\n",dim_data_fname);
        ret = -2; 
        goto parse_out;

	    /* printf("Can not open file!\n");
       		return -1; */
	    }

	    int capacity=INIT_MAX_NUM_ELMTS_OWN_SCALE;

      if(!doublealloc_ready(&data, capacity)) {

	    /* Should close both the dimension-dimension scale file handler and the dimension scale data file handler.
               Also change the error message in a more meaningful way. So change the code.  KY 2013-10-04*/
            fclose(fpointer);
            fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);   
            ret = -1;
	          goto parse_out;
	    }

      /* The original writer must assume that the maximum number (float or double) in decimal form will not exceed 126 bytes 
           This may be reasonable. KY 2013-10-04*/
      /* The following code will assure any size of data separated by comma,white space,tab or newlines will be read in 
           correctly. KY 2013-11-12 */

      char* saved_to_concat = NULL;
      int end_last_str_flag = 0;
     
      while(!feof(fpointer)) {

          char*str = NULL;
          char*cur = NULL;
          char*temp_cur = NULL;
          int concat_flag = 0;

          /*512 bytes not too big and not too small for dozens of numbers. */
          str = (char*)malloc(512);
          memset(str,0,512);

          if(fgets(str, 512, fpointer)) {

              if ((str[510]!='\0') && (str[510]!='\n')) {
                  concat_flag = 1;
              }

              int temp_flag_checker = 0;
              if ((str[510]==',') || (str[510]==' ') || (str[510]=='\t') || (str[510] == '\r'))
                  temp_flag_checker = 1;

              cur = strtok(str," ,\t\n\r");
              if(cur!=NULL) {
                  /* If by the end of string, it only reads '1234' out of '123456', we will
                     concatenate the last 1234 with 56 so that the number 123456 can be read through.  */

                  /* If needs to concatenate */
                  if (str[0]!=' ' && str[0]!=',' && str[0]!='\t'
                                  && str[0]!='\r' && end_last_str_flag == 0){
                      if(saved_to_concat!=NULL) {
                          char* to_concat = malloc(strlen(saved_to_concat)+strlen(cur)+1);
                          memset(to_concat,0,strlen(to_concat));
                          strncpy(to_concat,saved_to_concat,strlen(saved_to_concat));
                          strcat(to_concat,cur);
                          {
                              double n = atof(to_concat);

                              /* The maximum number of element of the own dimension scale can increase dynamically. */
                              if(data.len >= data.capacity)
                                  if(!doublealloc_readyplus(&data, capacity)) {
                                  /* Should close both the dimension-dimension scale file handler and the dimension scale data file handler.
                                     Also change the error message in a more meaningful way. So change the code.  KY 2013-10-04*/
                                      if(to_concat!=NULL)
                                          free(to_concat);
                                      if(saved_to_concat != NULL)
                                          free(saved_to_concat);
                                       if(str !=NULL)
                                          free(str);
                                      fclose(fpointer);
                                      fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                                      ret = -1;
                                      goto parse_out;
                                  }
                              data.s[data.len++] = n;

                          }
                          free(to_concat);
                          free(saved_to_concat);
                          saved_to_concat = NULL;
                      }
                      else {

                          /* This is the first element of the string, read it */
                              double n = atof(cur);

                              if(data.len >= data.capacity)
                                  if(!doublealloc_readyplus(&data, capacity)) {
                                      if(saved_to_concat !=NULL)
                                          free(saved_to_concat);
                                      if(str !=NULL)
                                          free(str);
                                      fclose(fpointer);
                                      fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                                      ret = -1;
                                      goto parse_out;
                                  }
                              data.s[data.len++] = n;

                      }
                  } 
                  /*Not necessary to concatenate,but remember the last element.*/
                  else {
                      if(saved_to_concat != NULL) {
                          {
                              double n = atof(saved_to_concat);

                              if(data.len >= data.capacity)
                                  if(!doublealloc_readyplus(&data, capacity)) {
                                      if(saved_to_concat !=NULL)
                                          free(saved_to_concat);
                                      if(str !=NULL)
                                          free(str);
                                      fclose(fpointer);
                                      fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                                      ret = -1;
                                      goto parse_out;
                                  }
                              data.s[data.len++] = n;


                          }
                          free(saved_to_concat);
                          saved_to_concat = NULL;
                      }

                      {
                          double n = atof(cur);

                          if(data.len >= data.capacity)
                              if(!doublealloc_readyplus(&data, capacity)) {
                                  if(saved_to_concat !=NULL)
                                      free(saved_to_concat);
                                  if(str !=NULL)
                                      free(str);
                                  fclose(fpointer);
                                  fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                                  ret = -1;
                                  goto parse_out;
                              }
                          data.s[data.len++] = n;

                      }
                  }
              }

              /* If this string's last character is a separator, remember it so that
                 we don't concatenate the strings.*/
              end_last_str_flag = temp_flag_checker;
            
              while(cur!=NULL) {

                  cur = strtok(NULL," ,\t\n\r");
                  if(cur !=NULL ){
                      if(concat_flag == 1) {/* May need to concatenate the last string */
                          if(temp_cur!=NULL) {
                              double n = atof(temp_cur);
                              if(data.len >= data.capacity)
                                  if(!doublealloc_readyplus(&data, capacity)) {
                                      if(saved_to_concat !=NULL)
                                      free(saved_to_concat);
                                      if(str !=NULL)
                                          free(str);
                                      fclose(fpointer);
                                      fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                                      ret = -1;
                                      goto parse_out;
                                  }
                              data.s[data.len++] = n;

                          }
                          temp_cur = cur;

                      }
                      else {/* No need to concantenate, just obtain the current number */
                               double n = atof(cur);
                              if(data.len >= data.capacity)
                                  if(!doublealloc_readyplus(&data, capacity)) {
                                      if(saved_to_concat !=NULL)
                                      free(saved_to_concat);
                                      if(str !=NULL)
                                          free(str);
                                      fclose(fpointer);
                                      fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                                      ret = -1;
                                      goto parse_out;
                                  }
                              data.s[data.len++] = n;

                      }
                  }
              }

              if(concat_flag == 1) {
                  if(temp_cur!=NULL) {
                      saved_to_concat = malloc(strlen(temp_cur)+1);
                      strcpy(saved_to_concat,temp_cur);
                  }
                  concat_flag = 0;
              }

          }

          free(str);
              
      } 
      ctx->dim_data[ctx->dim_data_len]=data;
      ++ctx->dim_data_len;

      if(saved_to_concat!=NULL)
          free(saved_to_concat);

      fclose(fpointer);

    } else if(2 == flag) { /* index */

        /* If this is actual 2* case, don't do anything */
        if(ctx->iaster_flag == 1)
           continue;


      	if(ctx->grid_swath_len > 0)
      	  ctx->ptr_index[ctx->grid_swath_len-1] = (++pos_index);
       
      	dim_index_dname = trim(strtok_r(NULL, " \t\n\r", &saveptr));
              if (dim_index_dname == NULL) {
                  printf("The dimension name that needs to fill in the natural number doesn't exist. \n Please check the mapping file %s and make the corrections.\n",dimgeonames);
                 ret = -2;
                 goto parse_out;
              }
      	
      	assert(ctx->dim_index_len <= ctx->dim_index_max);
        	if(ctx->dim_index_len == ctx->dim_index_max) {
             		ctx->dim_index_max += 1;
      		ctx->dim_index_name = realloc(ctx->dim_index_name, sizeof(char*)*ctx->dim_index_max);
      		if(ctx->dim_index_name == NULL) {
                          fprintf(stderr,"Cannot allocate enough memory. The error occurs at line %d of source file %s \n", __LINE__,__FILE__);
                          ret = -1;
                          goto parse_out;
      		}
      	}
      	ctx->dim_index_name[ctx->dim_index_len] = strdup(dim_index_dname);
      	++ctx->dim_index_len;
    }
    else {
       fprintf(stderr," Invalid number for the selection of the file option.\n Please check the number provided at the column 1 of the mapping file.\n The number must be 0,1,2 or 2*.\n ");
       ret = -2;
       goto parse_out;

    }
  }

  /* print_ptr_out(ctx); */
	
parse_out:
  fclose(fp);
  return ret;
}
/**     
 * @brief The internal routine to check if there are duplicate dimension names or field names
 *    
 * @param ctxid context identifier
 * @param si_field starting index for the data field 
 * @param ei_field ending index for the data field 
 * @param si_file starting index for the ASCII data file
 * @param ei_file ending index for the ASCII data file 
 * @param si_index starting index for the dimension that wants to use the nature number 
 * @param ei_field ending index for the dimension that wants to use the nature number 
 * @return 0 if successful,-1 if failed 
 */ 

static int check_duplicate_dim_field_names_internal(context_t *ctx,
                                                    int si_field,int ei_field, 
                                                    int si_file, int ei_file,
                                                    int si_index, int ei_index) {

    int succeeded = 1;
    int failinside = 0;
    int i = 0;
    int j = 0;

    /*Check dimnames(for field) within and then dim_data_name(ASCII dim scale files) and then dim_index_name */
    for (i = si_field; i <ei_field; i++) {
        for (j = i+1;j<ei_field;j++) {
            if((strcmp(ctx->dimnames[i],ctx->dimnames[j])) == 0) {
                print_dup_message(ctx->dimnames[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }

    /* field dim. with data file dim. */
    for (i = si_field; i <ei_field; i++) {
        for (j = si_file;j<ei_file;j++) {
            if((strcmp(ctx->dimnames[i],ctx->dim_data_name[j])) == 0) {
                print_dup_message(ctx->dimnames[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }       

    /* field dim. with index dim. */
    for (i = si_field; i <ei_field; i++) {
        for (j = si_index;j<ei_index;j++) {
            if((strcmp(ctx->dimnames[i],ctx->dim_index_name[j])) == 0) {
                print_dup_message(ctx->dimnames[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }

    /* data file dim. within */
    for (i = si_file; i <ei_file; i++) {
        for (j = i+1;j<ei_file;j++) {
            if((strcmp(ctx->dim_data_name[i],ctx->dim_data_name[j])) == 0) {
                print_dup_message(ctx->dim_data_name[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }

    /* data file dim. with index dim. */
    for (i = si_file; i <ei_file; i++) {
        for (j = si_index;j<ei_index;j++) {
            if((strcmp(ctx->dim_data_name[i],ctx->dim_index_name[j])) == 0) {
                print_dup_message(ctx->dim_data_name[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }

    /* data index dim. within */
    for (i = si_index; i <ei_index; i++) {
        for (j = i+1;j<ei_index;j++) {
            if((strcmp(ctx->dim_index_name[i],ctx->dim_index_name[j])) == 0) {
                print_dup_message(ctx->dim_index_name[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }

    /* Check out the field name duplication */
    /* We enforce that the dimension names link to a different fields, this only makes physical sense but not logical sense.
       This may help people not to accidently use the same field for different dimensions. But what if people want to do that?
       this may be relaxed later? 
       KY 2013-11-27 */

    for (i = si_field; i <ei_field; i++) {
        for (j = i+1;j<ei_field;j++) {
            if((strcmp(ctx->geonames[i],ctx->geonames[j])) == 0) {
                print_dup_message(ctx->geonames[i]);
                failinside = 1;
                break;
            }
        }
        if (failinside)
            break;
    }


    if(failinside) 
       succeeded = 0;
    
    return (succeeded?0:-1);

}
/**     
 * @brief The routine to check if there are duplicate dimension names or field names
 *    
 * @param ctx pointer to the context struct
 * @return 0 if successful,-1 if failed 
 */ 
static int check_duplicate_dim_field_names(context_t * ctx) {

    int succeeded = 1;
    int eosobj_index = 0;
    int failinside = 0;

    int si_field = 0;
    int ei_field = 0;
    int si_file  = 0;
    int ei_file  = 0;
    int si_index = 0;
    int ei_index = 0;
    

    /* Handle single HDFEOS object */
    if (0 == ctx->grid_swath_len) {
        
       si_field = 0;
       ei_field = ctx->dimgeo_len;
       si_file  = 0;
       ei_file  = ctx->dim_data_len;
       si_index = 0;
       ei_index = ctx->dim_index_len;

       if (check_duplicate_dim_field_names_internal(ctx,si_field,ei_field,si_file,ei_file,si_index,ei_index) ==-1)
          failinside = 1;
    }

    else { 

        for (eosobj_index = 0; eosobj_index < ctx->grid_swath_len; eosobj_index++) {
            if(eosobj_index == 0) {
                si_field = 0;
                si_file  = 0;
                si_index = 0;
            }

            else {
                si_field = ctx->ptr_field[eosobj_index-1];
                si_file  = ctx->ptr_file[eosobj_index-1];
                si_index = ctx->ptr_index[eosobj_index-1];
            }

            ei_field = ctx->ptr_field[eosobj_index];
            ei_file  = ctx->ptr_file[eosobj_index];
            ei_index = ctx->ptr_index[eosobj_index];

            if (check_duplicate_dim_field_names_internal(ctx,si_field,ei_field,si_file,ei_file,si_index,ei_index) ==-1) {
                failinside = 1;
                break;
            }

        }
    }

    if (failinside) 
       succeeded = 0;

    return (succeeded?0:-1);
}

/**     
 * @brief check if the dimension name in the HDF-EOS5 file(dim) exists in the mapping file
 * Note: Change the confused returned value, note: previously 0 means "exist", currently 0 means "don't exist"    
 * @param ctxid context identifier
 * @param dim  can retrieve all the information of the dimension
 * @return 1 if a there is a match of dimension name in the file(dim) and dimension name in the mapping file, otherwise 0
 */ 

int dim_name_exist(hid_t ctxid, const dimension_t *dim) 
{
  int l = 0;
  context_t *ctx = get_context(ctxid);
  
  for(l=0; l<ctx->dimgeo_len; l++)
    if(strcmp(dim->name.s, ctx->dimnames[l]) == 0)
      break;
  if(l < ctx->dimgeo_len)
    return 1;

  for(l=0; l<ctx->dim_data_len; l++)
    if(strcmp(dim->name.s, ctx->dim_data_name[l]) == 0)
      break;
  if(l < ctx->dim_data_len)
    return 1;

  for(l=0; l<ctx->dim_index_len; l++)
    if(strcmp(dim->name.s, ctx->dim_index_name[l]) == 0)
      break;
  if(l < ctx->dim_index_len)
    return 1;
 
  return 0;
}

/* Helper function to print out the index */
int print_index_out(hid_t ctxid) 
{
  int i = 0;
  context_t *ctx = get_context(ctxid);
  for(i=0; i<ctx->dim_index_len; i++)
    printf("dim_index_name[%i]=%s\n", i, ctx->dim_index_name[i]);  
  return 0;
}
/**
 * @brief Insert the swath dim. info. to the context struct for the index option
 * @param ctxid context identifier
 * @param grids can be used to retrieve the information about grids 
 * @param return 0 if successful, -1 if failed.
 */

/* The original developer who wrote this insert_swathza_index didn't provide 
   comments on the implementation logic. Here I try to add comments 
   based on my understanding. Comments may be subject to changes.
   If this tool becomes a higher priority, it may be necessary to
   revise the whole implementation approach.

   This routine as well as the insert_grid_index should only apply to the "index" case for the -f option.
   The "index" case refers to the situation when 
   the users want to fill in natural numbers(or index) to dimension scales for the dimensions not specified in
   the mapping file.
   In the dim. to dim. scale mapping file, the "index" case is represented as 2*.
   The expression 2* 
       2: fill in the natural number, 
       *: for the rest of dimension scales the users don't specify.

   KY 2013-10-07

*/

int insert_swathza_index(hid_t ctxid, const szaugalloc *swathzas) 
{

  /* If having 2* in the mapping file, 2* represents the case when  filling natural numbers for all non-specified dimension scales*/
  if(asterisk_len > 0) 
    {
      int i = 0;
      context_t *ctx = get_context(ctxid);
      int start =-1;
      int end = -1;      

      for(i=0; i<asterisk_len; i++)
      {
        int idx = asterisk[i];
        if(idx != -1) /* having grid or swath name */
        {
          char *str = ctx->grid_swath_name[idx];
          int k = 0;
      	  for (k=0; k<swathzas->len; ++k)
          {
           	const swath_zonal_aug_t *swathzaaux = &swathzas->s[k];
      	    if(strcmp(swathzaaux->swath_or_zonal_name.s, str) == 0) 
      	    {
      	      int l = 0;
              int m = 0;
      
      	      for (m=0; m<swathzaaux->swath_or_zonal.dim.len; ++m)
      	      { 
            		const dimension_t *dim = &swathzaaux->swath_or_zonal.dim.s[m];
            
                            /* Need to ignore the field case by looping through all the dimensions specified by fields inside the file KY 2013-10-07  */
            		start = (idx==0)?0: ctx->ptr_field[idx-1];
            		end = ctx->ptr_field[idx];
            		for(l=start; l<end; l++)
                  if(strcmp(dim->name.s, ctx->dimnames[l]) == 0)
                    break;
             	  if(l < end)
            		  continue;
            
                            /* Need to ignore the "data file" case by looping through all the dimensions of which dimension scales are specified in separate "data" files. KY 2013-10-07*/
            		start = (idx==0)?0: ctx->ptr_file[idx-1];
            		end = ctx->ptr_file[idx];
            		for(l=start; l<end; l++)
            		  if(strcmp(dim->name.s, ctx->dim_data_name[l]) == 0)
            		    break;
            		if(l < end)
            		  continue;
            
                /* If this dimension has already been indexed, also ignore this dimension  KY */
                start = (idx==0)?0: ctx->ptr_index[idx-1];
                end = ctx->ptr_index[idx];
                for(l=start; l<end; l++)
                  if(strcmp(dim->name.s, ctx->dim_index_name[l]) == 0)
                    break;
                if(l < end)
                  continue;	
            
            		/* Insert a new dimension for the index option*/
            		++ctx->dim_index_max;
                ctx->dim_index_name = realloc(ctx->dim_index_name, sizeof(char*)*ctx->dim_index_max); 
            		start = ctx->ptr_index[idx];
            		end = ctx->dim_index_len;
              	int n = 0;
            
                            /* Need to investigate why not appending the new dimension but prepending it in a costly way(for loop). KY 2013-10-07 */
            		for(n=end-1; n>start-1; n--) 
            		  ctx->dim_index_name[n+1] = ctx->dim_index_name[n];
            		ctx->dim_index_name[start] = strdup(dim->name.s);
            		++ctx->dim_index_len;
            		for(l=idx; l<ctx->grid_swath_len; l++)
            		  ctx->ptr_index[l]++;
      	      }
      	    }
      	  }
        }
      	else /* no grid or swath name */
      	{
      	  const swath_zonal_aug_t *swathzaaux = &swathzas->s[0];
      	  int m = 0 ;
      	  for (m=0; m<swathzaaux->swath_or_zonal.dim.len; ++m)
          {
            const dimension_t *dim = &swathzaaux->swath_or_zonal.dim.s[m];
      
                  /* If the dimension not found in the mapping file, add this dimension with the index option  KY */
        	  if(dim_name_exist(ctxid, dim) == 0)
            {
      	      ++ctx->dim_index_max;
      	      ctx->dim_index_name = realloc(ctx->dim_index_name, sizeof(char*)*ctx->dim_index_max);
      	      ctx->dim_index_name[ctx->dim_index_len] = strdup(dim->name.s);
      	      ++ctx->dim_index_len;
      	    }  	    
          }
      	} /* end of else */ 
      } /* end of for */	

      free(asterisk);
    }
  return 0;
}

/**
 * @brief Insert the grid dim. info. to the context struct for the index option
 * @param ctxid context identifier
 * @param grids can be used to retrieve the information about grids 
 * @param return 0 if successful, -1 if failed.
 */

/* The original developer who wrote this insert_swathza_index didn't provide 
   comments on the implementation logic. Here I try to add comments 
   based on my understanding. Comments may be subject to changes.
   If this tool becomes a higher priority, it may be necessary to
   revise the whole implementation approach.

   This routine as well as the insert_swathza_index should only apply to the "index"(aster case) case for the -f option.
   The "index" case refers to the situation when 
   the users want to fill in natural numbers(or index) to dimension scales for the dimensions not specified in
   the mapping file.
   In the dim. to dim. scale mapping file, the "index" case is represented as 2*.
   The expression 2* 
       2: fill in the natural number, 
       *: for the rest of dimension scales the users don't specify.

   KY 2013-10-07

*/
int insert_grid_index(hid_t ctxid, const gridaugalloc *grids) 
{

  /* If having 2* in the mapping file, 2* represents the case when filling natural numbers for all non-specified dimension scale.  KY 2013-10-07*/
  if(asterisk_len > 0) 
    {
      int i = 0;
      context_t *ctx = get_context(ctxid);
      int start = -1;
      int end = -1;      

      for(i=0; i<asterisk_len; i++)
      {
        int idx = asterisk[i];
        if(idx != -1) /* grid or swath name */
        {
          char *str = ctx->grid_swath_name[idx];
          int k = 0;
	        for (k=0; k<grids->len; ++k)
          {
      	    const grid_aug_t *gridaux = &grids->s[k];
      	    if(strcmp(gridaux->gridname.s, str) == 0) 
      	    {
      	      int l = 0;
              int m = 0;
      
      	      for (m=0; m<gridaux->grid.dim.len; ++m)
       	      {
            		const dimension_t *dim = &gridaux->grid.dim.s[m];
              		
                /* XDim and YDim correspond to the calculated lat/lon. So ignore them for the index case in the -f option. KY 2013-10-07 */
            		if(strcmp(dim->name.s, "XDim")==0 ||
                   strcmp(dim->name.s, "YDim")==0)
                  continue;
            		
                /* Need to ignore the field case by looping through all the dimensions specified by fields inside the file KY 2013-10-07  */
            		start = (idx==0)?0: ctx->ptr_field[idx-1];
            		end = ctx->ptr_field[idx];
            		for(l=start; l<end; l++)
                  if(strcmp(dim->name.s, ctx->dimnames[l]) == 0)
                    break;
             	  if(l < end)
            		  continue;
            
                /* Need to ignore the "data file" case by looping through all the dimensions of which dimension scales are specified in separate "data" files. KY 2013-10-07*/
            		start = (idx==0)?0: ctx->ptr_file[idx-1];
            		end = ctx->ptr_file[idx];
            		for(l=start; l<end; l++)
            		  if(strcmp(dim->name.s, ctx->dim_data_name[l]) == 0)
            		    break;
            		if(l < end)
            		  continue;
            
                /* If this dimension has already been indexed, also ignore this dimension  KY */
                start = (idx==0)?0: ctx->ptr_index[idx-1];
                end = ctx->ptr_index[idx];
                for(l=start; l<end; l++)
                  if(strcmp(dim->name.s, ctx->dim_index_name[l]) == 0)
                    break;
                if(l < end)
                  continue;	
            
                /* Insert a new dimension for the index option*/
            		++ctx->dim_index_max;
                ctx->dim_index_name = realloc(ctx->dim_index_name, sizeof(char*)*ctx->dim_index_max); 
            		start = ctx->ptr_index[idx];
            		end = ctx->dim_index_len;
              	int n = 0;
            
                /* Need to investigate why not appending the new dimension but prepending it in a costly way(for loop). KY 2013-10-07 */
            		for(n=end-1; n>start-1; n--) /*[start, end-1] => [start+1, end] */
            		  ctx->dim_index_name[n+1] = ctx->dim_index_name[n];
            		ctx->dim_index_name[start] = strdup(dim->name.s);
            		++ctx->dim_index_len;
            		for(l=idx; l<ctx->grid_swath_len; l++)
            		  ctx->ptr_index[l]++;
      	      }
      	    }
 	        }
        }
      	else /* no grid or swath name */
      	{
      	  const grid_aug_t *gridaux = &grids->s[0];
      	  int m = 0;
      	  for (m=0; m<gridaux->grid.dim.len; ++m)
          {
            const dimension_t *dim = &gridaux->grid.dim.s[m];
      	   
            /* XDim and YDim correspond to the calculated lat and lon, so ignore it */
      	    if(strcmp(dim->name.s, "XDim")==0 ||
               strcmp(dim->name.s, "YDim")==0)
               continue;
      
                  /* If the dimension not found in the mapping file, add this dimension with the index option  KY */
        	  if(dim_name_exist(ctxid, dim) == 0)
            {
      	      ++ctx->dim_index_max;
      	      ctx->dim_index_name = realloc(ctx->dim_index_name, sizeof(char*)*ctx->dim_index_max);
      	      ctx->dim_index_name[ctx->dim_index_len] = strdup(dim->name.s);
      	      ++ctx->dim_index_len;
      	    }  	    
          }
      	} /* end of else */ 
      } /* end of for */	

      free(asterisk);
    }
  return 0;
}

#if 0
This is no longer an issue. The code may be deleted in the next release. KY 2013-10-07
static int remove_prodhistory(hid_t ctxid)
{
  char *path = "/HDFEOS/ADDITIONAL/FILE_ATTRIBUTES";
  char *dsname = "ProductionHistory";
  hid_t gid;
  context_t *ctx;
 
  ctx = get_context(ctxid);
  if(!ctx)
      return -1;

  gid = H5GOPEN(ctx->hdf5fd, path);
  if(gid<=0)
    return 0;

  H5Ldelete(gid, dsname, H5P_DEFAULT);

  /* I do not care whether there was a ProductionHistory or Not.
     As long as it is not there now, this must return success
  */

  H5Gclose(gid);

  return 0;
}
#endif

/** 
 * @brief Augment the given file with the given method.
 * The augmentation procedure has two steps:
 *   1. Retrieving all HDF-EOS5 object information using the HDF-EOS5 library
 *   2. Augmenting the given file using the HDF5 library with information retrieved in the previous step
 * 
 * @param fname filename to be augmented
 * @param augmentmethod 1, 2, or 3 is available, check the following document at 
   http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
   We only test solution 1 and the current augmentation is always using 1.
 * @param dimgeonames: the name of the file that users can specify the mapping info. from dim. to dim. scale.  Such info. cannot be derived from the HDF-EOS5 file by the tool itself 
     This file is only valid for the -f option. Check the readme or usage at http://hdfeos.org/software/aug_eos5/
     for more information.
 * @param no_log_info: if not 0, no log info. will be printed
 * @param aug_option: either index (1) or file (2) or default (0)

 * @return 0 if successful
 */
int aug_eos5(const char *fname, int augmentmethod, const char* dimgeonames, int no_log_info, const int aug_option)
{

  g_no_log_info = no_log_info;
  hid_t ctxid = -1;
  gridaugalloc grids;
  szaugalloc swathzas;
  int succeeded = 0;

  /* Initialize grid and swath(including zonal) allocation to hold the whole structure
  */
  gridaugalloc_init(&grids);
  szaugalloc_init(&swathzas);

  do {

    /* Initialize the context info. that includes info. such as augmentaton method, file handles  etc. */
    context_t *ctx = NULL;
    if ((ctxid = create_context(augmentmethod)) == -1) earlyerr1_break(4, "create context")
    if ((ctx = get_context(ctxid)) == NULL) earlyerr1_break(4, "internal error?")
	
    /* If having the dim to dim scale mapping file, parse the info. from the mapping file. */
    if( 2 == aug_option) {

      int temp_parse_return = 0;
  
      temp_parse_return = parse_dim_geo_names(ctx,dimgeonames);

      /* Something is wrong with the code, record the file name and line number, break and out of the loop. */
      if(-1 == temp_parse_return) err1_break(4, "Error reading the mapping file")

      /* Something is wrong with the contents in the dimension map file, nothing is wrong with the code, print the 
         error message for the user to provide the correct info. in the mapping file */
      if (-2== temp_parse_return) 
         break; 
      

      /* Check if users accidently provide duplicate dimension name or field name within one EOS object */
      if (-1 == check_duplicate_dim_field_names(ctx)) 
          break;
    
    }

    /* Step 1: collect all information using the HDF-EOS5 library , open, read grid, swath, zonal and then close.*/
    /* g_no_log_info is used to turn on/off the logging/error messages. Now it is a global variable.*/
    if(!g_no_log_info)
        msg_push("read");

    if (open_hdfeos5(ctxid, fname, aug_option) == -1) err1_break(1, "open hdfeos5 file")
    if (read_grid(ctxid, &grids) == -1)  {
        /* Need to close HDF-EOS5 file handles first */
        if(close_hdfeos5(ctxid) == -1) warn1(1, "close hdfeos5")
        err1_break(1, "read hdfeos5 grid")
    }

    if (read_swath(ctxid, &swathzas) == -1) {
        if(close_hdfeos5(ctxid) == -1) warn1(1, "close hdfeos5")
        err1_break(1, "read hdfeos5 swath")
    }

    if (read_zonal(ctxid, &swathzas) == -1) {
        if(close_hdfeos5(ctxid) == -1) warn1(1, "close hdfeos5")
        err1_break(1, "read hdfeos5 zonal")
    }
 
    if(2 == aug_option) {
        if(check_obj_names(ctxid,&grids,&swathzas) == -1) {
           fprintf(stderr,"The swath, grid or zonal average names in the mapping file cannot be found in the HDF-EOS5 file\n");
           fprintf(stderr,"Please check the mapping file and make the corrections.\n");
           break;
        }
        if(check_dim_field_names(ctxid,&grids,&swathzas) == -1) {
           fprintf(stderr,"The dimension names or the data field names in the mapping file cannot be found in the HDF-EOS5 file\n");
           fprintf(stderr,"Please check the mapping file and make the corrections.\n");
           break;
        }
    }

    if (close_hdfeos5(ctxid) == -1) err1_break(1, "close hdfeos5")
    if(!g_no_log_info)
        msg_pop();

    /* Add the dimension names for swath(zonal) and grid for the index (2*) case in the -f option */
    if(swathzas.len != 0)    
      insert_swathza_index(ctxid, &swathzas);
    if(grids.len != 0)
      insert_grid_index(ctxid, &grids);

    /* Step 2: augment the given file using the HDF5 library */
    /* Open with the HDF5 interface, augment grids and swaths(zonal), then close */
    /* g_no_log_info is used to turn on/off the logging/error messages. Now it is a global variable.*/
    if(!g_no_log_info)
	    msg_push("augment");

    /* Open the HDF5 file */
    if (open_hdf5(ctxid, fname) == -1) err1_break(1, "open hdf5 file")

    /* Augment grids */
    if (augment_grids(ctxid, &grids,aug_option) == -1) {
        if(close_hdf5(ctxid)== -1)  warn1(1, "close hdf5");
        err1_break(1, "augment grid")
    }

    if(1 == augmented) {
        printf("%s has already been augmented. If you want to augment %s, you need to provide the original file to the augmentation tool.\n", ctx->infname, ctx->infname);
        if(close_hdf5(ctxid)== -1)  warn1(1, "close hdf5");
        break;
    }

    /* Augment swaths and zonal averages */
    if (augment_swathzas(ctxid, &swathzas,aug_option) == -1) {
        if(close_hdf5(ctxid)== -1)  warn1(1, "close hdf5");
        err1_break(1, "augment swath")
    }

    if(1 == augmented) {
        printf("%s has already been augmented. If you want to augment %s, you need to provide the original file to the augmentation tool.\n", ctx->infname, ctx->infname);
        if(close_hdf5(ctxid)== -1)  warn1(1, "close hdf5");
        break;
    }


    /* Close the HDF5 file */
    if (close_hdf5(ctxid) == -1) err1_break(1, "close hdf5")
    if(!g_no_log_info)
        msg_pop();

    succeeded = 1;
  } while (0);

  if (1) {
    gridaugalloc_free(&grids);
    szaugalloc_free(&swathzas);
    close_context(ctxid);
  }

  return succeeded ? 0 : -1;
}

/* @brief Usage of this augmentation tool
*/
void usage()
{
  printf("Usage: aug_eos5 [-i|-f <mapping-file>] <HDF-EOS5 file>\n");
}
