/* 
 Copyright by The HDF Group.
 All rights reserved.                                                      
                                                                           
 This file is part of the HDF-EOS5 Augmentation Tool. The full HDF-EOS5 
 Augmentation Tool copyright notice including terms governing use, modification,
 and redistribution, is contained in the file COPYING.     
 COPYING can be found at the root of the source code    
 distribution tree. If you do not have access to this file, you may request
 a copy from eoshelp@hdfgroup.org. 
*/


#ifndef DUMP_UTILITY_H
#define DUMP_UTILITY_H

#include "cvct.h"
#include "H5public.h"
#include "H5Ipublic.h"
#include "aug_eos5apicompatible.h"


/** 
 * @brief vector-like structures
 * The following preprocessor statements create C++ vectors like structures and functions.
 * e.g. DECLARE_VECTOR_ALL(challoc, char) will make bunch of functions that can be found in cvct.h.
 * A structure challoc contains 's' field for actual data and
 * 'len' field for the length.
 */
DECLARE_VECTOR_ALL(challoc, char)
DECLARE_VECTOR_ALLOCATOR(longalloc, long)
DECLARE_VECTOR_ALLOCATOR(doublealloc, double)
DECLARE_VECTOR_ALLOCATOR(hobjreftalloc, hobj_ref_t)

/* The structs defined on the following are 
   namelist_t,dimension_t,grid_t,dimmap_t,projinfo_t,gridinfo_t,fieldinfo_t,attribute_t,indexmap_t,swath_zonal_t, 
   followed by initialization(init_namelist_t etc.) and free (free_namelist_t etc.) functions */

typedef struct {
  challoc name;
} namelist_t;

void init_namelist_t(namelist_t *e);
void free_namelist_t(namelist_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(namelistalloc, namelist_t)

/** 
 * @brief A dimension in HDF-EOS5 is a pair of dimension name and size
 */
typedef struct {
  challoc name;
  hsize_t dimsize;
  challoc max;
  hsize_t maxdimsize;
} dimension_t;
void init_dimension_t(dimension_t *e);
void free_dimension_t(dimension_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(dimensionalloc, dimension_t)

/** 
 * @brief HDF-EOS5 attributes are represented as name,datatype and value.
 */
typedef struct {
  challoc name;
  hid_t type;
  challoc value;
} attribute_t;
void init_attribute_t(attribute_t *e);
void free_attribute_t(attribute_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(attributealloc, attribute_t)

/** 
 * @brief HDF-EOS5 Swath Datafield and Geofield are mapped to this structure
 * It consists of field name, array rank, data type, dimensions, data ,
 * the fillvalue attribute and other local attributes.
 */
typedef struct {
  challoc name;
  int rank;
  hid_t type;
  dimensionalloc dims;
  challoc data;
  challoc filler;
  attributealloc localattr; // local attributes from HE5_GDinqlocattrs
} fieldinfo_t;
void init_fieldinfo_t(fieldinfo_t *e);
void free_fieldinfo_t(fieldinfo_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(fieldinfoalloc, fieldinfo_t)

/** 
 * @brief Grid information
 * For more information, please refer to the API HE5_GDgridinfo at the HDF-EOS5 
 * reference manual(http://newsroom.gsfc.nasa.gov/sdptoolkit/docs/HDF-EOS5_REF.pdf)
 */
typedef struct {
  long xdim;
  long ydim;
  double upleft[2];
  double lowright[2];
} gridinfo_t;
void init_gridinfo_t(gridinfo_t *e);
void free_gridinfo_t(gridinfo_t *e);

/** 
 * @brief Grid projection information
 * For more information, please refer to the API HE5_GDprojinfo at the HDF-EOS5 
 * reference manual(http://newsroom.gsfc.nasa.gov/sdptoolkit/docs/HDF-EOS5_REF.pdf)
 */
typedef struct {
  int proj;
  int zone;
  int sphere;
  double param[16];
  int pix;
  int origin;
} projinfo_t;
void init_projinfo_t(projinfo_t *e);
void free_projinfo_t(projinfo_t *e);

/** 
 * @brief one big data structure that holds everything in one EOS5 Grid
 */
typedef struct {
  gridinfo_t info;
  projinfo_t proj;
  namelistalloc extendible_dimnames;
  dimensionalloc dim;
  fieldinfoalloc field;
  attributealloc attr; /* Object attributes from HE5_GDinqattrs */
  attributealloc grpattr; /* Group attributes from HE5_GDinqgrpattrs */
} grid_t;
void init_grid_t(grid_t *e);
void free_grid_t(grid_t *e);

/** 
 * @brief EOS5 swath dimension mapping
 */
typedef struct {
  challoc data;
  challoc geo;
  long offset;
  long increment;
} dimmap_t;
void init_dimmap_t(dimmap_t *e);
void free_dimmap_t(dimmap_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(dimmapalloc, dimmap_t)

/** 
 * @brief EOS5 swath index mapping
 */
typedef struct {
  challoc geo;
  challoc data;
  longalloc indices;
} indexmap_t;
void init_indexmap_t(indexmap_t *e);
void free_indexmap_t(indexmap_t *e);
DECLARE_VECTOR_ALLOCATOR_FREE(indexmapalloc, indexmap_t)

/** 
 * @brief one big data structure that holds everything in one EOS5 Swath or zonal average
 */
typedef struct {
  namelistalloc  extendible_dimnames;
  dimensionalloc dim;
  dimmapalloc dimmap;
  indexmapalloc index;
  fieldinfoalloc geofield;
  fieldinfoalloc datafield;
  attributealloc attr; /* attributes from HE5_SWinqattrs */
  attributealloc grpattr; /* attributes from HE5_SWinqgrpattrs */
} swath_zonal_t; /* Zonal shares the same data structure with swath, so swath_t is renamed as swath_zonal_t.  */
void init_swath_or_zonal_t(swath_zonal_t *e);
void free_swath_or_zonal_t(swath_zonal_t *e);

/** 
 * @brief The holder of all the information in an HDF-EOS5 file
 */
typedef void (*dtorfnc)(hid_t ctxid);
DECLARE_VECTOR_ALLOCATOR(dtorfncalloc, dtorfnc)
typedef struct {

  /* File name */
  const char *infname;

  /* The grid file ID handle when opening or creating an HDF5 file to create,read or write a grid. */
  hid_t gridfd;
  namelistalloc gridlist;

  /* The swath or zonal file ID handle when opening or creating an HDF5 file to create,read or write 
  a swath or a za.
  */
  hid_t swathfd;
  hid_t zonalfd;
  namelistalloc szlist;

  /* Global attributes retrieved from HE5_EHinqglbattrs.
     Check section 3.6 of HDF-EOS5 user's guide(http://newsroom.gsfc.nasa.gov/sdptoolkit/docs/HDF-EOS5_UG.pdf)
  */
  attributealloc globalattr; 

  /* HDF5 file ID */
  hid_t hdf5fd;

  /* augmentmethod reflects 3 solutions documented by an investigation report.
   The document can be found under 
   http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
   Eventually only solution 1 is accepted although the tool also implements other two solutions.
   Only solution 1 is used by NASA and we only test solution 1.
  */
  int augmentmethod;

  /* This is related to the C++ vector-like implementation. */
  dtorfncalloc dtors;

  /* Dimension name to GeoField name mapping */
  char **dimnames;
  char **geonames;
  int dimgeo_len;
  int dimgeo_max;

  /* Dimension name to dimension data mapping */
  char **dim_data_name;
  doublealloc *dim_data;
  int dim_data_len;
  int dim_data_max;

  /* Dimension name to index mapping */
  char **dim_index_name;
  int dim_index_len;
  int dim_index_max;
  int iaster_flag;

  /* Swath or grid or zonal average names specified by users at the dim to dim. scale mapping file */
  char **grid_swath_name;

  /* Swath or grid or zonal flag, necessary to check if the dimension name or
     the field name exist for a corresponding EOS5 object in the mapping file. 
     KY 2013-10-30 */
  int *ptr_obj_flag;

  /* The dimension field position at the dim to dim. scale mapping file */
  int *ptr_field;

  /* The position that specifies 'which' dim. scale data file at the dim to dim. scale mapping file */
  int *ptr_file;

  /* The position of an index used at the dim to dim. scale mapping file */
  int *ptr_index;

  /* The number of swath or grid specified in the dim to dim. scale mapping file */
  int grid_swath_len;

  /* The maximum number of swath or grid specified in the dim. to dim. scale mapping file.
     It seems this number is dynamically increased. */
  int grid_swath_max;

  /* For grid only */
  /* This is used to obtain the latitude and longitude values
     directly from latitude and longitude fields for TES
     since the projection info. is not accurate.
     We need both the name and the name length to assure
     that we got the correct value.
  */
  char *instrument_name;
  hsize_t len_of_instrument_name;

  /* The swath and zonal average share almost the same data structure. The tool internally
     treats them the same. za_index marks the end of swaths and the start of the zonal
     average objects in the file. */
  int za_index;

  /* This is the flag that shows this object is an zonal average object. */
  int za_object;
} context_t;

/* Create the context_t that holds the structure of an HDF-EOS5 file */
hid_t create_context(int augmentmethod);

/* Obtain the context_t */
context_t * get_context(hid_t ctxid);

/* C++ vector like memory operation routine 
   This routine seems not be used by the tool.
   KY 2013-10-18
*/
#if 0
int add_dtor(hid_t ctxid, dtorfnc d);
#endif

/* Close the context_t */
void close_context(hid_t ctxid);

/* parse name lists and store into individual names into C++ vector-like structures. */
int parse_list(const char *s, size_t len, char sep, namelistalloc *names);

/* The helper function will remove leading and trailing spaces.  
*/
char * trim(char *str);

/* error handling */
/* The following error handling macros are not totally used. Furthermore,
   there is no information on how the error code is used. For example, what doesn't it mean by
   calling err_break(3)? According to the implementation of these macros it seems to imply that H4toH5error_set
   has been used as a reference for this tool. So using the get_codestr function at h4toh5util.c as
   a guideline(http://www.hdfgroup.org/ftp/HDF5/tools/h4toh5/src/), I make the error code in the 
   following categories:
   code 1: Read HDF-EOS5 error
   code 2: Augment HDF-EOS5 error
   code 3: Memory error
   code 4: Other error
   The error message may already provide hints. Hopefully these code numbers can provide more hints.
*/ 
   
#define earlynomem_break() { _earlynomem1(__FILE__, __LINE__, ""); break; }
#define earlyerr1_break(l, msg) { _earlyerr1(l, __FILE__, __LINE__, msg); break; }
#define earlyerr1_return(l,msg) {_earlyerr1(l, __FILE__, __LINE__, msg); return FAIL; }
#define earlyerr1_flagbreak(l, var, msg) { _earlyerr1(l, __FILE__, __LINE__, msg); var = 1; break; }
#define nomem_break() { _nomem1(ctxid, __FILE__, __LINE__, ""); break; }
#define nomem_flagbreak(var) { _nomem1(ctxid, __FILE__, __LINE__, ""); var = 1; break; }
#define nomem1_break(msg) { _nomem1(ctxid, __FILE__, __LINE__, msg); break; }
#define nomem1_flagbreak(var, msg) { _nomem1(ctxid, __FILE__, __LINE__, msg); var = 1; break; }
#define nomem2_break(msg, a1) { _nomem2(ctxid, __FILE__, __LINE__, msg, a1); break; }
#define nomem2_flagbreak(var, msg, a1) { _nomem2(ctxid, __FILE__, __LINE__, msg, a1); var = 1; break; }
#define err_break(l) { _err1(ctxid, l, __FILE__, __LINE__, ""); break; }
#define err_flagbreak(l, var) { _err1(ctxid, l, __FILE__, __LINE__, ""); var = 1; break; }
#define err_return(l) { _err1(ctxid, l, __FILE__, __LINE__, ""); return FAIL; }
#define warn1(l, msg) { _warn1(ctxid, l, __FILE__, __LINE__, msg); }
#define err1_break(l, msg) { _err1(ctxid, l, __FILE__, __LINE__, msg); break; }
#define err1_flagbreak(l, var, msg) { _err1(ctxid, l, __FILE__, __LINE__, msg); var = 1; break; }
#define err1_return(l, msg) { _err1(ctxid, l, __FILE__, __LINE__, msg); return FAIL; }
#define err2_break(l, msg, a1) { _err2(ctxid, l, __FILE__, __LINE__, msg, a1); break; }
#define err2_flagbreak(l, var, msg, a1) { _err2(ctxid, l, __FILE__, __LINE__, msg, a1); var = 1; break; }
#define err2_return(l,msg,a1) {_err2(ctxid, l, __FILE__, __LINE__, msg, a1); return FAIL; }
#define err3_break(l, msg, a1, a2) { _err3(ctxid, l, __FILE__, __LINE__, msg, a1, a2); break; }
#define err3_flagbreak(l, var, msg, a1, a2) { _err3(ctxid, l, __FILE__, __LINE__, msg, a1, a2); var = 1; break; }
#define ASSERT(e) { if (!(e)) _assertfailure(__FILE__, __LINE__); }
void _nomem1(hid_t ctxid, const char *fname, int line, const char *msg);
void _nomem2(hid_t ctxid, const char *fname, int line, const char *msg, const char *a1);
void _earlynomem1(const char *fname, int line, const char *msg);
void _earlyerr1(int code, const char *fname, int line, const char *msg);
void _warn1(hid_t ctxid, int code, const char *fname, int line, const char *msg);
void _err1(hid_t ctxid, int code, const char *fname, int line, const char *msg);
void _err2(hid_t ctxid, int code, const char *fname, int line, const char *msg, const char *a1);
void _err3(hid_t ctxid, int code, const char *fname, int line, const char *msg, const char *a1, const char *a2);
void _assertfailure(const char *fname, int line);

/* message */
#define msg_push(s) _msg_push(ctxid, s)
#define msg_pop() _msg_pop(ctxid)
#define msg1(s) _msg1(ctxid, s)
#define msg2(s1, s2) _msg2(ctxid, s1, s2)
#define msg3(s1, s2, s3) _msg3(ctxid, s1, s2, s3)
#define msg4(s1, s2, s3, s4) _msg4(ctxid, s1, s2, s3, s4)
void _msg_push(hid_t ctxid, const char *s);
void _msg_pop(hid_t ctxid);
void _msg1(hid_t ctxid, const char *s);
void _msg2(hid_t ctxid, const char *s1, const char *s2);
void _msg3(hid_t ctxid, const char *s1, const char *s2, const char *s3);
void _msg4(hid_t ctxid, const char *s1, const char *s2, const char *s3, const char *s4);

/* Retrieve latitude and longitude values of HDF-EOS5 grids based on HDF-EOS5 APIs.*/
int generate_lonlat(hid_t ctxid, const grid_t *grid, int *ydimmajor, doublealloc *lons, doublealloc *lats);

/* Check if latitude and longitude are orthogonal, that is: if the horizonal domain is xdim by ydim.
   Can latitude and longitude be represented as lat[ydim],lon[xdim] or lat[xdim],lon[ydim]? If yes,
   lat and lon are treated as orthogonal. */
int check_ortho_dimscale(const grid_t *grid, int ydimmajor, const doublealloc *lon, const doublealloc *lat);

#endif

// vim:ts=8:sw=2:sts=2:cindent
