/* 
 Copyright by The HDF Group.
 All rights reserved.                                                      
                                                                           
 This file is part of the HDF-EOS5 Augmentation Tool. The full HDF-EOS5 
 Augmentation Tool copyright notice including terms governing use, modification,
 and redistribution, is contained in the file COPYING.     
 COPYING can be found at the root of the source code    
 distribution tree. If you do not have access to this file, you may request
 a copy from eoshelp@hdfgroup.org. 
*/
#include "hdf5.h"
#include "util.h"
#include "write.h"

/* These global variables are defined in run.c. Hope to remove them in the next release. KY 2013-10-16 */ 
extern int g_no_log_info;
extern int start_ptr_field, end_ptr_field, start_ptr_file, end_ptr_file, start_ptr_index, end_ptr_index;

#define EOSGEOFIELD  "/HDFEOS/SWATHS/%s/Geolocation Fields/%s"
#define EOSGEOFIELD2 "/HDFEOS/ZAS/%s/Data Fields/%s"
#define EOSDATAFIELD "/HDFEOS/GRIDS/%s/Data Fields/%s"

/** 
 * @brief Open the HDF-EOS5 file with the HDF5 library
 * 
 * @param ctxid context identifier
 * @param fname file name
 * 
 * @return 0 if successful
 */
int open_hdf5(hid_t ctxid, const char *fname)
{
  hid_t hdf5fd = -1;
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;

  do {
    if ((hdf5fd = H5Fopen(fname, H5F_ACC_RDWR, H5P_DEFAULT)) < 0) err1_break(3, "H5Fopen fail")
    succeeded = 1;
  } while (0);

  if (succeeded) {
    ctx->hdf5fd = hdf5fd;
  }
  else {
    if (hdf5fd != -1) 
        H5Fclose(hdf5fd);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Close the HDF5 file ID that is used to open the HDF-EOS5 file
 * 
 * @param ctxid context identifier
 * 
 * @return 0 if successful
 */
int close_hdf5(hid_t ctxid)
{
  context_t *ctx = get_context(ctxid);
  if (ctx != NULL) {
    if (ctx->hdf5fd != -1) 
        H5Fclose(ctx->hdf5fd);
    ctx->hdf5fd = -1;
    return 0;
  }
  return -1;
}

/**** *****************************WARNING**************************
 * NOTE: This function is NOT used by the current augmentation tool since it applies to solution 2 or 3 listed
 * at the document under 
    http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
 * The function is not reviewed.

 *******************************END OF WARNING ******************
*/
static int make_grouppath(hid_t ctxid, const char *eosobjname, int isgrid, challoc *gridpath)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);

    if (isgrid) {
      if (!challoc_copys(gridpath, "/HDFEOS/GRIDS/")) nomem1_break("/")
    }
    else {
      if (!challoc_copys(gridpath, (ctx->za_object==0)? "/HDFEOS/SWATHS/": "/HDFEOS/ZAS/")) nomem1_break("/") 
    }
    if (!challoc_cats(gridpath, eosobjname)) nomem1_break("gridswath groupname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}


/** 
 * @brief augment one grid using the HDF5 library
 * 
 * @param ctxid context identifier
 * @param eosobjname grid or swath name
 * @param isgrid Is this an EOS5 grid 
 * @param gridpath grid path
 * 
 * @return 0 if successful
 */

static int make_aug_grouppath(hid_t ctxid, const char *eosobjname, int isgrid, challoc *gridpath)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);
    int failinside = 0;

    switch (ctx->augmentmethod) {
      case 1:
	if (!challoc_copys(gridpath, "/HDFEOS/")) nomem1_flagbreak(failinside, "/")
	break;
      case 2:
      case 3:
	if (!challoc_copys(gridpath, "/NETCDF4/")) nomem1_flagbreak(failinside, "/")
	break;
      default:
	ASSERT(0);
	break;
    }
    if (failinside) break;

    if (isgrid) {
      if (!challoc_cats(gridpath, "GRIDS/")) nomem1_flagbreak(failinside, "/")
    }
    else {
      if (!challoc_cats(gridpath, (ctx->za_object==0)? "SWATHS/": "ZAS/")) nomem1_flagbreak(failinside, "/")
    }

    if (!challoc_cats(gridpath, eosobjname)) nomem1_break("eosobjname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief make the absolute path for the given dimension name
 * For example, if eosobjname is "grid1", isgrid is non-zero and dimname is "dim1",
 * This function will generate the absolute path of a new dim.. This new dim. may be
 * used to create an HDF5 dimensional scale HDF5 dataset 
 * The absolute path of the dim1 is "/HDFEOS/GRIDS/grid1/dim1".
 * 
 * @param ctxid context identifier
 * @param eosobjname grid name or swath name or zonal average name
 * @param isgrid is this dimension part of grid?
 * @param dimname dimension name
 * @param[out] dimpath absolute dimension path
 * 
 * @return 0 if successful
 */
static int make_dimpath(hid_t ctxid, const char *eosobjname, int isgrid, const char *dimname, challoc *dimpath)
{
  int succeeded = 0;

  do {
    if (make_aug_grouppath(ctxid, eosobjname, isgrid, dimpath) == -1) err_break(3)
    if (!challoc_cats(dimpath, "/")) nomem1_break("/")
    if (!challoc_cats(dimpath, dimname)) nomem1_break("dimname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief make the absolute path for the given data field name
 * For example, if eosobjname is "swath1", isgrid is zero and dsetname is "geofield1",
 * this function will put "/HDFEOS/SWATHS/swath1/Geolocation Fields/geofield1" as the dset path.
 * 
 * @param ctxid context identifier
 * @param eosobjname grid name or swath name or zonal average name
 * @param isgrid is this field part of grid?
 * @param dsetname dataset name
 * @param isdatafield is this data field or geo-location field?
 * @param[out] dsetpath absolute dataset path
 * 
 * @return 0 if successful
 */
static int make_fieldpath(hid_t ctxid, const char *eosobjname, int isgrid, const char *dsetname, int isdatafield, challoc *dsetpath)
{
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);

    if (make_aug_grouppath(ctxid, eosobjname, isgrid, dsetpath) == -1) err_break(3)
    if (isdatafield ==1) {/* general data fields */
      if (!challoc_cats(dsetpath, "/Data Fields/")) nomem1_break("/")
    }
    else if(isdatafield == 2) {/*  lat/lon cv fields */
      if (!challoc_cats(dsetpath, "/")) nomem1_break("/")
    }
    else {
      /* Swath in "Geolocaton Fields", grid in "Data Fields". */
      if (!challoc_cats(dsetpath, (ctx->za_object==0)? "/Geolocation Fields/" : "/Data Fields/")) nomem1_break("/")
    }
    if (!challoc_cats(dsetpath, dsetname)) nomem1_break("datasetname")

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief Helper function to detect if the datatype is an HDF5 string datatype
 * 
 * @param ctxid context identifier
 * @param type  HDF5 datatype ID
 * 
 * @return 0 if successful
 */

static int detect_stringtype(hid_t ctxid, hid_t type)
{
  H5T_class_t c = H5Tget_class(type);
  if (c == H5T_STRING)
    return 1;
  return 0;
}

/** 
 * @brief Check if an attribute exists
 *        Note: This routine currently is ONLY used for checking XDim and YDim dimension scale attributes when handling HDF-EOS5 grids
 *              Without modification, for future use, it should ONLY apply to the case when 
                the dimension scale names are added by the tool with the dimension scale path like "/HDFEOS/EOSOBJTYPE/EOSOBJNAME/SCALENAME"
                for example, "/HDFEOS/GRIDS/grid1/dim1" etc.
 * @param ctxid context identifier
 * @param eosobjname HDF-EOS5 object name
 * @param dname  dimension name 
 * @param isdim is this dimension part of grid?
 * @param isdatafield is this the data field 
          Swath has both geolocation and data fields. This parameter is useful when isgrid is not true and the user wants to check a field attribute
 * @param isgrid is this an HDF-EOS5 grid?
 * @param aname  attribute name
 * @return 1 if the attribute exists, 0 if the attribute doesn't exist, -1 if a failure occurs
 */
static int check_attribute_existence(hid_t ctxid, const char *eosobjname,const char *dname,int isdim, int isdatafield, int isgrid,const char *aname) {

  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;
  int attr_exist = 0;

  challoc_init(&dimpath);

  do {

    context_t *ctx = get_context(ctxid);

    /* If this an HDF-EOS5 dimensional scale, make the dimension name.*/
    if (isdim) {
       if (make_dimpath(ctxid, eosobjname, isgrid, dname, &dimpath) == -1) err_break(2)
    }
    else {
       if (make_fieldpath(ctxid, eosobjname, isgrid, dname, isdatafield, &dimpath) == -1) err_break(2)
    }

    if ((dsetid = H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err1_break(2, "H5DOPEN")
    attr_exist = H5Aexists(dsetid,aname);
    if (attr_exist < 0) err1_break(2,"H5Aexist");
      
    succeeded = 1;

  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
    challoc_free(&dimpath);
  }
  if (1 == succeeded)
     return ((attr_exist>0)?1:0) ;
  else 
     return -1;

}

/** 
 * @brief delete a dimension scale attribute
 * Note: This routine currently is ONLY used for checking XDim and YDim dimension scale attributes when handling HDF-EOS5 grids
 *       Without modification, for future use, it should ONLY apply to the case when 
         the dimension scale names are added by the tool with the dimension scale path like "/HDFEOS/EOSOBJTYPE/EOSOBJNAME/SCALENAME"
         for example, "/HDFEOS/GRIDS/grid1/dim1" etc.
 * @param ctxid context identifier
 * @param eosobjname HDF-EOS5 object name
 * @param dimname  dimension scale name 
 * @param isgrid is this an HDF-EOS5 grid?
 * @param aname  attribute name
 * @return 0 if the attribute is successfully deleted, -1 if a failure occurs
 */
static int delete_attribute(hid_t ctxid, const char *eosobjname,const char *dimname,int isgrid,const char *aname) {

  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {

    context_t *ctx = get_context(ctxid);

    /* If this an HDF-EOS5 dimensional scale, make the dimension name.*/
    if (make_dimpath(ctxid, eosobjname, isgrid, dimname, &dimpath) == -1) err_break(2)

    if ((dsetid = H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err1_break(2, "H5DOPEN")
    if (H5Adelete(dsetid, aname)<0) err2_break(2, "del att", dimpath.s);
      
    succeeded = 1;

  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
    challoc_free(&dimpath);
  }
  return succeeded? 0: -1;

}
/** 
 * @brief Write an HDF5 attribute
 * 
 * @param ctxid context identifier
 * @param obj_id HDF5 group ID or dataset ID that the attribute is attached to.
 * @param name Attribute name
 * @param type Attribute data type
 * @param data Attribute data values
 * @param datalen Attribute data length in byte
 * 
 * @return 0 if successful
 */

static int write_attribute(hid_t ctxid, hid_t obj_id, const char *name, hid_t type, const void *data, int datalen)
{
  hid_t space = -1;
  hid_t at = -1;
  int succeeded = 0;

  do {
    size_t typesize = 0;

    /* Check if this is a string datatype */
    if (detect_stringtype(ctxid, type)) {
      if ((space = H5Screate(H5S_SCALAR)) == -1) err1_break(2, "space create")
    }
    else {
      hsize_t count = 0;
      if ((typesize = H5Tget_size(type)) == 0) err1_break(2, "unknown type size")
      count = datalen / typesize;
      if ((space = H5Screate_simple(1, &count, &count)) == -1) err1_break(2, "space create")
    }

    if ((at = H5ACREATE(obj_id, name, type, space, H5P_DEFAULT)) < 0) err1_break(2, "attribute create")
    if (H5Awrite(at, type, data) == -1) err1_break(2, "attribute write")

    succeeded = 1;
  } while (0);

  if (1) {
    if (at != -1) H5Aclose(at);
    if (space != -1) H5Sclose(space);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Internal routine to write an HDF5 string attribute
 * 
 * @param ctxid context identifier
 * @param obj_id HDF5 group ID or dataset ID that the attribute is attached to.
 * @param name Attribute name
 * @param value Attribute value
 * @param addnull  Add NULL or not 
 * 
 * @return 0 if successful
 */

static int write_attribute_string_internal(hid_t ctxid, hid_t obj_id, const char *name, const char *value, int addnull)
{
  hid_t tid = -1;
  int succeeded = 0;

  do {
    int valuelen = strlen(value);

    if ((tid = H5Tcopy(H5T_C_S1)) == -1) err1_break(2, "type create")
    if (H5Tset_size(tid, strlen(value) + (addnull ? 1 : 0)) == -1) err1_break(2, "type size")
    if (H5Tset_strpad(tid, H5T_STR_NULLTERM) == -1) err1_break(2, "type pad")

    if (write_attribute(ctxid, obj_id, name, tid, value, valuelen) == -1) err_break(2)

    succeeded = 1;
  } while (0);

  if (1) {
    if (tid != -1) H5Tclose(tid);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Write an HDF5 string attribute when not adding NULL
 * 
 * @param ctxid context identifier
 * @param obj_id HDF5 group ID or dataset ID that the attribute is attached to.
 * @param name Attribute name
 * @param value Attribute value
 * 
 * @return 0 if successful
 */

static int write_attribute_string(hid_t ctxid, hid_t obj_id, const char *name, const char *value)
{
  return write_attribute_string_internal(ctxid, obj_id, name, value, 0);
}

/** 
 * @brief Write an HDF5 string attribute when adding NULL
 * 
 * @param ctxid context identifier
 * @param obj_id HDF5 group ID or dataset ID that the attribute is attached to.
 * @param name Attribute name
 * @param value Attribute value
 * 
 * @return 0 if successful
 */


static int write_attribute_string0(hid_t ctxid, hid_t obj_id, const char *name, const char *value)
{
  return write_attribute_string_internal(ctxid, obj_id, name, value, 1);
}

/** 
 * @brief Replace the "NAME" attribute of dim. scale with the dimension name.
 * NOTE: This routine should not be called for the <default> augmentation option.
         It should only be used for the <index> or the <file> option.
 * @param ctxid context identifier
 * @param dim   dimension
 * @param dimpath the absolute path of this dim.
 * 
 * @return 0 if successful
 */
int replace_dimscale_name_attr(hid_t ctxid, const dimension_t *dim, const challoc dimpath) {

  int succeeded=0;
  hid_t dest=-1;
  context_t *ctx;

  do {
    ctx = get_context(ctxid);
    if ((dest= H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0)  err2_break(2, "open dataset", dimpath.s)
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(2, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(2);
    if(H5Dclose(dest) < 0) err2_break(2, "close dataset", dimpath.s);
    
    succeeded = 1;
    } while(0);

  return succeeded? 0: -1;
}

/** *****************************WARNING**************************
 * NOTE: This function is NOT used by the current augmentation tool since it applies to solution 2 or 3 listed
 * at the document under 
    http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
 * The function is not reviewed.

 *******************************END OF WARNING ******************

 * @brief create a group specified by path and ancestor groups if necessary
 * 
 * @param ctxid context identifier
 * @param path absolute path to create
 * 
 * @return 0 if successful
 */
static hid_t create_group_rec(hid_t ctxid, const challoc *path)
{
  hid_t group = -1;
  challoc partial;
  int succeeded = 0;

  challoc_init(&partial);
  do {
    context_t *ctx = get_context(ctxid);
    int i = 0;
    int failinside = 0;

    for (i = 1; i <= path->len; ++i) {
      if (i == path->len || path->s[i] == '/') {
	if (!challoc_copyb(&partial, path->s, i)) nomem1_flagbreak(failinside, "path")
	
	if ((group = H5GCREATE(ctx->hdf5fd, partial.s)) < 0) err2_flagbreak(2, failinside, "create group", partial.s)

	if(!g_no_log_info)
	    msg2("create", partial.s);
      }
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (!succeeded) {
    if (group != -1) H5Gclose(group);
  }
  if (1) {
    challoc_free(&partial);
  }
  return succeeded ? group : -1;
}

/** *****************************WARNING**************************
 * NOTE: This function is NOT used by the current augmentation tool since it applies to solution 2 or 3 listed
 * at the document under 
    http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
 * The function is not reviewed.

 *******************************END OF WARNING ******************

 * @brief copy all elements in the source dataset, srcpath, to destpath
 * This is necessary for the third method. This function does not implement copying all elements yet.
 * 
 * @param ctxid context identifier
 * @param srcpath source dataset
 * @param destpath target path
 * 
 * @return 0 if successful
 */
static int clone_dataset(hid_t ctxid, const challoc *srcpath, const challoc *destpath)
{
  hid_t source = -1;
  hid_t dest = -1;
  hid_t tid = -1;
  hid_t spaceid = -1;
  hid_t plist = -1;
  hsize_t *chunksizes = NULL;
  void *buf = NULL;
  int succeeded = 0;

  do {
    context_t *ctx = get_context(ctxid);
    hssize_t npoints = 0;
    size_t elemsize  = 0;
   /* H5D_layout_t layout = H5D_CONTIGUOUS;*/

    if ((source = H5DOPEN(ctx->hdf5fd, srcpath->s)) < 0) err2_break(2, "open dataset", srcpath->s)
    if ((tid = H5Dget_type(source)) < 0) err2_break(2, "get type", srcpath->s)
    if ((spaceid = H5Dget_space(source)) < 0) err2_break(2, "get space", srcpath->s)

    if ((npoints = H5Sget_simple_extent_npoints(spaceid)) < 0) err2_break(2, "get npoints", srcpath->s)
    if ((elemsize = H5Tget_size(tid)) == 0) err2_break(2, "get size", srcpath->s)
    if ((buf = malloc(npoints * elemsize)) == NULL) nomem2_break("clone", srcpath->s)
    if (H5Dread(source, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(2, "read", srcpath->s)

    if ((plist = H5Dget_create_plist(source)) < 0) err2_break(2, "plist", srcpath->s)

     /*The checking filter call doesn't make sense. Why must the chunked dataset have filters?
        So comment this part out. KY 2013-10-16
      */
#if 0
    if ((layout = H5Pget_layout(plist)) < 0) err2_break(3, "layout", srcpath->s)
    if (layout == H5D_CHUNKED) {

      /*The checking filter call doesn't make sense. Why must the chunked dataset have filters?
        So comment this who part out. KY 2013-10-16
      */
      int nfilter = H5Pget_nfilters(plist);
      if (nfilter < 0) err1_break(3, "nfilters")
      /* This following line is not necessary. Chunking storage doesn't have to have a filter.
         If nfilter is 0, you don't want to set a dummy deflate filter. So comment this line out. KY 2013-10-16
      */
      /*
      else if (nfilter == 0) {
	if (H5Pset_filter(plist, H5Z_FILTER_DEFLATE, 0, 0, NULL) < 0) err1_break(3, "chunk")
      }
      */
    }
#endif

    if ((dest = H5DCREATE(ctx->hdf5fd, destpath->s, tid, spaceid, plist)) < 0) err2_break(2, "create dataset", destpath->s)
    if (H5Dwrite(dest, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(2, "write", destpath->s)

      if(!g_no_log_info)
	msg4("clone", srcpath->s, "to", destpath->s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (buf) free(buf);
    if (chunksizes) free(chunksizes);
    if (plist != -1) H5Pclose(plist);
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
    if (dest != -1) H5Dclose(dest);
    if (source != -1) H5Dclose(source);
  }
  return succeeded ? 0 : -1;
}

typedef struct {
  hid_t ctxid;
  const challoc *srcpath;
  const challoc *destpath;
} visit_data_t;

/**** *****************************WARNING**************************
 * NOTE: This function is NOT used by the current augmentation tool since it applies to solution 2 or 3 listed
 * at the document under 
    http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
 * The function is not reviewed.

 *******************************END OF WARNING ******************

 * @brief copy or link one dataset
 * This function is a call back function called by H5Giterate()
 * 
 * @param srcgrp an HDF5 group under HDFEOS group
 * @param name dataset name or group name
 * @param data an instance of visit_data_t structure
 * 
 * @return 0 if successful
 */
#if defined H5_USE_16_API
static herr_t link_grid_children(hid_t srcgrp, const char *name, void *data)
{
  challoc srcfullpath;
  challoc destfullpath;
  hid_t group = -1;
  int succeeded = 0;

  challoc_init(&srcfullpath);
  challoc_init(&destfullpath);
  do {
    const visit_data_t *visitdata = data;
    hid_t ctxid = visitdata->ctxid;
    context_t *ctx = get_context(ctxid);
    H5G_stat_t objinfo;
    int failinside = 0;

    if (!challoc_copy(&srcfullpath, visitdata->srcpath)) nomem1_break("srcpath")
    if (!challoc_cats(&srcfullpath, "/")) nomem1_break("srcpath")
    if (!challoc_cats(&srcfullpath, name)) nomem1_break("srcpath")
    if (!challoc_copy(&destfullpath, visitdata->destpath)) nomem1_break("destpath")
    if (!challoc_cats(&destfullpath, "/")) nomem1_break("destpath")
    if (!challoc_cats(&destfullpath, name)) nomem1_break("destpath")


    if (H5Gget_objinfo(srcgrp, name, 0, &objinfo) < 0) err2_break(2, "get objinfo", srcfullpath.s)
    switch (objinfo.type) {
      case H5G_GROUP:
	if ((group = H5GCREATE(ctx->hdf5fd, destfullpath.s)) < 0) err2_flagbreak(2, failinside, "create group", destfullpath.s)
	  if(!g_no_log_info)
	    msg2("create group", destfullpath.s);
	{
	  visit_data_t nested;
	  nested.ctxid = ctxid;
	  nested.srcpath = &srcfullpath;
	  nested.destpath = &destfullpath;
	  if(!g_no_log_info)
	    msg_push(srcfullpath.s);
	  if (H5Giterate(ctx->hdf5fd, srcfullpath.s, NULL, link_grid_children, &nested)) err2_flagbreak(2, failinside, "visit", srcfullpath.s)
	    if(!g_no_log_info)
	      msg_pop();
	}
	break;
      case H5G_DATASET:
	/* According to the augmentation method, this function will either create a link or clone a dataset */
	switch (ctx->augmentmethod) {
	  case 2:
	    if (H5LINK_CREATE_HARD(ctx->hdf5fd, srcfullpath.s, destfullpath.s) < 0) err3_flagbreak(2, failinside, "link", srcfullpath.s, destfullpath.s)
	    msg4("link", destfullpath.s, "to", srcfullpath.s);
	    break;
	  case 3:
	    if (clone_dataset(ctxid, &srcfullpath, &destfullpath) == -1) err3_flagbreak(2, failinside, "clone", srcfullpath.s, destfullpath.s)
	    break;
	  default:
	    ASSERT(0);
	}
	break;
      default:
	err2_flagbreak(3, failinside, "unknown objtype", srcfullpath.s)
	break;
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (group != -1) H5Gclose(group);
    challoc_free(&srcfullpath);
    challoc_free(&destfullpath);
  }
  return succeeded ? 0 : -1;
}

#else
static herr_t link_grid_children(hid_t srcgrp, const char *name, const H5L_info_t *linfo, void *data)
{
  challoc srcfullpath;
  challoc destfullpath;
  hid_t group = -1;
  int succeeded = 0;

  challoc_init(&srcfullpath);
  challoc_init(&destfullpath);
  do {
    const visit_data_t *visitdata = data;
    hid_t ctxid = visitdata->ctxid;
    context_t *ctx = get_context(ctxid);
    /*H5G_stat_t objinfo;*/
    H5O_info_t objinfo;
    int failinside = 0;

    if (!challoc_copy(&srcfullpath, visitdata->srcpath)) nomem1_break("srcpath")
    if (!challoc_cats(&srcfullpath, "/")) nomem1_break("srcpath")
    if (!challoc_cats(&srcfullpath, name)) nomem1_break("srcpath")
    if (!challoc_copy(&destfullpath, visitdata->destpath)) nomem1_break("destpath")
    if (!challoc_cats(&destfullpath, "/")) nomem1_break("destpath")
    if (!challoc_cats(&destfullpath, name)) nomem1_break("destpath")

    /*
    if (H5Gget_objinfo(srcgrp, name, 0, &objinfo) < 0) err2_break(2, "get objinfo", srcfullpath.s)
    if (H5Oget_info_by_name2(srcgrp, name,  &objinfo,H5O_INFO_BASIC,H5P_DEFAULT) < 0) err2_break(2, "get objinfo", srcfullpath.s)
    */
    if (H5OGET_INFO_BY_NAME(srcgrp, name,  &objinfo,H5P_DEFAULT) < 0) err2_break(2, "get objinfo", srcfullpath.s)
    switch (objinfo.type) {
      case H5O_TYPE_GROUP:
	if ((group = H5GCREATE(ctx->hdf5fd, destfullpath.s)) < 0) err2_flagbreak(2, failinside, "create group", destfullpath.s)
	  if(!g_no_log_info)
	    msg2("create group", destfullpath.s);
	{
	  visit_data_t nested;
	  nested.ctxid = ctxid;
	  nested.srcpath = &srcfullpath;
	  nested.destpath = &destfullpath;
	  if(!g_no_log_info)
	    msg_push(srcfullpath.s);
	  /*if (H5Giterate(ctx->hdf5fd, srcfullpath.s, NULL, link_grid_children, &nested)) err2_flagbreak(2, failinside, "visit", srcfullpath.s) */
	  if (H5Literate_by_name(ctx->hdf5fd, srcfullpath.s, H5_INDEX_NAME,H5_ITER_INC,NULL, link_grid_children, &nested,H5P_DEFAULT)) err2_flagbreak(2, failinside, "visit", srcfullpath.s)
	    if(!g_no_log_info)
	      msg_pop();
	}
	break;
      case H5O_TYPE_DATASET:
	/* According to the augmentation method, this function will either create a link or clone a dataset */
	switch (ctx->augmentmethod) {
	  case 2:
	    /*if (H5Glink(ctx->hdf5fd, H5G_LINK_HARD, srcfullpath.s, destfullpath.s) < 0) err3_flagbreak(2, failinside, "link", srcfullpath.s, destfullpath.s) */
	    if (H5LINK_CREATE_HARD(ctx->hdf5fd, srcfullpath.s, destfullpath.s) < 0) err3_flagbreak(2, failinside, "link", srcfullpath.s, destfullpath.s)
	    msg4("link", destfullpath.s, "to", srcfullpath.s);
	    break;
	  case 3:
	    if (clone_dataset(ctxid, &srcfullpath, &destfullpath) == -1) err3_flagbreak(2, failinside, "clone", srcfullpath.s, destfullpath.s)
	    break;
	  default:
	    ASSERT(0);
	}
	break;
      default:
	err2_flagbreak(3, failinside, "unknown objtype", srcfullpath.s)
	break;
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (group != -1) H5Gclose(group);
    challoc_free(&srcfullpath);
    challoc_free(&destfullpath);
  }
  return succeeded ? 0 : -1;
}
#endif
/**** *****************************WARNING**************************
 * NOTE: This function is NOT used by the current augmentation tool since it applies to solution 2 or 3 listed
 * at the document under 
    http://hdfeos.org/software/aug_eos5/doc/Investigation%20Report%20for%20Allowing%20netCDF-4%20to%20Access%20HDF-EOS5%20Files.pdf
 * The function is not reviewed.

 *******************************END OF WARNING ******************
 * @brief create an HDF5 group called NETCDF4 and copy or link every decendant group of HDFEOS group under NETCDF4 group
 * 
 * @param ctxid context identifier
 * @param eosobjname grid name or swath name
 * @param isgrid is this part of grid?
 * 
 * @return 0 if successful
 */
int create_aug_group(hid_t ctxid, const char *eosobjname, int isgrid)
{
  challoc srcpath;
  challoc destpath;
  hid_t destgrp = -1;
  int succeeded = 0;

  challoc_init(&srcpath);
  challoc_init(&destpath);
  do {
    context_t *ctx = get_context(ctxid);
    visit_data_t visitdata;

    if (make_grouppath(ctxid, eosobjname, isgrid, &srcpath) == -1) err_break(2)
    if (make_aug_grouppath(ctxid, eosobjname, isgrid, &destpath) == -1) err_break(2)
    if ((destgrp = create_group_rec(ctxid, &destpath)) < 0) err2_break(2, "create group", destpath.s)

      if(!g_no_log_info)
	msg_push(srcpath.s);
    visitdata.ctxid = ctxid;
    visitdata.srcpath = &srcpath;
    visitdata.destpath = &destpath;
#if defined H5_USE_16_API
    if (H5Giterate(ctx->hdf5fd, srcpath.s, NULL, link_grid_children, &visitdata)) err2_break(2, "visit", srcpath.s)
      if(!g_no_log_info)
#else
    if (H5Literate_by_name(ctx->hdf5fd, srcpath.s, H5_INDEX_NAME,H5_ITER_INC,NULL, link_grid_children, &visitdata,H5P_DEFAULT)) err2_break(2, "visit", srcpath.s)
      if(!g_no_log_info)
#endif
	msg_pop();

    succeeded = 1;
  } while (0);

  if (1) {
    if (destgrp != -1) H5Gclose(destgrp);
    challoc_free(&destpath);
    challoc_free(&srcpath);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Obtain the created dimension scale datatype of an HDF-EOS5 grid
 * 
 * @param ctx pointer to the context struct
 * @param gridname grid name
 * @param dim  this dimension
 * @param aug_option augmentation option(1-Index, 2-File, 0-Default)
 * 
 * @return dim. scale datatype if successful
 */

hid_t get_grid_dim_type(context_t *ctx, const char *gridname, const dimension_t *dim, const int aug_option) { 

  hid_t tid = H5Tcopy(H5T_IEEE_F32LE); /* for default or dim not specified */
  int i = 0;

  switch(aug_option) {
    case 1: /* -i for the dim. scale other than "XDim" and "YDim". "XDim" and "YDim" are for lat and lon.  */
      {
	      if(!(strcmp(dim->name.s, "XDim")==0 || (strcmp(dim->name.s, "YDim")==0)))

        /* For the index option, the datatype is always an integer. */
	      tid = H5Tcopy(H5T_NATIVE_INT);
	      break;
      }
    case 2: /* -f */
    {
      for(i=start_ptr_file; i<end_ptr_file; i++) {
          /* Check if the dim. scale values come from an ASCII data file */
	      if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
	      break;     
	    } 
	    if(i < end_ptr_file) {/* use the ASCII data file */
           /* Always set the datatype to floating data type */
	      tid = H5Tcopy(H5T_IEEE_F32LE);
	      goto TID_READY;
	    } 

    	for(i=start_ptr_index; i<end_ptr_index; i++) {
    	  if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
    	    break;
    	}
    	if(i < end_ptr_index) {/* use the index */
    
              /* Not for XDim or YDim */
    	  if(strcmp(dim->name.s, "XDim") == 0 || strcmp(dim->name.s, "YDim") == 0) {
                tid = -1;
    	  }
              /* For the index option, the datatype is always integer */
    	  tid = H5Tcopy(H5T_NATIVE_INT);
    	  goto TID_READY;
    	}

        /* For the option to use data fields in this file. */
      for(i=start_ptr_field; i<end_ptr_field; i++) { 
	      if(strcmp(dim->name.s, ctx->dimnames[i])==0)
	         break;
        }

        if(i < end_ptr_field) {
          /* "XDim" and "YDim" should be allowed to use geofield option
              since files may provide lat and lon to geofield, KY 2012-11-20 */
#if 0
	  if(strcmp(dim->name.s, "XDim") == 0 || strcmp(dim->name.s, "YDim") == 0) {
	    printf("%s is not allowed to use geofield option.\n", dim->name.s);
	    exit(0);
	  }
#endif
      	  const char *geoname = ctx->geonames[i];
                /* Not good to use the static array. But for HDF-EOS5, 1024 should be bigger enough*/
      	  char geopath[1024];
      	  hid_t source = -1;
      	  sprintf(geopath, (ctx->za_object==0)? EOSGEOFIELD: EOSGEOFIELD2,  gridname, geoname);
      
          source = H5DOPEN(ctx->hdf5fd, geopath);
#if 0
          // Somehow this data field geopath doesn't have to be existed. Need to check, 2018-09-13
          //if(source <0) {
          //  printf("Cannot open the geolocation field %s.\n", geopath);
          //  exit(0);
          //}
#endif
      	  if (source >=0) {
      	    tid = H5Dget_type(source);
            H5Dclose(source);
          }
      	}
      }
    }
TID_READY:
  return tid;
}

/** 
 * @brief Obtain the created dimension scale datatype of an HDF-EOS5 swath or zonal average
 * 
 * @param ctxid context identifier
 * @param swzaname swath or zonal average name
 * @param dim  this dimension
 * @param aug_option augmentation option(1-Index, 2-File, 0-Default)
 * 
 * @return dim. scale datatype if successful
 */


hid_t get_swza_dim_type(context_t *ctx, const char *swzaname, const dimension_t *dim, const int aug_option)
{
  char geopath[1024];
  hid_t tid=-1;
  hid_t source = -1;

  tid= H5Tcopy(H5T_IEEE_F32LE); /* default */

  switch(aug_option)
  {
  case 1:
    {
      /*For the index option, the dim. scale datatype is always integer */
      tid = H5Tcopy(H5T_NATIVE_INT); /* -i */
      break;
    }
  case 2: 
    {
      int i;

      /* If the dim. scale is from an ASCII data file, 
         the datatype is always floating-point */
      for(i=start_ptr_file; i<end_ptr_file; i++) {
      	if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
      	  break;
      }
      if(i < end_ptr_file) /* [dim][file] */
      { 
      	tid = H5Tcopy(H5T_IEEE_F32LE);
      	goto TID_OBTAINED;
      }
     
      for(i=start_ptr_index; i<end_ptr_index; i++)
      {
	      if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
	      break;
      }
      if(i < end_ptr_index) /* [dim] */
      {
        /*For the index option, the dim. scale datatype is always integer*/
	      tid = H5Tcopy(H5T_NATIVE_INT);
	      goto TID_OBTAINED;
      }

      for(i=start_ptr_field; i<end_ptr_field; i++) {
	      if(strcmp(dim->name.s, ctx->dimnames[i])==0)
	        break;
      }
      if(i < end_ptr_field) /* [dim][field] */
      {
      	const char *geoname = ctx->geonames[i];
      
              /* Here we always assume that dimension scale fields for a swath 
                 comes from the "Geolocation Fields". If having requests, this
                 assumption can be relaxed by also including the "Data Fields".
                 KY 2013-10-16
              */
      	sprintf(geopath, (ctx->za_object==0)? EOSGEOFIELD: EOSGEOFIELD2,  swzaname, geoname);
      
      	if (!((source = H5DOPEN(ctx->hdf5fd, geopath)) < 0))
      	   tid = H5Dget_type(source);
        if (source != -1) 
           H5Dclose(source);
      }
    }
  }

TID_OBTAINED:
  return tid;
}

/** 
 * @brief Create an HDF5 dimension scale dataset
 * 
 * @param ctxid context identifier
 * @param eosobjname grid or swath name
 * @param isgrid  if this EOS object is grid
 * @param ispure  if this dimension is a pure dimension without dimension scale
 * @param dim     dimension info(name and length)
 * @param ext_dimnames extensible dimension names (Need this to determine if this dimension is extensible)
 * @param aug_option    which augmentation otpion: index(1), file(2) or default(0)
 * 
 * @return 0 if successful
 */

int write_hdf5_dimension(hid_t ctxid, const char *eosobjname, int isgrid, int ispure, const dimension_t *dim, const namelistalloc *ext_dimnames, const int aug_option)
{
  hid_t tid = -1;
  hid_t spaceid = -1;
  hid_t propid = -1;
  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {
    context_t *ctx = get_context(ctxid);
    hsize_t dimsize = dim->dimsize;

    hsize_t maxdimsize = dim->dimsize;
    /* hsize_t maxdimsize = dim->maxdimsize; */
    int i = 0; 

    for (i = 0; i <ext_dimnames->len; i++) {

        /* Only when the dataset is extensible, the extensible dimension names will match the dim names, so set maxdimsize to H5S_UNLIMITED. */
        if(strcmp(dim->name.s,ext_dimnames->s[i].name.s) == 0) {
            /*printf("Maximum dim is unlimited and the dimension name is %s\n",dim->name.s);*/
            maxdimsize = H5S_UNLIMITED; 
            break;
        }
    }

    /* If dimsize is set to be H5S_UNLIMITED, it indicates(or the tool assumes) that the dim. size starts with 0, may need to investigate later KY 2013-10-16 */
    if (dimsize == H5S_UNLIMITED)
      dimsize = 0;

    if (make_dimpath(ctxid, eosobjname, isgrid, dim->name.s, &dimpath) == -1) err_break(2)

    if ((spaceid = H5Screate_simple(1, &dimsize, &maxdimsize)) < 0) err3_break(2, "H5Screate_simple", eosobjname, dim->name.s)

    if(isgrid) {
      tid = get_grid_dim_type(ctx, eosobjname, dim, aug_option);
      if (-1 == tid) err1_break(4,"For the index option, XDim or YDim are not allowed to use as dimension names in the mapping file. ") 
    } 
    else
      tid = get_swza_dim_type(ctx, eosobjname, dim, aug_option);

    if ((propid = H5Pcreate(H5P_DATASET_CREATE)) < 0) err1_break(2, "H5Pcreate")

    /* In HDF5, the unlimited dimension requires the chunking storage  */
    if (maxdimsize == H5S_UNLIMITED) {
      hsize_t chunksize = 0;
      if (dimsize != 0) 
         chunksize = dimsize;
      else 
         /* This is really not the right way to set the chunk to be 1 since the performance will be very bad. Can we find a better way?
            KY 2013-10-18 */
         /* I decide to set the chunk size to be 100 although it may also be not ideal. */
         chunksize = 100;
         /*chunksize = 1;*/
      if (H5Pset_chunk(propid, 1, &chunksize) < 0) err1_break(2, "H5Pset_chunk")
    }

    /* Create the HDF5 dataset */
    if ((dsetid = H5DCREATE(ctx->hdf5fd, dimpath.s, tid, spaceid, propid)) < 0) err2_break(2, "H5DCREATE", dimpath.s)

    /* Follow the HDF5 dimension scale spec. */ 
    if (write_attribute_string0(ctxid, dsetid, "CLASS", "DIMENSION_SCALE") == -1) err_break(2)

    if (ispure) {
      /* The created dimension will not contain any meaningful value. */
      if (write_attribute_string0(ctxid, dsetid, "NAME", "This is a netCDF dimension but not a netCDF variable.") == -1) err_break(2)
    }

    if(!g_no_log_info)
    {
      if (ispure)
	      msg2("create a pure dimension, mostly for the default option", dimpath.s);
      else
      	msg2("create a dimensional scale", dimpath.s);
    }

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&dimpath);
    if (dsetid != -1) H5Dclose(dsetid);
    if (propid != -1) H5Pclose(propid);
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Create an HDF5 dimension scale dataset for XDim and YDim in the HDF-EOS5 grid case.
 * The corresponding dim. scales for XDim and YDim are lat and lon. So the dim. scales may be
 * calcuated based on HDF-EOS5 APIs rather than explicitly provided.  
 * @param ctxid    context identifier
 * @param gridname grid  name
 * @param grid     can retrieve all info. related to this grid
 * @param isxdim   Is this dimension XDim?
 * @param ispure   Is the corresponding lat/lon 1-D(orthogonal) ?
 * @param aug_option    which augmentation otpion: index(1), file(2) or default(0)
 * 
 * @return 0 if successful
 */


int write_hdf5_implicit_dimension(hid_t ctxid, const char *gridname, const grid_t *grid, int isxdim, int ispure, const int aug_option)
{
  dimension_t dim;
  int succeeded = 0;

  init_dimension_t(&dim);
  do {

    /* Obtain XDim/YDim name and size. */
    if (!challoc_copys(&dim.name, isxdim ? "XDim" : "YDim")) nomem1_break("dim name")
    dim.dimsize = isxdim ? grid->info.xdim : grid->info.ydim;
    if (write_hdf5_dimension(ctxid, gridname, 1, ispure, &dim, &(grid->extendible_dimnames), aug_option) == -1) err_break(2)

    succeeded = 1;
  } while (0);

  if (1) {
    free_dimension_t(&dim);
  }
  return succeeded ? 0 : -1;
}

/* The functions listed below implement the HDF5 dimension scales for HDF-EOS5 by following the HDF5 dimensional scale specification.
   If necessary, these functions may be replaced by HDF5 dimension scale APIs and other handlings in the future. Documented at 
   https://jira.hdfgroup.uiuc.edu/browse/AUGEOSFV-12

   write_dimensionlist_attr,
   write_referencelist_attr,
   attach_dimension,
   associate_hdf5_dimension

   Generally the attach_dimension will hook up the connections between data fields and dimension scales.
   Comments and descriptions for these four functions will not be added for the time being. 
   KY 2013-10-16 
*/

static const char *HDF5_DIMENSION_LIST = "DIMENSION_LIST";

static int write_dimensionlist_attr(hid_t ctxid, hid_t h5dataset, const hobjreftalloc *dimscales)
{
  hvl_t *dimrefs = NULL;
  int deepfreecount = 0;
  hid_t spaceid = -1;
  hid_t tid = -1;
  hid_t attrid = -1;
  int succeeded = 0;

  do {

    if (dimscales->len) {

      hssize_t numelems = 0;
      int i = 0;
      int failinside = 0;

      hsize_t spacedimsizes = dimscales->len;

      H5E_BEGIN_TRY {
	      attrid = H5Aopen_name(h5dataset, HDF5_DIMENSION_LIST);
      } H5E_END_TRY;

      if (attrid != -1) {

      	/* This routine may be called more than once for fake dimscales such as lon and lat.
      	 * Check if previously written data are same as new input
      	 */
      	if ((spaceid = H5Aget_space(attrid)) == -1) err1_break(2, "H5Aget_space")
      	if ((tid = H5Aget_type(attrid)) == -1) err1_break(2, "H5Aget_type")
      	if ((numelems = H5Sget_simple_extent_npoints(spaceid)) == -1) err1_break(2, "get extent_npoints")
      	ASSERT(numelems == dimscales->len)
      
      	if ((dimrefs = malloc(dimscales->len * sizeof(hvl_t))) == NULL) nomem1_break("dim ref vl type")
      	if (H5Aread(attrid, tid, dimrefs) == -1) err1_break(2, "read dimension_list")
      
              /* The following for-loop just verifies that the "DIMENSION_LIST" value is what it should be.*/
      	for (i = 0; i < dimscales->len; ++i) {
      	  ASSERT(dimrefs[i].len == 1)
      	  ASSERT(*(hobj_ref_t *)dimrefs[i].p == dimscales->s[i])
      	}
      }
      else {

      	if ((spaceid = H5Screate_simple(1, &spacedimsizes, NULL)) == -1) err1_break(2, "H5Screate_simple")
      
      	/* I think the specification differs from the implementation. Both netCDF4 and H5DS
      	 * implementation uses VLEN of hobj_ref_t as datatype, and I follow this.
               * 
      	 */
      	if ((tid = H5Tvlen_create(H5T_STD_REF_OBJ)) == -1) err1_break(2, "H5Tvlen_create failure")
      
              /* Fill in dimrefs with object reference numbers from dimscales */
      	if ((dimrefs = malloc(dimscales->len * sizeof(hvl_t))) == NULL) nomem1_break("dim ref vl type")
      	for (i = 0; i < dimscales->len; ++i) {
      	  if ((dimrefs[i].p = malloc(sizeof(hobj_ref_t))) == NULL) nomem1_flagbreak(failinside, "dim ref vl[i].p")
      	  ++deepfreecount;
      	  *((hobj_ref_t *)dimrefs[i].p) = dimscales->s[i];
      	  dimrefs[i].len = 1;
      	}
      	if (failinside) 
          break;
      	if ((attrid = H5ACREATE(h5dataset, HDF5_DIMENSION_LIST, tid, spaceid, H5P_DEFAULT)) == -1) err1_break(2, "dataset attribute")
      	if (H5Awrite(attrid, tid, dimrefs) == -1) err1_break(2, "write dataset attribute")
      }
    }
    succeeded = 1;
  } while (0);

  if (1) {
    if (spaceid != -1) H5Sclose(spaceid);
    if (tid != -1) H5Tclose(tid);
    if (attrid != -1) H5Aclose(attrid);
    if (dimrefs != NULL) {
      while (deepfreecount--)
	      free(dimrefs[deepfreecount].p);
      free(dimrefs);
    }
  }
  return succeeded ? 0 : -1;
}

typedef struct {
  hobj_ref_t ref;
  int index;
} reference_backpointer_t;

static const char *HDF5_REFERENCE_LIST = "REFERENCE_LIST";

static int write_referencelist_attr(hid_t ctxid, hobj_ref_t dimscaleref, hobj_ref_t datasetref, int dimindex)
{
  hid_t dimscaleid = -1;
  hid_t spaceid = -1;
  hid_t oldspaceid = -1;
  hid_t tid = -1;
  hid_t attrid = -1;
  reference_backpointer_t *refs = NULL;
  int succeeded = 0;

  do {
    hssize_t numelems = 0;
    hsize_t dimsize = 0;
    context_t *ctx = get_context(ctxid);

    if ((dimscaleid = H5RDEREFERENCE(ctx->hdf5fd, H5R_OBJECT, &dimscaleref)) < 0) err1_break(2, "H5RDEREFERENCE")

    /* If "REFERENCE_LIST" is already there, one element should be appended. It seems that appending
     * is not supported. To handle this, append in the memory first, and then remove and recreate the attribute
     */ 
    H5E_BEGIN_TRY {
      attrid = H5Aopen_name(dimscaleid, HDF5_REFERENCE_LIST);
    } H5E_END_TRY;

    if (attrid != -1) {
      /* REFERENCE_LIST exists in ds dataset */
      if ((tid = H5Aget_type(attrid)) == -1) err1_break(2, "H5Aget_type")
      if ((oldspaceid = H5Aget_space(attrid)) == -1) err1_break(2, "H5Aget_space")
      if ((numelems = H5Sget_simple_extent_npoints(oldspaceid)) == -1) err1_break(2, "H5Sget_simple_extent_npoints")
      numelems++;
      if ((refs = malloc((int)numelems * sizeof(reference_backpointer_t))) == NULL) nomem1_break("reference_backpointer_t")
      if (H5Aread(attrid, tid, refs) == -1) err1_break(2, "H5Aread")
      if (H5Sclose(oldspaceid) == -1) err1_break(2, "H5Sclose")
      oldspaceid = -1;
      if (H5Aclose(attrid) == -1) err1_break(2, "H5Aclose")
      attrid = -1;
      if (H5Adelete(dimscaleid, HDF5_REFERENCE_LIST) == -1) err1_break(2, "H5Adelete")
    }
    else {
      numelems = 1;
      if ((refs = malloc((int)numelems * sizeof(reference_backpointer_t))) == NULL) nomem1_break("reference_backpointer_t 1")
    }
    refs[numelems - 1].ref = datasetref;
    refs[numelems - 1].index = dimindex;

    dimsize = numelems;
    if ((spaceid = H5Screate_simple(1, &dimsize, NULL)) == -1) err1_break(2, "H5Screate_simple")
    if (tid == -1) {
      if ((tid = H5Tcreate(H5T_COMPOUND, sizeof(reference_backpointer_t))) == -1) err1_break(2, "H5Tcreate")
      if (H5Tinsert(tid, "dataset", HOFFSET(reference_backpointer_t, ref), H5T_STD_REF_OBJ) == -1) err1_break(2, "H5Tinsert DATASET")
      if (H5Tinsert(tid, "index", HOFFSET(reference_backpointer_t, index), H5T_NATIVE_INT) == -1) err1_break(2, "H5Tinsert INDEX")
    }
    if ((attrid = H5ACREATE(dimscaleid, HDF5_REFERENCE_LIST, tid, spaceid, H5P_DEFAULT)) == -1) err1_break(2, "H5ACREATE")
    if (H5Awrite(attrid, tid, refs) == -1) err1_break(2, "write dimension scale data set attribute")

    succeeded = 1;
  } while (0);

  if (1) {
    if (refs != NULL) free(refs);
    if (attrid != -1) H5Aclose(attrid);
    if (tid != -1) H5Tclose(tid);
    if (oldspaceid != -1) H5Sclose(oldspaceid);
    if (spaceid != -1) H5Sclose(spaceid);
    if (dimscaleid != -1) H5Dclose(dimscaleid);
  }
  return succeeded ? 0 : -1;
}

static int attach_dimension(hid_t ctxid, const char *dsetpath, const hobjreftalloc *dimscales)
{
  hid_t dsetid = -1;
  hobj_ref_t dsetref;
  int succeeded = 0;

  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    /* DIMENSION_LIST */
    if ((dsetid = H5DOPEN(ctx->hdf5fd, dsetpath)) < 0) err2_break(2, "H5DOPEN", dsetpath)

    /* Write the dimension list attribute in the data field */
    if (write_dimensionlist_attr(ctxid, dsetid, dimscales) == -1) err_break(2)

    /* Write the REFERENCE_LIST attribute in the dimension scale */
    if (H5Rcreate(&dsetref, ctx->hdf5fd, dsetpath, H5R_OBJECT, -1) < 0) err2_break(2, "H5Rcreate dset", dsetpath)
    for (i = 0; i < dimscales->len; ++i) {
      if (write_referencelist_attr(ctxid, dimscales->s[i], dsetref, i) == -1) err_flagbreak(2, failinside)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
  }
  return succeeded ? 0 : -1;
}

int associate_hdf5_dimension(hid_t ctxid, const char *eosobjname, int isgrid, const fieldinfo_t *field, int isdatafield)
{
  hobjreftalloc dimscales;
  challoc dimpath;
  challoc dimnames;
  challoc dsetpath;
  int succeeded = 0;

  hobjreftalloc_init(&dimscales);
  challoc_init(&dimpath);
  challoc_init(&dimnames);
  challoc_init(&dsetpath);
  do {
    int i = 0;
    int failinside = 0;

    context_t *ctx = get_context(ctxid);

    if (!hobjreftalloc_ready(&dimscales, field->dims.len)) nomem1_break("obj ref")

    /* For each field, generate the object references to the dimension scale corresponding to the dimension */
    for (i = 0; i < field->dims.len; ++i) {
      const dimension_t *dim = &field->dims.s[i];

      if (make_dimpath(ctxid, eosobjname, isgrid, dim->name.s, &dimpath) == -1) err_flagbreak(2, failinside)
      if (!challoc_cats(&dimnames, i == 0 ? "[" : ",")) nomem1_flagbreak(failinside, "dim names")
      if (!challoc_cats(&dimnames, dim->name.s)) nomem1_flagbreak(failinside, "dim names")
      if (H5Rcreate(&dimscales.s[i], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err2_flagbreak(2, failinside, "H5Rcreate dim", dimpath.s)
      dimscales.len++;
    }

    if (failinside) break;
    if (!challoc_cats(&dimnames, "]")) nomem1_flagbreak(failinside, "dim names")

    if (make_fieldpath(ctxid, eosobjname, isgrid, field->name.s, isdatafield, &dsetpath) == -1) err_break(2)
    if (attach_dimension(ctxid, dsetpath.s, &dimscales) == -1) err_break(2)

      if(!g_no_log_info)
	      msg4("associate dimension", dsetpath.s, "with", dimnames.s);

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&dimpath);
    challoc_free(&dimnames);
    challoc_free(&dsetpath);
    hobjreftalloc_free(&dimscales);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Helper function to obtain the latitude and longitude HDF5 datatype id
 * 
 * @param ctxid context identifier
 * @param gridname grid name
 * @param islongitude: Is this longitude?
 * 
 * @return Latitude or Longitude HDF5 datatype ID if successful
 */

static hid_t get_lonlat_type(hid_t ctxid, const char *gridname, int islongitude) {

  hid_t source = -1;
  hid_t tid = -1;
  char path[1024];
  context_t *ctx = get_context(ctxid);
  int succeeded = 0;
  
  if(!islongitude)
    sprintf(path, EOSDATAFIELD, gridname, "Latitude");
  else
    sprintf(path, EOSDATAFIELD, gridname, "Longitude");
  
  do {
    if((source = H5DOPEN(ctx->hdf5fd, path)) < 0) err2_break(2, "H5DOPEN", path)
    if((tid = H5Dget_type(source)) < 0) err2_break(2, "H5Dget_type", path)
    succeeded = 1;
  } while(0);

  if(source != -1) 
     H5Dclose(source);

  return succeeded? tid: -1;
}
/** 
 * @brief Obtain the Latitude or Longitude values. This is currently only used to retrieve TES data.
 * NOTE: This may be used to support the 2-D swath case or the general grid case. KY 2013-10-18
 * @param ctxid context identifier
 * @param gridname grid name
 * @param islongitude: Is this longitude?
 * 
 * @return latitude or longitude data if successful
 */
void* get_coord_from_geofield(hid_t ctxid, const char *gridname, int islongitude) {

  int succeeded = 0;
  void* coord = NULL;
  
  do {
    hid_t source = -1;
    hid_t tid = -1;
    hid_t spaceid = -1;
    hssize_t npoints = 0;
    size_t elemsize = 0;

    context_t* ctx = get_context(ctxid);
    char path[1024];

    if(!islongitude)
      sprintf(path, EOSDATAFIELD, gridname, "Latitude");
    else
      sprintf(path, EOSDATAFIELD, gridname, "Longitude");

    if((source = H5DOPEN(ctx->hdf5fd, path)) < 0) err2_break(2, "H5DOPEN", path)
    if((tid = H5Dget_type(source)) < 0) err2_break(2, "H5Dget_type", path)
    if((spaceid = H5Dget_space(source)) < 0) err2_break(2, "H5Dget_space", path) 
    
    if((npoints = H5Sget_simple_extent_npoints(spaceid)) < 0) err2_break(2, "H5Sget_simple_extent_npoints", path)  						
    if((elemsize = H5Tget_size(tid)) == 0) err2_break(2, "H5Tget_size", path)
    if((coord = malloc(npoints*elemsize)) == NULL) nomem2_break("malloc", path)
    if(H5Dread(source, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, coord) < 0) err2_break(2, "H5Dread", path)										
    if(source != -1)
       H5Dclose(source);
    if(tid != -1) 
       H5Tclose(tid);

    succeeded = 1;
  } while(0); 

  return succeeded? coord: NULL;
} 

/* Need to check if this function works or not. No comments or descriptions are currently added. AUGEOSFV-8 KY 2013-10-16*/
#if defined H5_USE_16_API
herr_t fill_hdf5_dimension_on_attr_iter1(hid_t src, const char *attrname, void *arg)
{
  hid_t aid = -1;
  hid_t tid = -1;
  hid_t aid2 = -1;
  hid_t sid = -1;
  hid_t dest = *(hid_t*)arg;
  hid_t retval = 0;
  size_t size = 0;
  
  /* We copy all atttibutes except "Title" and "DIMENSION_LIST" */
  if(strcmp(attrname, "Title")==0)
    return 0;
  else if(strcmp(attrname, "DIMENSION_LIST")==0) 
    return 0; 

  if((aid = H5Aopen(src, attrname, H5P_DEFAULT)) < 0)
    return -1;
  if((tid = H5Aget_type(aid))<0) {
    H5Aclose(aid);
    return -1;
  }
  if((sid = H5Aget_space(aid))<0){
    H5Aclose(aid);
    return -1;
  }
  if((size = H5Aget_storage_size(aid))==0){
    H5Aclose(aid);
    return -1;
  }

  void *buf = malloc(size);
  if(buf==NULL) {
    H5Aclose(aid);
    return -1;
  }
  
  if((retval = H5Aread(aid, tid, buf)) < 0){
    H5Aclose(aid);
    free(buf);
    return -1;
  }

  H5Aclose(aid);
  if((aid2=H5ACREATE(dest, attrname, tid, sid, H5P_DEFAULT))<0){
    free(buf);
    return -1;
  }

  if((retval = H5Awrite(aid2, tid, buf))<0){
    free(buf);
    H5Aclose(aid2);
    return -1;
  }

  if((retval = H5Aclose(aid2))<0) {
     free(buf);
     return -1;
  }

  free(buf);
  
  return 0;
}
#else


/* Need to check if this function works or not. No comments or descriptions are currently added. AUGEOSFV-8 KY 2013-10-16*/
herr_t fill_hdf5_dimension_on_attr(hid_t src, const char *attrname, const H5A_info_t *ainfo, void *arg)
{
  hid_t aid = -1;
  hid_t tid = -1;
  hid_t aid2 = -1;
  hid_t sid = -1;
  hid_t dest = *(hid_t*)arg;
  hid_t retval = 0;
  size_t size = 0;
  
  /* We copy all atttibutes except "Title" and "DIMENSION_LIST" */
  if(strcmp(attrname, "Title")==0)
    return 0;
  else if(strcmp(attrname, "DIMENSION_LIST")==0) 
    return 0; 

  if((aid = H5Aopen(src, attrname, H5P_DEFAULT)) < 0)
    return -1;
  if((tid = H5Aget_type(aid))<0) {
    H5Aclose(aid);
    return -1;
  }
  if((sid = H5Aget_space(aid))<0){
    H5Aclose(aid);
    return -1;
  }
  if((size = H5Aget_storage_size(aid))==0){
    H5Aclose(aid);
    return -1;
  }

  void *buf = malloc(size);
  if(buf==NULL) {
    H5Aclose(aid);
    return -1;
  }
  
  if((retval = H5Aread(aid, tid, buf)) < 0){
    H5Aclose(aid);
    free(buf);
    return -1;
  }

  H5Aclose(aid);
  if((aid2=H5ACREATE(dest, attrname, tid, sid, H5P_DEFAULT))<0){
    free(buf);
    return -1;
  }

  if((retval = H5Awrite(aid2, tid, buf))<0){
    free(buf);
    H5Aclose(aid2);
    return -1;
  }

  if((retval = H5Aclose(aid2))<0) {
     free(buf);
     return -1;
  }

  free(buf);
  
  return 0;
}

#endif
/** 
 * @brief Fill dimension scales with index numbers.
 * 
 * @param ctxid   context identifier
 * @param dim     Can retrieve all the information of this dimension
 * @param dimpath The absolute path and the size of the added dimension scale dataset
 * @return FAIL if failed, SUCCEED if successful
 */

int fill_dim_scale_with_index(hid_t ctxid, const dimension_t *dim, const challoc dimpath) {

  int succeeded=0;

  if(strcmp(dim->name.s,"Unlim")==0) 
     return succeeded;

  hid_t dest=-1;
  context_t *ctx;

  do {
    int i = 0;
    int ds = dim->dimsize;

    ctx = get_context(ctxid);
    
    /* We will skip dimension "Scale" Unlim since this is just used by HDF-EOS lib internally.
       The information is already used by the augmentation tool. */

    if ((dest= H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0)  err2_break(2, "open dataset", dimpath.s)

    /* The NAME attribute needs to be changed from netCDF predefined pure dimension attribute
       to its own name, otherwise, the dim. scale will not be read by the netCDF library. KY 2013-10-18 */
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(2, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(2);
 
    int* padding_data =  malloc(ds*sizeof(int));
    for(i=0; i<ds; i++)
      padding_data[i] = i+1;
    if (H5Dwrite(dest, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, padding_data) < 0) err2_break(2, "write", dimpath.s)
    
    free(padding_data);
    if(H5Dclose(dest) < 0) err2_break(2, "close dataset", dimpath.s);
    
    succeeded = 1;
  } while(0);

  return succeeded? 0: -1;
}

/** 
 * @brief Fill dimension scales of XDim and YDim with data values obtained from an ASCII file.
 * 
 * @param ctxid   context identifier
 * @param dimname   dimension name 
 * @param dimsize   dimension size (for XDim or YDim)
 * @param dimpath The absolute path and the size of the added dimension scale dataset
 * @param pos     The position of this data file among all possible data files
 * @return FAIL if failed, SUCCEED if successful
 * 
 */

/* This function is almost the same as fill_dim_scale_with_data except the input parameters
   The function that calls this function(fill_hdf5_1d_dimension) doesn't have dim. info as input.
   Need to investigate why. May be due to the special handling of XDim and YDim. KY 2013-10-18
*/
int fill_xydim_scale_with_data(hid_t ctxid, const char *dimname, const int dimsize, const challoc dimpath,int pos) {
 
  hid_t dest=-1;
  context_t *ctx;
 
  int succeeded=0;

  do {
    ctx = get_context(ctxid);
    if ((dest= H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err2_break(2, "open dataset", dimpath.s)

     /* The NAME attribute needs to be changed from netCDF predefined pure dimension attribute
       to its own name, otherwise, the dim. scale will not be read by the netCDF library. KY 2013-10-18 */
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(2, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dimname) == -1) err_break(2);
   
    float* coord = malloc(dimsize*sizeof(float));
    if(NULL == coord) {
       H5Dclose(dest);
       err1_break(2,"Cannot allocate memory for the coordinate");
    } 

    /* The data values stored in the ASCII file have been read into the memory and stored inside dim_data.
       For better performance, the data may just be retrieved to save some memories. However, since the 
       dim. scale data should be small, so it may be okay to store the data in the memory at the beginning.
       Will see if it needs to get improved.

       KY 2013-10-16
    */
    doublealloc da = ctx->dim_data[pos];
    if(dimsize != da.len) {
      printf("The required data length [%d] is different from the length [%lu] of data file. Please check the data file to eliminate the inconsistency.\n", dimsize, da.len);
      free(coord);
      err_break(4);
    }
    int k = 0;
    for(k=0; k<dimsize; k++) {
      coord[k] = (float)da.s[k];
    }
    if (H5Dwrite(dest, H5T_IEEE_F32LE, H5S_ALL, H5S_ALL, H5P_DEFAULT, coord) < 0) {
       free(coord);
       err2_break(2, "write", dimpath.s);
    }

    free(coord);

    if(H5Dclose(dest) < 0) err2_break(2, "close dataset", dimpath.s);
    
    succeeded = 1;
  } while(0);

  return succeeded? 0: -1;
}

/** 
 * @brief For HDF-EOS5 grid, fill dimension scales of non-XDim/YDim dimensions with data values obtained from an ASCII file.
 * 
 * @param ctxid   context identifier
 * @param dim     Can retrieve all the info. related to this dimension
 * @param dimpath The absolute path and the size of the added dimension scale dataset
 * @param pos     The index(or position) of this data file among all possible ASCII data files in the mapping file
 * @return FAIL if failed, SUCCEED if successful
 * 
 */
int fill_dim_scale_with_data(hid_t ctxid, const dimension_t *dim, const challoc dimpath,int pos) {
  
  hid_t dest=-1;
  context_t *ctx;

  int succeeded = 0;

  do {
    int ds = dim->dimsize;
    ctx = get_context(ctxid);
    if ((dest= H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err2_break(2, "open dataset", dimpath.s)

    /* Rename NAME attribute
       The original value of the NAME attribute is the predefined attribute for pure dimension
       "This is a netCDF dimension but not a netCDF variable".
       With this value, NetCDF cannot read any value in it. */
    if (H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(2, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(2);

    float* coord = malloc(ds*sizeof(float));
    doublealloc da = ctx->dim_data[pos];
    if(ds != da.len) {
      printf("The required data length [%d] is different from the length [%lu] of data file. Please check the data file to eliminate the inconsistency.\n", ds, da.len);
      free(coord);
      err_break(4);
    }
    int k = 0;
    for(k=0; k<ds; k++) {  
      coord[k] = (float)da.s[k];
    }
    if (H5Dwrite(dest, H5T_IEEE_F32LE, H5S_ALL, H5S_ALL, H5P_DEFAULT, coord) < 0) {
       free(coord);
       err2_break(2, "write", dimpath.s);
    }
    free(coord);

    if(H5Dclose(dest) < 0) err2_break(2, "close dataset", dimpath.s);
    
    succeeded = 1;
  } while(0);

  return succeeded? 0: -1;
}

/** 
 * @brief Fill dimension scale values from a data field in the HDF-EOS5 file 
 * 
 * @param ctxid   context identifier
 * @param eosobjname EOS5 object(grid,swath or zonal average) name
 * @param dim     Can retrieve all the info. related to this dimension
 * @param dimpath The absolute path and the size of the added dimension scale dataset
 * @param isgrid  Is this HDF-EOS5 grid, 1 for grid 0 for other objects
 * @param field_pos     The position(or index) of this data field among all data fields in the mapping file
 * @return FAIL if failed, SUCCEED if successful
 * 
 */
int fill_dim_scale_with_field(hid_t ctxid, const char *eosobjname, const dimension_t *dim, const challoc dimpath, int isgrid, const int field_pos) {

  hid_t source=-1;
  hid_t tid=-1;
  hid_t spaceid=-1;
  hid_t  dest=-1;

  /*dims_eos5 means the dims obtained from the EOS5 API */
  hid_t dims_eos5_spaceid = -1;
  int geo_field_ndims = -1;
 
  hssize_t dims_eos5_npoints = 0;

  hssize_t npoints = 0;
  size_t elemsize = 0;
  void *buf = NULL;

  char geopath[1024];
  int succeeded = 0;
      
  do {
    context_t *ctx = get_context(ctxid);
    const char *geoname = ctx->geonames[field_pos];

    hsize_t dimsize = dim->dimsize;

    if(dimsize == H5S_UNLIMITED)  
      dimsize = 0;

    if(isgrid)
      sprintf(geopath, EOSDATAFIELD, eosobjname, geoname);
    else
      sprintf(geopath, (ctx->za_object==0)? EOSGEOFIELD: EOSGEOFIELD2,  eosobjname, geoname);
    if ((source = H5DOPEN(ctx->hdf5fd, geopath)) < 0) 
    {
      fprintf(stderr, "Warning: Dimension name %s cannot be filled since %s does not exist.\n", dim->name.s, geopath);
	break;
    }

    if ((tid = H5Dget_type(source)) < 0) err2_break(2, "get type", geopath)
    if ((spaceid = H5Dget_space(source)) < 0) err2_break(2, "get space", geopath)

    if ((npoints = H5Sget_simple_extent_npoints(spaceid)) < 0) err2_break(2, "get npoints", geopath)

    if ((geo_field_ndims = H5Sget_simple_extent_ndims(spaceid)) <0) err2_break(2,"get number of dimensions", geopath);
    if (geo_field_ndims != 1) err2_break(2,"Current the dimensional scale must be 1 ",geopath);

    if ((elemsize = H5Tget_size(tid)) == 0) err2_break(2, "get size", geopath)
    if ((buf = malloc(npoints * elemsize)) == NULL) nomem2_break("clone", geopath)
    if (H5Dread(source, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(2, "read", geopath)

    if ((dest= H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err2_break(2, "open dataset", dimpath.s);

    if((dims_eos5_spaceid = H5Dget_space(dest)) <0) err2_break(2,"get space",dimpath.s)
    if ((dims_eos5_npoints = H5Sget_simple_extent_npoints(dims_eos5_spaceid)) < 0) err2_break(2, "get dim. npoints based on eos5 APIs", dimpath.s)

    if (dims_eos5_npoints < npoints) {
        hsize_t dim_size[1];
        dim_size[0] = npoints;
        /* Using this function doesn't make sense, see AUGEOSFV-9 KY 2013-10-18*/
        if ((H5Dset_extent(dest,dim_size)) < 0) err2_break(2,"cannot extend the dataset ",dimpath.s);
    }

    if (H5Dwrite(dest, tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, buf) < 0) err2_break(2, "write", dimpath.s)

    /* Rename NAME attribute
       The original value of the NAME attribute is the netCDF predefined attribute for pure dimension
       "This is a netCDF dimension but not a netCDF variable".
       With this value, NetCDF cannot read any value in it. */
    if(H5Aexists(dest, "NAME") > 0) 
      if(H5Adelete(dest, "NAME")<0) err2_break(2, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(2);

    /* Copy attributes from the original dimension scale */
    /* Seems that the attributes are not copied over. (AUGEOSFV-8) for more details KY 2013-10-18*/
#if defined H5_USE_16_API
    {
        if(H5Aiterate1(source, NULL, fill_hdf5_dimension_on_attr_iter1, (void*)&dest)<0)
            err2_break(2, "copy attributes", dimpath.s)
    }
#else
    {
     hsize_t n=0;
     if(H5Aiterate2(source, H5_INDEX_CRT_ORDER, H5_ITER_NATIVE, &n, fill_hdf5_dimension_on_attr, (void*)&dest)<0)
       err2_break(2, "copy attributes", dimpath.s)
    }
#endif

    /* Remove DIMENSION_LIST, comment out the following line, removing DIMENSION_LIST by H5Aiterate2. */
    /*if(H5Adelete(dest, "DIMENSION_LIST")<0) err2_break(2, "del dimension list", dimpath.s); */
    succeeded = 1;
  } while(0);

  if(1) {
    if(buf) free(buf);
    if(spaceid != -1) H5Sclose(spaceid);
    if(tid != -1) H5Tclose(tid);
    if(dest != -1) H5Dclose(dest);
    if(source != -1) H5Dclose(source);
  }

  return succeeded? 0: -1;
}

/** 
 * @brief Fill dimension scale values for dimensions that are not for horizontal planes
 * 
 * @param ctxid       context identifier
 * @param gridname    grid name
 * @param dim         Can retrieve all the info. related to this dimension
 * @param aug_option  Index(1) or File(2) or the default option(0)
 * @return FAIL if failed, SUCCEED if successful
 *  
 */
int fill_hdf5_1d_3rd_dimension(hid_t ctxid, const char *gridname, const dimension_t  *dim, const int aug_option) {

   challoc dimpath;
   hobjreftalloc dimscales;
   int succeeded = 0;
   int failinside = 0;

   challoc_init(&dimpath);
   hobjreftalloc_init(&dimscales);

   do {
     context_t *ctx = get_context(ctxid);
     if(make_dimpath(ctxid, gridname, 1, dim->name.s, &dimpath) == -1) err_break(3)
     switch(aug_option) {
       case 0:
    	 {
               /* For the default option, we should NOT change the NAME. We should still
                  keep the netCDF predefined attribute.  KY 2013-10-17*/
    	   /*replace_dimscale_name_attr(ctxid, dim, dimpath);*/
    	   break;
    	 }
           case 1:
    
             /* Index option */
    	 {
    	   if(fill_dim_scale_with_index(ctxid, dim, dimpath) == -1) err_flagbreak(2,failinside)
               else
                  goto FINISHED;
    	 }
           case 2:
    	 {
    
               /* -f option, the dimension scale can come from a data field, or an ASCII file, or
                     can be just natural numbers. */
    
    	   int i = 0;
    	   for(i=start_ptr_index; i<end_ptr_index; i++) {
    	     if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
    	        break;
    	   }
    	  
    	   if(i < end_ptr_index) {/* use index */
    	     if(fill_dim_scale_with_index(ctxid, dim, dimpath) == -1) err_flagbreak(2,failinside)
           else 
    	        goto FINISHED; 
    	   }
    
    	   for(i=start_ptr_file; i<end_ptr_file; i++) {
           if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
             break;
         } 
         if(i < end_ptr_file) {/* use file */
           if(fill_dim_scale_with_data(ctxid,dim,dimpath,i) == -1) err_flagbreak(2,failinside) 
           else 
             goto FINISHED;
         }
    
    	   for(i=start_ptr_field; i<end_ptr_field; i++) {/* use data field */
    	     if(strcmp(dim->name.s, ctx->dimnames[i])==0)
                   break;
    	   }
    	   if(i < end_ptr_field) {
    	     if(fill_dim_scale_with_field(ctxid, gridname, dim, dimpath, 1, i) == -1) err_flagbreak(2,failinside)
           else 
             goto FINISHED;
    	   }
         /*none of the above, use the default option */
    
    	 } /* end of case */
     } /* end of switch */

     if( 1== failinside)
        break;
     /* Investigation reveals that the code lines from "hobjreftalloc ..." to "attach_dimension ..." may not be necessary since
       it just creates a "DIMENSION_LIST" attribute to a dimensional scale itself, which is not necessary.
       I just checked the aug. file with ncdump and found that by doing so it doesn't violate netCDF-4 data model. 
       So not remove them because that may break
       the backward compatibility of existing HDF-EOS5 files that were created this way. Will take a serious look
       at these if priority of this tool bumps and users give feedback. KY 2013-10-17 
    */
FINISHED:	         
     if (!hobjreftalloc_ready(&dimscales, 1)) nomem1_break("obj ref")
     if (H5Rcreate(&dimscales.s[0], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err1_break(2, "H5Rcreate")	
     dimscales.len++;

     if (attach_dimension(ctxid, dimpath.s, &dimscales) == -1) err_break(2)

     if(!g_no_log_info)
       msg2("fill 1d dimension", dimpath.s);

      succeeded = 1;
  } while (0);

  if (failinside == 1)
     succeeded = 0;
  if (1) {
    hobjreftalloc_free(&dimscales);
    challoc_free(&dimpath);
  }

  return succeeded ? 0 : -1;
}

/** 
 * @brief Fill in the 1-D dimension 
 * 
 * @param ctxid       context identifier
 * @param gridname    grid name
 * @param islongitude  Is this longitude
 * @param coord        dimension scale value
 * @param aug_option  Index(1) or File(2) or the default option(0)
 * @return FAIL if failed, SUCCEED if successful
 *  
 */
int fill_hdf5_1d_dimension(hid_t ctxid, const grid_t *grid, const char *gridname, int islongitude, const doublealloc *coord,const int aug_option)
{
  hid_t dsetid  = -1;
  hid_t memtype = -1;
  hid_t spaceid = -1;
  hid_t atype   = -1;
  hid_t attr    = -1;

  challoc dimpath;
  hobjreftalloc dimscales;
  int succeeded = 0;
  int failinside = 0;

  float* fbuf = NULL;

    dimension_t dim;
    init_dimension_t(&dim);

  challoc_init(&dimpath);
  hobjreftalloc_init(&dimscales);
  do {

    context_t *ctx = get_context(ctxid);

    if (!challoc_copys(&dim.name, islongitude ? "XDim" : "YDim")) nomem1_break("dim name")
    dim.dimsize = islongitude ? grid->info.xdim : grid->info.ydim;

    if (make_dimpath(ctxid, gridname, 1, (&dim)->name.s, &dimpath) == -1) err_break(3)

    if(aug_option == 2) {
      int i;
      for(i=start_ptr_file; i<end_ptr_file; i++) {
        if(strcmp((&dim)->name.s, ctx->dim_data_name[i]) == 0)
          break;
      }
      if(i < end_ptr_file) {/* use file */
        if(fill_xydim_scale_with_data(ctxid, (&dim)->name.s, coord->len, dimpath,i) == -1) err_flagbreak(2,failinside)
        else 
	       goto FILLED;
      }
      for(i=start_ptr_field; i<end_ptr_field; i++) {/* use data field */
        if(strcmp((&dim)->name.s, ctx->dimnames[i])==0)
          break;
      }
      if(i < end_ptr_field) {
        if(fill_dim_scale_with_field(ctxid, gridname, &dim, dimpath, 1, i) == -1) err_flagbreak(2,failinside) 
        else
          goto FILLED;
      }
    } 

    atype = H5T_NATIVE_DOUBLE;

    char *path = "/HDFEOS/ADDITIONAL/FILE_ATTRIBUTES/";
    if((dsetid = H5GOPEN(ctx->hdf5fd, path)) < 0) err1_break(2, "H5GOPEN")
    attr = H5Aexists(dsetid, "InstrumentName");

    if(attr>0 && strncmp("TES", ctx->instrument_name,ctx->len_of_instrument_name) == 0) /* the attribure "InstrumentName" is present  */
    {
      if((fbuf = get_coord_from_geofield(ctxid, gridname, islongitude)) != NULL)
        atype = get_lonlat_type(ctxid, gridname, islongitude);
    }

    hsize_t dimsize = coord->len;

    if ((memtype = H5Tcopy(atype)) < 0) err1_break(2, "H5Tcopy")
    if ((spaceid = H5Screate_simple(1, &dimsize, NULL)) < 0) err1_break(2, "H5Screate_simple")

    if ((dsetid = H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err1_break(2, "H5DOPEN")
    if (fbuf == NULL) { 
      if (H5Dwrite(dsetid, memtype, spaceid, spaceid, H5P_DEFAULT, coord->s) < 0) err1_break(2, "H5Dwrite")
    } else {
      if (H5Dwrite(dsetid, memtype, spaceid, spaceid, H5P_DEFAULT, fbuf) < 0) err1_break(2, "H5Dwrite")
    }

    /* The code lines from "hobjreftalloc ..." to "attach_dimension ..." may not be necessary since
       it just creates a "DIMENSION_LIST" attribute to a dimensional scale itself, which is not necessary.
       I just checked the aug. file with ncdump and found that by doing so it doesn't violate netCDF-4 data model. 
        So not remove them because that may break
       the backward compatibility of existing HDF-EOS5 files that were created this way. Will take a serious look
       at these if priority of this tool bumps and users give feedback. KY 2013-10-17 
    */   
FILLED:
    if (!hobjreftalloc_ready(&dimscales, 1)) nomem1_break("obj ref")
    if (H5Rcreate(&dimscales.s[0], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err1_break(2, "H5Rcreate")
    dimscales.len++;

    if (attach_dimension(ctxid, dimpath.s, &dimscales) == -1) err_break(2)

    if(!g_no_log_info)
      msg2("fill 1d dimension", dimpath.s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (spaceid != -1) H5Sclose(spaceid);
    if (memtype != -1) H5Tclose(memtype);
    if (dsetid != -1) H5Dclose(dsetid);
    if (atype != -1) H5Tclose(atype);
    hobjreftalloc_free(&dimscales);
    free_dimension_t(&dim);
    challoc_free(&dimpath);
    if(fbuf != NULL) free(fbuf);
  }

  if(failinside == 1) 
     succeeded = 0;

  return succeeded ? 0 : -1;
}

/** 
 * @brief Fill dimension scales for 2-D lat/lon.
 * Note: This function needs more investigation.
 * @param ctxid   context identifier
 * @param gridname   Grid name
 * @param islongitude  is this the longitude
 * @param coord        lat or lon data
 * @param ydimmajor    Is this ydim major 
 * @param xsize        xdim size
   @param ysize        ydim size
 * @return FAIL if failed, SUCCEED if successful
 */


int fill_hdf5_2d_dimension(hid_t ctxid, const char *gridname, int islongitude, const doublealloc *coord, int ydimmajor, long xsize, long ysize)
{
  hid_t dsetid = -1;
  hid_t memtype = -1;
  hid_t dsettype = -1;
  hid_t spaceid = -1;
  challoc cvarpath;
  challoc dimpath;
  hobjreftalloc dimscales;
  int succeeded = 0;

  challoc_init(&cvarpath);
  challoc_init(&dimpath);
  hobjreftalloc_init(&dimscales);
  do {
    int i =0;
    int failinside = 0;
    context_t *ctx = get_context(ctxid);
    const char *cvname = islongitude ? "lon" : "lat";
    hsize_t dimsizes[2];

    if (ydimmajor) { 
       dimsizes[0] = ysize; 
       dimsizes[1] = xsize; 
    }
    else { 
       dimsizes[0] = xsize; 
       dimsizes[1] = ysize; 
    }

    /* why memtype is NATIVE_DOUBLE and dsettype is IEEE_F32LE , need to investigate...  
       Find the reason: HDF-EOS5 returns lat/lon in double but for normal lat/lon we want to obtain in floating-point,
       This approach will not save the disk space and the HDF5 lib. will do the conversion. It is an OK approach for this single dataset.
       Need to watch for performance for big arrays.
       KY 2013-10-18*/
    if ((memtype = H5Tcopy(H5T_NATIVE_DOUBLE)) < 0) err1_break(2, "H5Tcopy")
    if ((dsettype = H5Tcopy(H5T_IEEE_F32LE)) < 0) err1_break(2, "H5Tcopy")
    if ((spaceid = H5Screate_simple(2, dimsizes, NULL)) < 0) err1_break(2, "H5Screate_simple")

    if (make_dimpath(ctxid, gridname, 1, cvname, &cvarpath) == -1) err_break(3)
    if ((dsetid = H5DCREATE(ctx->hdf5fd, cvarpath.s, dsettype, spaceid, H5P_DEFAULT)) < 0) err2_break(2, "H5DCREATE", cvarpath.s)
    if (H5Dwrite(dsetid, memtype, spaceid, spaceid, H5P_DEFAULT, coord->s) < 0) err1_break(2, "H5Dwrite")

    /* The code from hobjreftalloc_ready to attach_dimension may be necessary since this is for 2-D lat/lon case. Will investigate this in the future. 
       KY 2013-10-17 
    */
    if (!hobjreftalloc_ready(&dimscales, 2)) nomem1_break("obj ref")
    for (i = 0; i < 2; ++i) {
      const char *dimname = ydimmajor ? (i == 0 ? "YDim" : "XDim") : (i == 0 ? "XDim": "YDim");

      /* failinside will be set to 1 if error occurs and the program will exit from the for loop. */
      if (make_dimpath(ctxid, gridname, 1, dimname, &dimpath)) err_flagbreak(2, failinside)

      if (H5Rcreate(&dimscales.s[i], ctx->hdf5fd, dimpath.s, H5R_OBJECT, -1) < 0) err1_flagbreak(2, failinside, "H5Rcreate")
      dimscales.len++;
    }

    if (failinside) 
       break;

    if (attach_dimension(ctxid, cvarpath.s, &dimscales) == -1) err_break(2)
    if(!g_no_log_info)
	    msg2("fill 2d dimension", cvarpath.s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (spaceid != -1) H5Sclose(spaceid);
    if (dsettype != -1) H5Tclose(dsettype);
    if (memtype != -1) H5Tclose(memtype);
    if (dsetid != -1) H5Dclose(dsetid);
    hobjreftalloc_free(&dimscales);
    challoc_free(&dimpath);
    challoc_free(&cvarpath);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief write CF attributes by the augmentation tool
 * 
 * @param ctxid      context identifier
 * @param eosobjname HDF-EOS5 object name
 * @param isgrid     Is this an HDF-EOS5 grid
 * @param isdim      Is this an HDF-EOS5 dimension scale dataset
 * @param isdatafield is this the data field 
          Swath has both geolocation and data fields. This parameter is useful when isgrid is not true and the user wants to check a field attribute
          If this parameter is set to 2 , it is also used to make_field_path for 2d lat/lon cvs. 
 
 * @param dimname    HDF-EOS5 dimension name
 * @param numattr    Number of attributes to be added
 * @param attrnames  Attribute name lists
 * @param attrvalues Attribute values
 * @return FAIL if failed, SUCCEED if successful
 *
 */
static int write_attributes(hid_t ctxid, const char *eosobjname, int isgrid, int isdim, int isdatafield, const char *dimname, int numattr, const char **attrnames, const char **attrvalues)
{
  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {
    int i = 0;
    int failinside = 0;
    context_t *ctx = get_context(ctxid);

    /* If this is an HDF-EOS5 dimensional scale, make the dimension name.*/
    if (isdim) {
      if (make_dimpath(ctxid, eosobjname, isgrid, dimname, &dimpath) == -1) err_break(2)
    }
    else { 
      /* 
         This will happen for the 2-D lat/lon grid case. So far we don't have a testing file for this. KY 2013-10-17 
         Pass isdatafield to_fieldpath. I don't know why it always passes 0(geofield) before. Correct it.  KY 2013-10-25 
         dimpath struct is the same as fieldpath struct regarding the name and length. So borrow here. */ 
    
      if (make_fieldpath(ctxid, eosobjname, isgrid, dimname, isdatafield, &dimpath) == -1) err_break(2)
    }

    if ((dsetid = H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err1_break(2, "H5DOPEN")

    for (i = 0; i < numattr; ++i) {
      if (write_attribute_string0(ctxid, dsetid, attrnames[i], attrvalues[i]) == -1) err2_flagbreak(2, failinside, "attribute", attrnames[i])
    }
    if (failinside == 1) 
       break;

    succeeded = 1;
  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
    challoc_free(&dimpath);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Add CF attributes to latitude and longitude for the 1D lat/lon HDF-EOS5 grid case
 * 
 * @param ctxid      context identifier
 * @param gridname   HDF-EOS5 grid name
 * @param islongitude  Is this HDF-EOS5 longitude 
 * @return FAIL if failed, SUCCEED if successful
 *
 */

int conform_cf_conventions_grid_1d(hid_t ctxid, const char *gridname, int islongitude)
{
  int succeeded = 0;

  do {
    const char *dimname = islongitude ? "XDim" : "YDim";
    const char *longname = islongitude ? "Longitude" : "Latitude";
    const char *unit = islongitude ? "degrees_east" : "degrees_north";
    const char *attrname1 = "long_name";
    const char *attrname2 = "units";

    int attr_exist = 0;

    /* Need to check if "units" and "long_name" exist.   */
    /* Check if long_name exists. If yes, keep the original long_name, don't rewrite */
    attr_exist = check_attribute_existence(ctxid,gridname,dimname,1,1,1,attrname1); 
    if (-1 == attr_exist) err2_break(2,"check_attribute_exist",attrname1)
    else if ( 0 == attr_exist) {
      const char *aname_list[] ={attrname1};
      const char *avalue_list[] = {longname};
      if(write_attributes(ctxid,gridname,1,1,1,dimname,1,aname_list,avalue_list) == -1) err2_break(2,"write_attributes",attrname1)
    }

    /* Always use CF units for lat/lon. */
    attr_exist = check_attribute_existence(ctxid,gridname,dimname,1,1,1,attrname2);
    if (-1 == attr_exist) err2_break(2,"check_attribute_exist",attrname2)
    else if ( 1 == attr_exist) {
      if(delete_attribute(ctxid,gridname,dimname,1,attrname2) == -1) err2_break(2,"delete_attribute",attrname2)
    }
    {
      const char *aname_list[] ={attrname2};
      const char *avalue_list[] = {unit};
      if(write_attributes(ctxid,gridname,1,1,1,dimname,1,aname_list,avalue_list) == -1) err2_break(2,"write_attributes",attrname1)
    }

    if(!g_no_log_info)
        msg2("conform to CF convention 1.5 section 5.1", dimname);

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief Add CF attributes to latitude and longitude for the 2D lat/lon HDF-EOS5 grid case
 * 
 * @param ctxid      context identifier
 * @param gridname   HDF-EOS5 grid name
 * @param islongitude  Is this HDF-EOS5 longitude 
 * @return FAIL if failed, SUCCEED if successful
 *
 */

int conform_cf_conventions_grid_2d_dim(hid_t ctxid, const char *gridname, int islongitude)
{
  int succeeded = 0;

  do {
    /* Only "axis" and "long_name' will be written. This is slightly different from
     * the CF Conventions in that "units" attribute is missing.
     */
    const char *dimname = islongitude ? "XDim" : "YDim";
    const char *axis = islongitude ? "X" : "Y";
    const char *longname = islongitude ? "x-coordinate in Cartesian system" : "y-coordinate in Cartesian system";
    const char *attrnames[2] = { "axis", "long_name" };
    const char *attrvalues[2] = { axis, longname };

    if (write_attributes(ctxid, gridname, 1, 1, 1,dimname, 2, attrnames, attrvalues) == -1) err_break(2)

    if(!g_no_log_info)
        msg2("conform to CF convention 1.5 section 5.2", dimname);

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief Add CF attributes to latitude and longitude for the 1D lat/lon HDF-EOS5 grid case
 * 
 * @param ctxid      context identifier
 * @param gridname   HDF-EOS5 grid name
 * @param islongitude  Is this HDF-EOS5 longitude 
 * @param coordinates  The CF "coordinates" string that consists of coordinate variable(in this case,lat/lon) names
 * @return FAIL if failed, SUCCEED if successful
 *
 */

int conform_cf_conventions_grid_2d_coordvar(hid_t ctxid, const char *gridname, int islongitude, challoc *coordinates)
{
  int succeeded = 0;

  do {
    /* Both "lon" and "lat" are two-dimensional coordinate variables that
     * this tool generates to handle two-dimensional latitude and longitude case
     * such as sinusoidal projection, polar stereographic projection, etc.
     */
    const char *cvname = islongitude ? "lon" : "lat";
    const char *longname = islongitude ? "Longitude" : "Latitude";
    const char *unit = islongitude ? "degrees_east" : "degrees_north";
    const char *attrnames[2] = { "long_name", "units" };
    const char *attrvalues[2] = { longname, unit };

    if (write_attributes(ctxid, gridname, 1, 0, 2,cvname, 2, attrnames, attrvalues) == -1) err_break(2)
    if (coordinates->len > 0) {
      if (!challoc_cats(coordinates, " ")) nomem1_break("coordinates")
    }
    if (!challoc_cats(coordinates, cvname)) nomem1_break("coordinates")

    if(!g_no_log_info)
        msg2("conform to CF convention 1.5 section 5.2", cvname);

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief Add CF attributes to latitude and longitude for the 1D lat/lon HDF-EOS5 grid case
 * 
 * @param ctxid      context identifier
 * @param eosobjname   HDF-EOS5 object name
 * @param isgrid       Is this an HDF-EOS5 grid
 * @param field        HDF-EOS5 field info.
 * @param coordinates  The CF "coordinates" string that consists of coordinate variable(in this case,lat/lon) names
 * @return FAIL if failed, SUCCEED if successful
 *
 */

int conform_cf_conventions_2d_data(hid_t ctxid, const char *eosobjname, int isgrid, const fieldinfo_t *field, const challoc *coordinates)
{
  hid_t dsetid = -1;
  challoc dimpath;
  int succeeded = 0;

  challoc_init(&dimpath);
  do {
    context_t *ctx = get_context(ctxid);

    if (make_fieldpath(ctxid, eosobjname, isgrid, field->name.s, 1, &dimpath) == -1) err_break(2)

    if ((dsetid = H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0) err1_break(2, "H5DOPEN")

    /* Need to check if the string needs NULL TERM. KY 2013-11-18 */
    if (write_attribute_string0(ctxid, dsetid, "coordinates", coordinates->s) == -1) err_break(2)
    /*if (write_attribute_string(ctxid, dsetid, "coordinates", coordinates->s) == -1) err_break(2)*/

    if(!g_no_log_info)
        msg2("conform to CF convention 1.5 section 5.2 (coordinates)", dimpath.s);

    succeeded = 1;
  } while (0);

  if (1) {
    if (dsetid != -1) H5Dclose(dsetid);
    challoc_free(&dimpath);
  }
  return succeeded ? 0 : -1;
}

/** 
 * @brief Fill in the data values for HDF5 swath dimensional scale
 * 
 * @param ctxid      context identifier
 * @param eosobjname HDF-EOS5 object name
 * @param dim        Can be used to retrieve HDF-EOS5 dim. info
 * @param aug_option index(1), file(2) or the default(0)
 * @return FAIL if failed, SUCCEED if successful
 *
 */

int fill_hdf5_swath_dimension(hid_t ctxid, const char *eosobjname, const dimension_t *dim, const int aug_option)
{
  challoc dimpath;
  int succeeded = 0;
  int failinside = 0;
  hid_t source=-1;
  hid_t tid=-1;
  hid_t spaceid=-1;
  hid_t  dest=-1;
  void *buf = NULL;

  challoc_init(&dimpath);
  do {
    int i = 0;

    context_t *ctx = get_context(ctxid);
    
    hsize_t dimsize = dim->dimsize;


    if (dimsize == H5S_UNLIMITED)
      dimsize = 0;

    if (make_dimpath(ctxid, eosobjname, 0, dim->name.s, &dimpath) == -1) err_break(2);
      
    switch(aug_option) {
      case 0:
    	{
              /* For the default option, we should NOT change the NAME. We should still
                  keep the netCDF predefined attribute.  KY 2013-10-17*/
    	  /*replace_dimscale_name_attr(ctxid, dim, dimpath);*/
    	  break;
    	}
      case 1:
     	{
     	  if(fill_dim_scale_with_index(ctxid, dim, dimpath)== -1){
          failinside = 1;
     	    break;
        }
        else 
          goto DONE;
     	}
      case 2:
    	{
    	  for(i=start_ptr_field; i<end_ptr_field; i++)
    	  {
    	    if(strcmp(dim->name.s, ctx->dimnames[i])==0)
    	      break;
    	  }
    	  if(i < end_ptr_field) /* [dim, field] */
    	  {
    	    if(fill_dim_scale_with_field(ctxid, eosobjname, dim, dimpath, 0, i) == -1) err_flagbreak(2,failinside)
          else
    	       goto DONE;
    	  } 
    
    	  for(i=start_ptr_file; i<end_ptr_file; i++) {
    	    if(strcmp(dim->name.s, ctx->dim_data_name[i]) == 0)
    	      break;
    	  }
    	  if(i < end_ptr_file) {/* [dim, file] */
    	    if(fill_dim_scale_with_data(ctxid, dim, dimpath,i) == -1) err_flagbreak(2,failinside)
          else 
    	        goto DONE;
    	  }
    
    	  for(i=start_ptr_index; i<end_ptr_index; i++) {
          if(strcmp(dim->name.s, ctx->dim_index_name[i]) == 0)
    	       break;
    	  }
    	  if(i < end_ptr_index) {/* [dim] */
    	    if(fill_dim_scale_with_index(ctxid, dim, dimpath) == -1) err_flagbreak(2,failinside)
          else 
            goto DONE;
       }
    	}/* end of case 2 */	  
    }/* end of switch */
DONE:
    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&dimpath);
    if(buf) 
      free(buf);
    if (spaceid != -1) 
       H5Sclose(spaceid);
    if (tid != -1) 
       H5Tclose(tid);
    if (dest != -1) 
       H5Dclose(dest);
    if (source != -1) 
       H5Dclose(source);
  }
 
  /* Temporary fix, subject to change */
  if (failinside == 1)
      succeeded = 0;
  return succeeded ? 0 : -1;
}

/* The following routine is defined elsewhere. Remove the routine. */
#if 0
int replace_dimscale_name_attr(hid_t ctxid, const dimension_t *dim, const challoc dimpath) {
  int succeeded=0;
  hid_t dest=-1;
  context_t *ctx;

  do {
    ctx = get_context(ctxid);
    /* Using HDF5 1.8 APIs. [06/25/2012 -LD] */
    if ((dest= H5DOPEN(ctx->hdf5fd, dimpath.s)) < 0)  err2_break(3, "open dataset", dimpath.s)
    if(H5Aexists(dest, "NAME") > 0) 
      if (H5Adelete(dest, "NAME")<0) err2_break(3, "del att", dimpath.s);
    if (write_attribute_string0(ctxid, dest, "NAME", dim->name.s) == -1) err_break(3);
    if(H5Dclose(dest) < 0) err2_break(3, "close dataset", dimpath.s);
    
    succeeded = 1;
    } while(0);

  return succeeded? 0: -1;
}
#endif

